//
//	Pelistina on Cocoa - PoCo -
//	Application 
//
//	Copyright (C) 2005-2018 KAENRYUU Koutoku.
//

#import "PoCoAppController.h"

#import "PoCoControllerFactory.h"
#import "PoCoGlobalObserver.h"
#import "PoCoEditInfo.h"
#import "PoCoPenStyle.h"
#import "PoCoTilePattern.h"
#import "PoCoMyDocument.h"
#import "PoCoPicture.h"

#import "PoCoPreferencesPanel.h"
#import "PoCoNewDocumentPanel.h"
#import "PoCoSubViewWindow.h"
#import "PoCoPaletteWindow.h"
#import "PoCoPenStyleWindow.h"
#import "PoCoLayerWindow.h"
#import "PoCoInformationWindow.h"
#import "PoCoToolbarWindow.h"
#import "PoCoColorPatternWindow.h"

// ѿ
static NSString *SUBVIEW_OPEN_NAME = @"PoCoSubViewWindowOpened";
static NSString *PALETTE_OPEN_NAME = @"PoCoPaletteWindowOpened";
static NSString *PENSTYLE_OPEN_NAME = @"PoCoPenStyleWindowOpened";
static NSString *LAYER_OPEN_NAME = @"PoCoLayerWindowOpened";
static NSString *INFORMATION_OPEN_NAME = @"PoCoInformationWindowOpened";
static NSString *TOOLBAR_OPEN_NAME = @"PoCoToolbarWindowOpened";
static NSString *COLORPATTERN_OPEN_NAME = @"PoCoColorPaternWindowOpened";
static NSString *DEFAULT_RESOLUTION = @"PoCoNewDocumentDefaultResolution";
static NSString *DEFAULT_COLOR = @"PoCoNewDocumentDefaultColor";

// ============================================================================
@implementation PoCoAppController

// --------------------------------------------------------- instance - private
//
// ɥľε
//
//  Call
//    name : user default ̾
//    type : ľ
//
//  Return
//    None
//
-(void)setOpenedState:(NSString *)name 
                 type:(BOOL)type
{
    [[NSUserDefaults standardUserDefaults] setBool:type
                                            forKey:name];

    return;
}


// ------------------------------------------------------------- class - public
//
// 
//
//  Call
//    None
//
//  Return
//    None
//
+(void)initialize
{
    NSMutableDictionary *dic;

    dic = [NSMutableDictionary dictionary];

    // ƽͤ
    [dic setObject:[NSNumber numberWithBool:YES] forKey:SUBVIEW_OPEN_NAME];
    [dic setObject:[NSNumber numberWithBool:YES] forKey:PALETTE_OPEN_NAME];
    [dic setObject:[NSNumber numberWithBool:YES] forKey:PENSTYLE_OPEN_NAME];
    [dic setObject:[NSNumber numberWithBool:YES] forKey:LAYER_OPEN_NAME];
    [dic setObject:[NSNumber numberWithBool:YES] forKey:TOOLBAR_OPEN_NAME];
    [dic setObject:[NSNumber numberWithBool:NO]  forKey:COLORPATTERN_OPEN_NAME];
    [dic setObject:[NSNumber numberWithBool:NO]  forKey:INFORMATION_OPEN_NAME];

    // default 
    [[NSUserDefaults standardUserDefaults] registerDefaults:dic];

    return;
}


// ---------------------------------------------------------- instance - public
//
// initialize
//
//  Call
//    None
//
//  Return
//    function            : 
//    willFinish_         : λ椫(instance ѿ)
//    preferencesPanel_   : Ķѥͥ(instance ѿ)
//    newDocumentPanel_   : ѥͥ(instance ѿ)
//    subViewWindow_      : (instance ѿ)
//    toolbarWindow_      : ġС(instance ѿ)
//    paletteWindow_      : ѥå(instance ѿ)
//    colorPatternWindow_ : 顼ѥ(instance ѿ)
//    layerWindow_        : 쥤䡼(instance ѿ)
//    penStyleWindow_     : ڥ(instance ѿ)
//    informationWindow_  : Խ(instance ѿ)
//    factory_            : Խ(instance ѿ)
//    observer_           : μ(instance ѿ)
//    editInfo_           : Խ(instance ѿ)
//    penStyle_           : ڥ(instance ѿ)
//    penSteadyStyle_     : ڥ(instance ѿ)
//    tilePattern_        : ѥ(instance ѿ)
//    tileSteadyPattern_  : 勵ѥ(instance ѿ)
//
-(id)init
{
    DPRINT((@"[PoCoAppController init]\n"));

    // super class ν
    self = [super init];

    // Ȥν
    if (self != nil) {
        self->willFinish_ = NO;

        self->preferencesPanel_ = nil;
        self->newDocumentPanel_ = nil;
        self->subViewWindow_ = nil;
        self->toolbarWindow_ = nil;
        self->paletteWindow_ = nil;
        self->colorPatternWindow_ = nil;
        self->layerWindow_ = nil;
        self->penStyleWindow_ = nil;
        self->informationWindow_ = nil;

        self->factory_ = nil;
        self->observer_ = nil;
        self->editInfo_ = nil;
        self->penStyle_ = nil;
        self->penSteadyStyle_ = nil;
        self->tilePattern_ = nil;
        self->tileSteadyPattern_ = nil;

        // Խ
        self->factory_ = [[PoCoControllerFactory alloc] init];
        if (self->factory_ == nil) {
            DPRINT((@"can't create PoCoControllerFactory\n"));
            [self release];
            self = nil;
            goto EXIT;
        }

        // μ
        self->observer_ = [[PoCoGlobalObserver alloc] init];
        if (self->observer_ == nil) {
            DPRINT((@"can't create PoCoGlobalObserver\n"));
            [self release];
            self = nil;
            goto EXIT;
        }

        // Խ
        self->editInfo_ = [[PoCoEditInfo alloc] init];
        if (self->editInfo_ == nil) {
            DPRINT((@"can't create PoCoEditInfo\n"));
            [self release];
            self = nil;
            goto EXIT;
        }

        // ڥ
        self->penStyle_ = [[PoCoPenStyle alloc] init];
        if (self->penStyle_ == nil) {
            DPRINT((@"can't create PoCoPenStyle\n"));
            [self release];
            self = nil;
            goto EXIT;
        }
        self->penSteadyStyle_ = [[PoCoPenSteadyStyle alloc] init];
        if (self->penSteadyStyle_ == nil) {
            DPRINT((@"can't create PoCoPenSteadyStyle\n"));
            [self release];
            self = nil;
            goto EXIT;
        }

        // ѥ
        self->tilePattern_ = [[PoCoTilePattern alloc] init];
        if (self->tilePattern_ == nil) {
            DPRINT((@"can't create PoCoTilePattern\n"));
            [self release];
            self = nil;
            goto EXIT;
        }
        self->tileSteadyPattern_ = [[PoCoTileSteadyPattern alloc] init];
        if (self->tileSteadyPattern_ == nil) {
            DPRINT((@"can't create PoCoPenSteadyPattern\n"));
            [self release];
            self = nil;
            goto EXIT;
        }
    }

EXIT:
    return self;
}


//
// deallocate
//
//  Call
//    None
//
//  Return
//    preferencesPanel_   : Ķѥͥ(instance ѿ)
//    newDocumentPanel_   : ѥͥ(instance ѿ)
//    subViewWindow_      : (instance ѿ)
//    toolbarWindow_      : ġС(instance ѿ)
//    paletteWindow_      : ѥå(instance ѿ)
//    colorPatternWindow_ : 顼ѥ(instance ѿ)
//    layerWindow_        : 쥤䡼(instance ѿ)
//    penStyleWindow_     : ڥ(instance ѿ)
//    informationWindow_  : Խ(instance ѿ)
//    factory_            : Խ(instance ѿ)
//    observer_           : μ(instance ѿ)
//    editInfo_           : Խ(instance ѿ)
//    penStyle_           : ڥ(instance ѿ)
//    penSteadyStyle_     : ڥ(instance ѿ)
//    tilePattern_        : ѥ(instance ѿ)
//    tileSteadyPattern_  : 勵ѥ(instance ѿ)
//
-(void)dealloc
{
    DPRINT((@"[PoCoAppController dealloc]\n"));

    // ƥѥͥ/ɥ release
    [self->penStyleWindow_ release];
    [self->paletteWindow_ release];
    [self->subViewWindow_ release];
    [self->layerWindow_ release];
    [self->informationWindow_ release];
    [self->toolbarWindow_ release];
    [self->colorPatternWindow_ release];
    [self->newDocumentPanel_ release];
    [self->preferencesPanel_ release];

    //  pointer  nil ˤ
    self->informationWindow_ = nil;
    self->penStyleWindow_ = nil;
    self->layerWindow_ = nil;
    self->colorPatternWindow_ = nil;
    self->paletteWindow_ = nil;
    self->toolbarWindow_ = nil;
    self->subViewWindow_ = nil;
    self->newDocumentPanel_ = nil;
    self->preferencesPanel_ = nil;

    // ѥ
    [self->tilePattern_ release];
    [self->tileSteadyPattern_ release];
    self->tilePattern_ = nil;
    self->tileSteadyPattern_ = nil;

    // ڥ
    [self->penStyle_ release];
    [self->penSteadyStyle_ release];
    self->penStyle_ = nil;
    self->penSteadyStyle_ = nil;

    // Խ
    [self->editInfo_ release];
    self->editInfo_ = nil;

    // μ
    [self->observer_ release];
    self->observer_ = nil;

    // Խ
    [self->factory_ release];
    self->factory_ = nil;

    // super class 
    [super dealloc];

    return;
}


//
// untitled window ޻
//  ưľζɥ޻ߤ
//
//  Call
//    sender : ׵θ(api )
//
//  Return
//    function : NO()
//
-(BOOL)applicationShouldOpenUntitledFile:(NSApplication *)sender
{
    DPRINT((@"[PoCoAppController applicationShouldOpenUntitledFile]\n"));

    if ([self->editInfo_ noOpenNewDocPanel]) {
        // ϳʤ
        ;
    } else {
        // ɥȺˤʤ
        [self showNewDocumentPanel:sender];
    }

    return NO;                          //  NO
}


//
// nib ɤ߹ޤ줿
//
//  Call
//    editInfo_ : Խ(instance ѿ)
//
//  Return
//    None
//
-(void)awakeFromNib
{
    NSUserDefaults *def;

    def = [NSUserDefaults standardUserDefaults];

    // ƥɥ򳫤
    if ([def boolForKey:SUBVIEW_OPEN_NAME]) {
        [self showSubViewWindow:nil];
    }
    if ([def boolForKey:PALETTE_OPEN_NAME]) {
        [self showPaletteWindow:nil];
    }
    if ([def boolForKey:PENSTYLE_OPEN_NAME]) {
        [self showPenStyleWindow:nil];
    }
    if ([def boolForKey:LAYER_OPEN_NAME]) {
        [self showLayerWindow:nil];
    }
    if ([def boolForKey:INFORMATION_OPEN_NAME]) {
        [self showInformationWindow:nil];
    }
    if ([def boolForKey:TOOLBAR_OPEN_NAME]) {
        [self showToolbarWindow:nil];
    }
    if ([def boolForKey:COLORPATTERN_OPEN_NAME]) {
        [self showColorPatternWindow:nil];
    }

    return;
}


//
// ץꥱ󤬽λ
//
//  Call
//    note : (api )
//
//  Return
//    willFinish_ : λ椫(instance ѿ)
//
-(void)applicationWillTerminate:(NSNotification *)note
{
    self->willFinish_ = YES;
   
    return;
}


//
// ġС
//
//  Call
//    toolbarWindow_ : ġС(instance ѿ)
//
//  Return
//    function : ġС
//
-(PoCoToolbarWindow *)toolbarWindow
{
    return self->toolbarWindow_;
}


//
// 
//
//  Call
//    sender : (api ѿ)
//
//  Return
//    None
//
-(IBAction)newWindowForTab:(id)sender
{
    id newDoc;

    //  document 
    newDoc = [[MyDocument alloc] initWidth:[[[[[NSDocumentController sharedDocumentController] currentDocument] picture] bitmapPoCoRect] width]
                                initHeight:[[[[[NSDocumentController sharedDocumentController] currentDocument] picture] bitmapPoCoRect] height]
                            initResolution:(int)([[NSUserDefaults standardUserDefaults] integerForKey:DEFAULT_RESOLUTION])
                              defaultColor:(unsigned char)([[NSUserDefaults standardUserDefaults] integerForKey:DEFAULT_COLOR])];
    if (newDoc == nil) {
        DPRINT((@"can't make new document.\n"));
    } else {
        [[NSDocumentController sharedDocumentController] addDocument:newDoc];
        [newDoc makeWindowControllers];
        [newDoc showWindows];
#if 0   // ⤽̾ΤǤʤΤǳĥҤͿ褦ʤ
        if ([[self->name_ stringValue] length] > 0) {
            nm = [[NSString stringWithString:[self->name_ stringValue]]
              stringByAppendingPathExtension:@"poco"];
            [newDoc setFileURL:[NSURL URLWithString:nm]];
        }
#endif  // 0

        // DocumentController ˰ϤΤ˺
        [newDoc release];
    }

    return;
}


// ------------------------- instance - public - ץꥱ̾μ
//
// λǤϤʤ
//
//  Call
//    willFinish_ : λ椫(instance ѿ)
//
//  Return
//    function : λǤϤʤ
//
-(BOOL)noFinishProc
{
    return !(self->willFinish_);
}


//
// Խμ
//
//  Call
//    factory_ : Խ(instance ѿ)
//
//  Return
//    function : Խ
//
-(PoCoControllerFactory *)factory
{
    return self->factory_;
}


//
// Խμ
//
//  Call
//    editInfo_ : Խ(instance ѿ)
//
//  Return
//    function : Խμ
//
-(PoCoEditInfo *)editInfo
{
    return self->editInfo_;
}


//
// ڥμ
//
//  Call
//    penStyle_ : ڥ(instance ѿ)
//
//  Return
//    function : ڥμ
//
-(PoCoPenStyle *)penStyle
{
    return self->penStyle_;
}


//
// ڥμ
//
//  Call
//    penSteadyStyle_ : ڥ(instance ѿ)
//
//  Return
//    function : ڥμ
//
-(PoCoPenSteadyStyle *)penSteadyStyle
{
    return self->penSteadyStyle_;
}


//
// ѥμ
//
//  Call
//    tilePattern_ : ѥ(instance ѿ)
//
//  Return
//    function : ѥμ
//
-(PoCoTilePattern *)tilePattern
{
    return self->tilePattern_;
}


//
// ˥塼򹹿
//
//  Call
//    menu           : menu(api ѿ)
//    editInfo_      : Խ(instance ѿ)
//    paletteWindow_ : ѥå(instance ѿ)
//
//  Return
//    function : 
//
-(BOOL)validateMenuItem:(NSMenuItem *)menu
{
    BOOL result;

    result = YES;

    if ([menu action] == @selector(showSubViewWindow:)) {
        [menu setState:(([[NSUserDefaults standardUserDefaults] boolForKey:SUBVIEW_OPEN_NAME]) ? NSOnState : NSOffState)];
    } else if ([menu action] == @selector(showPaletteWindow:)) {
        [menu setState:(([[NSUserDefaults standardUserDefaults] boolForKey:PALETTE_OPEN_NAME]) ? NSOnState : NSOffState)];
    } else if ([menu action] == @selector(showPenStyleWindow:)) {
        [menu setState:(([[NSUserDefaults standardUserDefaults] boolForKey:PENSTYLE_OPEN_NAME]) ? NSOnState : NSOffState)];
    } else if ([menu action] == @selector(showLayerWindow:)) {
        [menu setState:(([[NSUserDefaults standardUserDefaults] boolForKey:LAYER_OPEN_NAME]) ? NSOnState : NSOffState)];
    } else if ([menu action] == @selector(showInformationWindow:)) {
        [menu setState:(([[NSUserDefaults standardUserDefaults] boolForKey:INFORMATION_OPEN_NAME]) ? NSOnState : NSOffState)];
    } else if ([menu action] == @selector(showToolbarWindow:)) {
        [menu setState:(([[NSUserDefaults standardUserDefaults] boolForKey:TOOLBAR_OPEN_NAME]) ? NSOnState : NSOffState)];
    } else if ([menu action] == @selector(showColorPatternWindow:)) {
        [menu setState:(([[NSUserDefaults standardUserDefaults] boolForKey:COLORPATTERN_OPEN_NAME]) ? NSOnState : NSOffState)];
    } else if ([menu action] == @selector(changeFunction:)) {
        [menu setState:(((int)([self->editInfo_ drawModeType]) == [menu tag]) ? NSOnState : NSOffState)];
    } else if ([menu action] == @selector(changeDrawing:)) {
        [menu setState:(((int)([self->editInfo_ penStyleType]) == [menu tag]) ? NSOnState : NSOffState)];
    } else if ([menu action] == @selector(continuesLine:)) {
        [menu setState:(([self->editInfo_ continuationType]) ? NSOnState : NSOffState)];
    } else if ([menu action] == @selector(flipPattern:)) {
        [menu setState:(([self->editInfo_ flipType]) ? NSOnState : NSOffState)];
    } else if ([menu action] == @selector(useHandle:)) {
        [menu setState:(([self->editInfo_ useHandle]) ? NSOnState : NSOffState)];
    } else if ([menu action] == @selector(changePointMove:)) {
        [menu setState:(((int)([self->editInfo_ pointModeType]) == [menu tag]) ? NSOnState : NSOffState)];
    } else if ([menu action] == @selector(changePenSizeProp:)) {
        [menu setState:(((int)([self->editInfo_ sizePropType]) == [menu tag]) ? NSOnState : NSOffState)];
    } else if ([menu action] == @selector(changeDensityProp:)) {
        [menu setState:(((int)([self->editInfo_ densityPropType]) == [menu tag]) ? NSOnState : NSOffState)];
    } else if ([menu action] == @selector(normalTone:)) {
        [menu setState:(([self->editInfo_ atomizerType]) ? NSOffState : NSOnState)];
    } else if ([menu action] == @selector(halfTone:)) {
        [menu setState:(([self->editInfo_ atomizerType]) ? NSOnState : NSOffState)];
    } else if ([menu action] == @selector(changeAtomizerSkip:)) {
        [menu setState:(((int)([self->editInfo_ atomizerSkip]) == [menu tag]) ? NSOnState : NSOffState)];
    } else if ([menu action] == @selector(colorMixingRGB:)) {
        [menu setState:(([self->editInfo_ colorMode] == PoCoColorMode_RGB) ? NSOnState : NSOffState)];
    } else if ([menu action] == @selector(colorMixingHLS:)) {
        [menu setState:(([self->editInfo_ colorMode] == PoCoColorMode_HLS) ? NSOnState : NSOffState)];
    } else if ([menu action] == @selector(lockPalette:)) {
        [menu setState:(([self->editInfo_ lockPalette]) ? NSOnState : NSOffState)];
    } else if ([menu action] == @selector(paletteDisclose:)) {
        [menu setState:(([self->paletteWindow_ isDisclose]) ? NSOnState : NSOffState)];
    } else if ([menu action] == @selector(registorColorPattern:)) {
        result = (!([[self->editInfo_ selRect] empty]));
    }

    return result;
}


//
// 勵ѥμ
//
//  Call
//    tileSteadyPattern_ : 勵ѥ(instance ѿ)
//
//  Return
//    function : ѥμ
//
-(PoCoTileSteadyPattern *)tileSteadyPattern
{
    return self->tileSteadyPattern_;
}


// ------------------------------------- instance - public - ƥɥγ
//
// Ķѥͥ򳫤
//
//  Call
//    sender : (api ѿ)
//
//  Return
//    preferencesPanel_ : Ķѥͥ(instance ѿ)
//
-(IBAction)showPreferencesPanel:(id)sender
{
    // Τ϶ͭ
    if (self->preferencesPanel_ == nil) {
        self->preferencesPanel_ = [[PoCoPreferencesPanel alloc] init];
    }

    // ɥ򳫤
    [[self->preferencesPanel_ window] center];
    [self->preferencesPanel_ startWindow];
    [self->preferencesPanel_ showWindow:self];

    return;
}


//
// ѥͥ򳫤
//
//  Call
//    sender : (api )
//
//  Return
//    newDocumentPanel_ : ѥͥ(instance ѿ)
//
-(IBAction)showNewDocumentPanel:(id)sender
{
    // Τ϶ͭ
    if (self->newDocumentPanel_ == nil) {
        self->newDocumentPanel_ = [[PoCoNewDocumentPanel alloc] init];
    }

    // ɥ򳫤
    [[self->newDocumentPanel_ window] center];
    [self->newDocumentPanel_ startWindow];

    return;
}


//
// ȥɥ򳫤
//
//  Call
//    sender    : (api )
//    editInfo_ : Խ(instance ѿ)
//
//  Return
//    subViewWindow_ : ȥɥ(instance ѿ)
//
-(IBAction)showSubViewWindow:(id)sender
{
    // Τ϶ͭ
    if (self->subViewWindow_ == nil) {
        self->subViewWindow_ = [[PoCoSubViewWindow alloc] init];
    }

    // ɥ򳫤
    [self->subViewWindow_ showWindow:self];

    // Ƥ뤳Ȥ򵭲
    [self setOpenedState:SUBVIEW_OPEN_NAME
                    type:YES];

    // ָ
    [[self->subViewWindow_ window] setMovable:!([self->editInfo_ holdSubWindowPos])];

    return;
}


//
// ѥåȥɥ򳫤
//
//  Call
//    sender    : (api )
//    editInfo_ : Խ(instance ѿ)
//
//  Return
//    paletteWindow_ : ѥåȥɥ(instance ѿ)
//
-(IBAction)showPaletteWindow:(id)sender
{
    // Τ϶ͭ
    if (self->paletteWindow_ == nil) {
        self->paletteWindow_ = [[PoCoPaletteWindow alloc] init];
    }

    // ɥ򳫤
    [self->paletteWindow_ showWindow:self];

    // Ƥ뤳Ȥ򵭲
    [self setOpenedState:PALETTE_OPEN_NAME
                    type:YES];

    // ָ
    [[self->paletteWindow_ window] setMovable:!([self->editInfo_ holdSubWindowPos])];

    return;
}


//
// ڥ襦ɥ򳫤
//
//  Call
//    sender    : (api )
//    editInfo_ : Խ(instance ѿ)
//
//  Return
//    penStyleWindow_ : ڥ襦ɥ(instance ѿ)
//
-(IBAction)showPenStyleWindow:(id)sender
{
    // Τ϶ͭ
    if (self->penStyleWindow_ == nil) {
        self->penStyleWindow_ = [[PoCoPenStyleWindow alloc] init];
    }

    // ɥ򳫤
    [self->penStyleWindow_ showWindow:self];

    // Ƥ뤳Ȥ򵭲
    [self setOpenedState:PENSTYLE_OPEN_NAME
                    type:YES];

    // ָ
    [[self->penStyleWindow_ window] setMovable:!([self->editInfo_ holdSubWindowPos])];

    return;
}


//
// 쥤䡼ɥ򳫤
//
//  Call
//    sender    : (api )
//    editInfo_ : Խ(instance ѿ)
//
//  Return
//    layerWindow_ : 쥤䡼ɥ(instance ѿ)
//
-(IBAction)showLayerWindow:(id)sender
{
    // Τ϶ͭ
    if (self->layerWindow_ == nil) {
        self->layerWindow_ = [[PoCoLayerWindow alloc] init];
    }

    // ɥ򳫤
    [self->layerWindow_ showWindow:self];

    // Ƥ뤳Ȥ򵭲
    [self setOpenedState:LAYER_OPEN_NAME
                    type:YES];

    // ָ
    [[self->layerWindow_ window] setMovable:!([self->editInfo_ holdSubWindowPos])];

    return;
}


//
// Խ󥦥ɥ򳫤
//
//  Call
//    sender    : (api )
//    editInfo_ : Խ(instance ѿ)
//
//  Return
//    informationWindow_ : 󥦥ɥ(instance ѿ)
//
-(IBAction)showInformationWindow:(id)sender
{
    // Τ϶ͭ
    if (self->informationWindow_ == nil) {
        self->informationWindow_ = [[PoCoInformationWindow alloc] init];
    }

    // ɥ򳫤
    [self->informationWindow_ showWindow:self];

    // Ƥ뤳Ȥ򵭲
    [self setOpenedState:INFORMATION_OPEN_NAME
                    type:YES];

    // ָ
    [[self->informationWindow_ window] setMovable:!([self->editInfo_ holdSubWindowPos])];

    return;
}


//
// ġС򳫤
//
//  Call
//    sender    : (api )
//    editInfo_ : Խ(instance ѿ)
//
//  Return
//    toolbarWindow_ : ġС(instance ѿ)
//
-(IBAction)showToolbarWindow:(id)sender
{
    // Τ϶ͭ
    if (self->toolbarWindow_ == nil) {
        self->toolbarWindow_ = [[PoCoToolbarWindow alloc] init];
    }

    // ɥ򳫤
    [self->toolbarWindow_ showWindow:self];

    // Ƥ뤳Ȥ򵭲
    [self setOpenedState:TOOLBAR_OPEN_NAME
                    type:YES];

    // ָ
    [[self->toolbarWindow_ window] setMovable:!([self->editInfo_ holdSubWindowPos])];

    return;
}


//
// 顼ѥ󥦥ɥ򳫤
//
//  Call
//    sender    : (api )
//    editInfo_ : Խ(instance ѿ)
//
//  Return
//    colorPatternWindow_ : 顼ѥ(instance ѿ)
//
-(IBAction)showColorPatternWindow:(id)sender
{
    // Τ϶ͭ
    if (self->colorPatternWindow_ == nil) {
        self->colorPatternWindow_ = [[PoCoColorPatternWindow alloc] init];
    }

    // ɥ򳫤
    [self->colorPatternWindow_ showWindow:self];

    // Ƥ뤳Ȥ򵭲
    [self setOpenedState:COLORPATTERN_OPEN_NAME
                    type:YES];

    // ָ
    [[self->colorPatternWindow_ window] setMovable:!([self->editInfo_ holdSubWindowPos])];

    return;
}


//
// ȥɥĤ줿
//
//  Call
//    None
//
//  Return
//    None
//
-(void)closedSubViewWindow
{
    // ĤȤ򵭲
    if ([self noFinishProc]) {
        [self setOpenedState:SUBVIEW_OPEN_NAME
                        type:NO];
    }

    return;
}


//
// ѥåȥɥĤ줿
//
//  Call
//    None
//
//  Return
//    None
//
-(void)closedPaletteWindow
{
    // ĤȤ򵭲
    if ([self noFinishProc]) {
        [self setOpenedState:PALETTE_OPEN_NAME
                        type:NO];
    }

    return;
}


//
// ڥ襦ɥĤ줿
//
//  Call
//    None
//
//  Return
//    None
//
-(void)closedPenStyleWindow
{
    // ĤȤ򵭲
    if ([self noFinishProc]) {
        [self setOpenedState:PENSTYLE_OPEN_NAME
                        type:NO];
    }

    return;
}


//
// 쥤䡼ɥĤ줿
//
//  Call
//    None
//
//  Return
//    None
//
-(void)closedLayerWindow
{
    // ĤȤ򵭲
    if ([self noFinishProc]) {
        [self setOpenedState:LAYER_OPEN_NAME
                        type:NO];
    }

    return;
}


//
// Խ󥦥ɥĤ줿
//
//  Call
//    None
//
//  Return
//    None
//
-(void)closedInformationWindow
{
    // ĤȤ򵭲
    if ([self noFinishProc]) {
        [self setOpenedState:INFORMATION_OPEN_NAME
                        type:NO];
    }

    return;
}


//
// ġСĤ줿
//
//  Call
//    None
//
//  Return
//    None
//
-(void)closedToolbarWindow
{
    // ĤȤ򵭲
    if ([self noFinishProc]) {
        [self setOpenedState:TOOLBAR_OPEN_NAME
                        type:NO];
    }

    return;
}


//
// 顼ѥ󥦥ɥĤ줿
//
//  Call
//    None
//
//  Return
//    None
//
-(void)closedColorPatternWindow
{
    // ĤȤ򵭲
    if ([self noFinishProc]) {
        [self setOpenedState:COLORPATTERN_OPEN_NAME
                        type:NO];
    }

    return;
}


//
// ɥΰư޻
//
//  Call
//    type                : 
//                          YES : ޻
//                          NO  : 
//    subViewWindow_      : (instance ѿ)
//    toolbarWindow_      : ġС(instance ѿ)
//    paletteWindow_      : ѥå(instance ѿ)
//    colorPatternWindow_ : 顼ѥ(instance ѿ)
//    layerWindow_        : 쥤䡼(instance ѿ)
//    penStyleWindow_     : ڥ(instance ѿ)
//    informationWindow_  : Խ(instance ѿ)
//
//  Return
//    None
//
-(void)subWindowHold:(BOOL)type
{
    type = !(type);

    // ڤؤ
    [[self->subViewWindow_      window] setMovable:type];
    [[self->toolbarWindow_      window] setMovable:type];
    [[self->paletteWindow_      window] setMovable:type];
    [[self->colorPatternWindow_ window] setMovable:type];
    [[self->layerWindow_        window] setMovable:type];
    [[self->penStyleWindow_     window] setMovable:type];
    [[self->informationWindow_  window] setMovable:type];

    return;
}


// ------------------------------------- instance - public - Խǽڤؤ
//
// 赡ǽ
//
//  Call
//    sender         : (api )
//    toolbarWindow_ : ġС(instance ѿ)
//
//  Return
//    None
//
-(IBAction)changeFunction:(id)sender
{
    [self->toolbarWindow_ drawMode:sender];

    return;
}


//
// ڥ()
//
//  Call
//    sender         : (api )
//    toolbarWindow_ : ġС(instance ѿ)
//
//  Return
//    None
//
-(IBAction)changeDrawing:(id)sender
{
    [self->toolbarWindow_ penStyle:sender];

    return;
}


// --------------------------------- instance - public - Խǽڤؤ
//
// Ϣ³/Ϣ³
//
//  Call
//    sender         : (api )
//    toolbarWindow_ : ġС(instance ѿ)
//
//  Return
//    None
//
-(IBAction)continuesLine:(id)sender
{
    [self->toolbarWindow_ continuation:sender];

    return;
}


//
// ѥؤ
//
//  Call
//    sender         : (api )
//    toolbarWindow_ : ġС(instance ѿ)
//
//  Return
//    None
//
-(IBAction)flipPattern:(id)sender
{
    [self->toolbarWindow_ flip:sender];

    return;
}


//
// ϥɥ̵ͭ
//
//  Call
//    sender         : (api )
//    toolbarWindow_ : ġС(instance ѿ)
//
//  Return
//    None
//
-(IBAction)useHandle:(id)sender
{
    [self->toolbarWindow_ handle:sender];

    return;
}


//
// ()
//
//  Call
//    sender         : (api )
//    toolbarWindow_ : ġС(instance ѿ)
//
//  Return
//    None
//
-(IBAction)changePointMove:(id)sender
{
    [self->toolbarWindow_ pointMode:sender];

    return;
}


//
// ɮ()
//
//  Call
//    sender         : (api )
//    toolbarWindow_ : ġС(instance ѿ)
//
//  Return
//    None
//
-(IBAction)changePenSizeProp:(id)sender
{
    [self->toolbarWindow_ penSize:sender];

    return;
}


//
// ɮ(ǻ)
//
//  Call
//    sender         : (api )
//    toolbarWindow_ : ġС(instance ѿ)
//
//  Return
//    None
//
-(IBAction)changeDensityProp:(id)sender
{
    [self->toolbarWindow_ density:sender];

    return;
}


//
// ̸̾᤭
//
//  Call
//    sender         : (api )
//    toolbarWindow_ : ġС(instance ѿ)
//
//  Return
//    None
//
-(IBAction)normalTone:(id)sender
{
    [self->toolbarWindow_ normalTone:sender];

    return;
}


//
// Ⱦ̸᤭
//
//  Call
//    sender         : (api )
//    toolbarWindow_ : ġС(instance ѿ)
//
//  Return
//    None
//
-(IBAction)halfTone:(id)sender
{
    [self->toolbarWindow_ halfTone:sender];

    return;
}


//
// ̸᤭ư
//
//  Call
//    sender         : (api )
//    toolbarWindow_ : ġС(instance ѿ)
//
//  Return
//    None
//
-(IBAction)changeAtomizerSkip:(id)sender
{
    [self->toolbarWindow_ atomizerSkip:sender];

    return;
}


// ------------------------ instance - public - ڥ/ѥڤؤ
//
// Υڥ
//
//  Call
//    sender          : (api )
//    penStyleWindow_ : ڥ襦ɥ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)nextPenStyle:(id)sender
{
    [self->penStyleWindow_ nextPenStyle];

    return;
}


//
// Υڥ
//
//  Call
//    sender          : (api )
//    penStyleWindow_ : ڥ襦ɥ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)prevPenStyle:(id)sender
{
    [self->penStyleWindow_ prevPenStyle];

    return;
}


//
// Υڥ
//
//  Call
//    sender          : (api )
//    penStyleWindow_ : ڥ襦ɥ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)nextPenSize:(id)sender
{
    [self->penStyleWindow_ nextPenSize];

    return;
}


//
// Υڥ
//
//  Call
//    sender          : (api )
//    penStyleWindow_ : ڥ襦ɥ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)prevPenSize:(id)sender
{
    [self->penStyleWindow_ prevPenSize];

    return;
}


//
// Υѥ
//
//  Call
//    sender          : (api )
//    penStyleWindow_ : ڥ襦ɥ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)nextTilePattern:(id)sender
{
    [self->penStyleWindow_ nextTilePattern];

    return;
}


//
// Υѥ
//
//  Call
//    sender          : (api )
//    penStyleWindow_ : ڥ襦ɥ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)prevTilePattern:(id)sender
{
    [self->penStyleWindow_ prevTilePattern];

    return;
}


//
// ǻ٤ѹ
//
//  Call
//    sender          : (api )
//    penStyleWindow_ : ڥ襦ɥ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)changeDensity:(id)sender
{
    [self->penStyleWindow_ addDensity:(int)([sender tag])];

    return;
}


// ------------------------------------------- instance - public - ѥåԽ
//
// ѥåȸ
//
//  Call
//    sender         : (api )
//    paletteWindow_ : ѥå(instance ѿ)
//
//  Return
//    None
//
-(IBAction)lockPalette:(id)sender
{
    [self->paletteWindow_ lockPalette:sender];

    return;
}


//
// ɥȽ
//
//  Call
//    sender         : (api )
//    paletteWindow_ : ѥå(instance ѿ)
//
//  Return
//    None
//
-(IBAction)paletteDisclose:(id)sender
{
    [self->paletteWindow_ changeDisclose:sender];

    return;
}


//
// RGB 黻
//
//  Call
//    sender         : (api )
//    paletteWindow_ : ѥå(instance ѿ)
//
//  Return
//    None
//
-(IBAction)colorMixingRGB:(id)sender
{
    [self->paletteWindow_ changeMode:sender];

    return;
}


//
// HLS 黻
//
//  Call
//    sender         : (api )
//    paletteWindow_ : ѥå(instance ѿ)
//
//  Return
//    None
//
-(IBAction)colorMixingHLS:(id)sender
{
    [self->paletteWindow_ changeMode:sender];

    return;
}


//
// ѥå°
//
//  Call
//    sender         : (api )
//    paletteWindow_ : ѥå(instance ѿ)
//
//  Return
//    None
//
-(IBAction)alternateColorSetting:(id)sender
{
    [self->paletteWindow_ raiseAttributeSheet:sender];

    return;
}


//
// ѥåȥǡ
//
//  Call
//    sender         : (api )
//    paletteWindow_ : ѥå(instance ѿ)
//
//  Return
//    None
//
-(IBAction)createGradation:(id)sender
{
    [self->paletteWindow_ raiseGradationSheet:sender];

    return;
}


//
// 
//
//  Call
//    sender         : (api )
//    paletteWindow_ : ѥå(instance ѿ)
//
//  Return
//    None
//
-(IBAction)exchangeColor:(id)sender
{
    [self->paletteWindow_ raiseExchangeSheet:sender];

    return;
}


//
// ʣ
//
//  Call
//    sender         : (api )
//    paletteWindow_ : ѥå(instance ѿ)
//
//  Return
//    None
//
-(IBAction)pastePalette:(id)sender
{
    [self->paletteWindow_ raisePasteSheet:sender];

    return;
}


// ------------------------------------------- instance - public - 쥤䡼Խ
//
// 쥤䡼ɲ
//
//  Call
//    sender       : (api )
//    layerWindow_ : 쥤䡼(instance ѿ)
//
//  Return
//    None
//
-(IBAction)addBitmapLayer:(id)sender
{
    [self->layerWindow_ newBitmapLayer:sender];

    return;
}


//
// ʸ쥤䡼ɲ
//
//  Call
//    sender       : (api )
//    layerWindow_ : 쥤䡼(instance ѿ)
//
//  Return
//    None
//
-(IBAction)addStringLayer:(id)sender
{
    [self->layerWindow_ newStringLayer:sender];

    return;
}


//
// 쥤䡼
//
//  Call
//    sender       : (api )
//    layerWindow_ : 쥤䡼(instance ѿ)
//
//  Return
//    None
//
-(IBAction)deleteLayer:(id)sender
{
    [self->layerWindow_ deleteLayer:sender];

    return;
}


//
// ɽ쥤䡼
//
//  Call
//    sender       : (api )
//    layerWindow_ : 쥤䡼(instance ѿ)
//
//  Return
//    None
//
-(IBAction)unificateLayer:(id)sender
{
    [self->layerWindow_ unificateLayer:sender];

    return;
}


// ------------------------------------------------- instance - public - ¾
//
// 顼ѥϿ
//
//  Call
//    sender              : (api )
//    colorPatternWindow_ : 顼ѥ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)registorColorPattern:(id)sender
{
    [self->colorPatternWindow_ setPattern:sender];

    return;
}


// -------------------------------------------------- instance - public - إ
//
// PoCo Υۡڡ
//
//  Call
//    sender : (api )
//
//  Return
//    None
//
-(IBAction)visitPoCoHomepage:(id)sender
{
    if ([[[NSLocale currentLocale] localeIdentifier] isEqualToString:@"ja_JP"]) {
        [[NSWorkspace sharedWorkspace] openURL:[NSURL URLWithString:@"http://www.poco256.org/"]];
    } else {
        [[NSWorkspace sharedWorkspace] openURL:[NSURL URLWithString:@"http://www.poco256.org/en/"]];
    }

    return;
}


//
// PoCo Υޥ˥奢
//
//  Call
//    sender : (api )
//
//  Return
//    None
//
-(IBAction)visitPoCoManualPage:(id)sender
{
    if ([[[NSLocale currentLocale] localeIdentifier] isEqualToString:@"ja_JP"]) {
        [[NSWorkspace sharedWorkspace] openURL:[NSURL URLWithString:@"http://www.poco256.org/man/man.html"]];
    } else {
        [[NSWorkspace sharedWorkspace] openURL:[NSURL URLWithString:@"http://www.poco256.org/en/man/man.html"]];
    }

    return;
}

@end
