//
//	Pelistina on Cocoa - PoCo -
//	
//
//	Copyright (C) 2005-2017 KAENRYUU Koutoku.
//

#import "PoCoControllerEditterBase.h"

#import "PoCoControllerFactory.h"
#import "PoCoControllerPictureBitmapReplacer.h"

#import "PoCoCalcCurveWithPoints.h"
#import "PoCoCalcRotation.h"

// ============================================================================
@implementation PoCoControllerEditterBase

// ---------------------------------------------------------- instance - public
//
// initialize
//
//  Call
//    pict   : Խоݲ
//    info   : Խ
//    undo   : ä
//    eraser : äѲ
//    buf    : ݻ
//    pen    : ڥ
//    tile   : ѥ
//    idx    : оݥ쥤䡼ֹ
//
//  Return
//    function       : 
//    penStyle_      : ڥ(instance ѿ)
//    tilePattern_   : ѥ(instance ѿ)
//    eraserPattern_ : äѲ(instance ѿ)
//    colorBuffer_   : ݻ(instance ѿ)
//    index_         : оݥ쥤䡼ֹ(instance ѿ)
//    layer_         : оݥ쥤䡼(instance ѿ)
//    drawRect_      : ΰ(instance ѿ)
//    undoRect_      : äΰ(instance ѿ)
//    undoBitmap_    : ä(instance ѿ)
//    drawPattern_   : ѥ(instance ѿ)
//    drawPen_       : ѥڥ(instance ѿ)
//    drawTile_      : ѥѥ(instance ѿ)
//
-(id)initPict:(PoCoPicture *)pict
     withInfo:(PoCoEditInfo *)info
     withUndo:(NSUndoManager *)undo
   withEraser:(PoCoColorPattern *)eraser
   withBuffer:(PoCoColorBuffer *)buf
      withPen:(PoCoPenStyle *)pen
     withTile:(PoCoTilePattern *)tile
      atIndex:(int)idx
{
//    DPRINT((@"[PoCoControllerEditterBase init]\n"));

    // super class 
    self = [super initPict:pict
                  withInfo:info
                  withUndo:undo];

    // Ȥ
    if (self != nil) {
        self->penStyle_ = pen;
        self->tilePattern_ = tile;
        self->eraserPattern_ = eraser;
        self->colorBuffer_ = buf;
        self->index_ = idx;
        self->layer_ = nil;
        self->drawRect_ = nil;
        self->undoRect_ = nil;
        self->undoBitmap_ = nil;
        self->drawPattern_ = nil;
        self->drawPen_ = nil;
        self->drawTile_ = nil;

        // ƴ retain
        [self->penStyle_ retain];
        [self->tilePattern_ retain];
        [self->eraserPattern_ retain];
        [self->colorBuffer_ retain];

        // оݥ쥤䡼
        self->layer_ = [self->picture_ layer:self->index_];
        if (self->layer_ != nil) {
            [self->layer_ retain];
        } else {
            DPRINT((@"can't get target layer.\n"));
            [self release];
            self = nil;
        }
    }

    return self;
}


//
// deallocate
//
//  Call
//    None
//
//  Return
//    penStyle_      : ڥ(instance ѿ)
//    tilePattern_   : ѥ(instance ѿ)
//    eraserPattern_ : äѲ(instance ѿ)
//    colorBuffer_   : ݻ(instance ѿ)
//    layer_         : оݥ쥤䡼(instance ѿ)
//    drawRect_      : ΰ(instance ѿ)
//    undoRect_      : äΰ(instance ѿ)
//    undoBitmap_    : ä(instance ѿ)
//    drawPattern_   : ѥ(instance ѿ)
//    drawPen_       : ѥڥ(instance ѿ)
//    drawTile_      : ѥѥ(instance ѿ)
//
-(void)dealloc
{
//    DPRINT((@"[PoCoControllerEditterBase dealloc]\n"));

    // 񸻤
    [self->penStyle_ release];
    [self->tilePattern_ release];
    [self->eraserPattern_ release];
    [self->colorBuffer_ release];
    [self->layer_ release];
    [self->drawRect_ release];
    [self->undoRect_ release];
    [self->undoBitmap_ release];
    [self->drawPattern_ release];
    [self->drawPen_ release];
    [self->drawTile_ release];
    self->penStyle_ = nil;
    self->tilePattern_ = nil;
    self->eraserPattern_ = nil;
    self->colorBuffer_ = nil;
    self->layer_ = nil;
    self->drawRect_ = nil;
    self->undoRect_ = nil;
    self->undoBitmap_ = nil;
    self->drawPattern_ = nil;
    self->drawPen_ = nil;
    self->drawTile_ = nil;

    // super class 
    [super dealloc];

    return;
}


//
// ϰϤ(ϰ)
//
//  Call
//    r        : 
//    picture_ : Խоݲ( instance ѿ)
//
//  Return
//    r : 
//
-(void)correctRect:(PoCoRect *)r
{
    PoCoRect *pr;

    pr = [self->picture_ bitmapPoCoRect];
    if ([r left] < [pr left]) {
        [r setLeft:[pr left]];
    }
    if ([r right] >= [pr right]) {
        [r setRight:[pr right]];
    }
    if ([r top] < [pr top]) {
        [r setTop:[pr top]];
    }
    if ([r bottom] >= [pr bottom]) {
        [r setBottom:[pr bottom]];
    }
    [pr release];

    return;
}


//
// Խ
//  ¿ŸƤӽФػ(⤽¿Ť˸Ƥ֤褦ʼˤϤʤʤϤ)
//
//  Call
//    dr             : ΰ
//    un             : äΰ
//    picture_       : Խоݲ( instance ѿ)
//    editInfo_      : Խ( instance ѿ)
//    eraserPattern_ : äѲ(instance ѿ)
//    index_         : оݥ쥤䡼ֹ(instance ѿ)
//    layer_         : оݥ쥤䡼(instance ѿ)
//
//  Return
//    function     : 
//    penStyle_    : ڥ(instance ѿ)
//    tilePattern_ : ѥ(instance ѿ)
//    drawRect_    : ΰ(instance ѿ)
//    undoRect_    : äΰ(instance ѿ)
//    undoBitmap_  : ä(instance ѿ)
//    drawPattern_ : ѥ(instance ѿ)
//    drawPen_     : ѥڥ(instance ѿ)
//    drawTile_    : ѥѥ(instance ѿ)
//
-(BOOL)startEdit:(PoCoRect *)dr
    withUndoRect:(PoCoRect *)ur
{
    PoCoSelColor *selColor;             // 򿧾

    // ΰ򵭲
    self->drawRect_ = dr;
    self->undoRect_ = ur;
    [self->drawRect_ retain];
    [self->undoRect_ retain];

    // äѲ
    if ([self->editInfo_ enableUndo]) {
        self->undoBitmap_ = [[self->layer_ bitmap] getBitmap:self->undoRect_];
    }           

    // ѥڥ
    if (self->penStyle_ != nil) {
        self->drawPen_ = [self->penStyle_ pattern:[self->editInfo_ penNumber]];
        [self->drawPen_ retain];
    }

    // ѥѥ
    if (self->tilePattern_ != nil) {
        self->drawTile_ = [self->tilePattern_ pattern:[self->editInfo_ tileNumber]];
        [self->drawTile_ retain];
    }

    // ѥ
    selColor = [self->editInfo_ selColor];
    if (([self->editInfo_ eraserType]) && (self->eraserPattern_ != nil)) {
        // β(alloc ϤʤΤ retain Ƥ)
        self->drawPattern_ = self->eraserPattern_;
        [self->drawPattern_ retain];
    } else if (([selColor isUnder]) && (self->index_ > 0)) {
        // ̥쥤䡼
        self->drawPattern_ = [[PoCoColorPattern alloc]
                                 initWithBitmap:[[self->picture_ layer:(self->index_ - 1)] bitmap]];
    } else if ([selColor isPattern]) {
        // 顼ѥ(alloc ϤʤΤ retain Ƥ)
        self->drawPattern_ = [self->picture_ colpat:[selColor num]];
        [self->drawPattern_ retain];
    } else {
        // ̾ο
        self->drawPattern_ = [[PoCoColorPattern alloc]
                                    initWidth:1
                                   initHeight:1
                                 defaultColor:[selColor num]];
    }

    return YES;                         //  YES
}


//
// Խλ
//
//  Call    
//    name         : ä̾ʸ
//    undoManager_ : ä( instance ѿ)
//    factory_     : Խ( instance ѿ)
//    index_       : оݥ쥤䡼ֹ(instance ѿ)
//    drawRect_    : ΰ(instance ѿ)
//    undoRect_    : äΰ(instance ѿ)
//    undoBitmap_  : ä(instance ѿ)
//
//  Return  
//    undoManager_ : ä( instance ѿ)
//          
-(void)endEdit:(NSString *)name
{
    // äϿ
    if (self->undoBitmap_ != nil) {
        [[self->undoManager_ prepareWithInvocationTarget:self->factory_]
            createPictureBitmapReplacer:YES
                                  layer:self->index_
                                   rect:self->undoRect_
                                 bitmap:self->undoBitmap_
                                   name:name];
        [super setUndoName:name];
    }

    // Խ
    [self postNotify:self->drawRect_];

    return;
}


//
// Խ
//
//  Call    
//    r      : ΰ
//    index_ : оݥ쥤䡼ֹ(instance ѿ)
//
//  Return  
//    None
//
-(void)postNotify:(PoCoRect *)r
{
    PoCoEditPictureNotification *note;

    note = [[PoCoEditPictureNotification alloc] initWithRect:r
                                                     atIndex:self->index_];
    [[NSNotificationCenter defaultCenter] postNotificationName:PoCoEditPicture
                                                        object:note];
    [note release];

    return;
}


//
// Խ
//  undo ȼʤ
//
//  Call    
//    r      : ΰ
//    index_ : оݥ쥤䡼ֹ(instance ѿ)
//
//  Return  
//    None
//
-(void)postNotifyNoEdit:(PoCoRect *)r
{
    PoCoEditPictureNotification *note;

    note = [[PoCoEditPictureNotification alloc] initWithRect:r
                                                     atIndex:self->index_];
    [[NSNotificationCenter defaultCenter] postNotificationName:PoCoRedrawPicture
                                                        object:note];
    [note release];

    return;
}

@end




// ============================================================================
@implementation PoCoControllerBoxEditterBase

// ---------------------------------------------------------- instance - public
//
// initialize
//
//  Call
//    pict   : Խоݲ
//    info   : Խ
//    undo   : ä
//    eraser : äѲ
//    buf    : ݻ
//    pen    : ڥ
//    tile   : ѥ
//    s      : (濴)
//    e      : 轪(濴)
//    o      : 
//    idx    : оݥ쥤䡼ֹ
//
//  Return
//    function  : 
//    startPos_ : (濴)(instance ѿ)
//    endPos_   : 轪(濴)(instance ѿ)
//    ortPos_   : (instance ѿ)
//    diagonal_ : žгѤ(instance ѿ)
//
-(id)initPict:(PoCoPicture *)pict
     withInfo:(PoCoEditInfo *)info
     withUndo:(NSUndoManager *)undo
   withEraser:(PoCoColorPattern *)eraser
   withBuffer:(PoCoColorBuffer *)buf
      withPen:(PoCoPenStyle *)pen
     withTile:(PoCoTilePattern *)tile
    withStart:(PoCoPoint *)s
      withEnd:(PoCoPoint *)e
  ifAnyOrient:(PoCoPoint *)o
      atIndex:(int)idx;
{
//    DPRINT((@"[PoCoControllerBoxEditterBase init]\n"));

    // super class 
    self = [super initPict:pict
                  withInfo:info
                  withUndo:undo
                withEraser:eraser
                withBuffer:buf
                   withPen:pen
                  withTile:tile
                   atIndex:idx];

    // Ȥ
    if (self != nil) {
        self->diagonal_ = NO;

        // ʣ(塢֤ݻ)
        self->startPos_ = [[PoCoPoint alloc] initX:MIN([s x], [e x])
                                             initY:MIN([s y], [e y])];
        self->endPos_   = [[PoCoPoint alloc] initX:MAX([s x], [e x])
                                             initY:MAX([s y], [e y])];
        if (o != nil) {
            self->ortPos_ = [[PoCoPoint alloc] initX:[o x]
                                               initY:[o y]];
            if (([self->startPos_ isEqualPos:s]) ||
                ([self->startPos_ isEqualPos:e]) ||
                ([self->endPos_ isEqualPos:s]) ||
                ([self->endPos_ isEqualPos:e])) {
                ;
            } else {
                self->diagonal_ = YES;
            }
        } else {
            self->ortPos_ = nil;
        }
    }

    return self;
}


//
// deallocate
//
//  Call
//    None
//
//  Return
//    startPos_ : (濴)(instance ѿ)
//    endPos_   : 轪(濴)(instance ѿ)
//    ortPos_   : (instance ѿ)
//
-(void)dealloc
{
//    DPRINT((@"[PoCoControllerBoxEditterBase dealloc]\n"));

    // 񸻤
    [self->startPos_ release];
    [self->endPos_ release];
    [self->ortPos_ release];
    self->startPos_ = nil;
    self->endPos_ = nil;
    self->ortPos_ = nil;

    // super class 
    [super dealloc];

    return;
}


//
// ͽΰλ
//  ֤ͤ release ƤӽФǹԤ
//
//  Call
//    corr      : 
//    picture_  : Խоݲ( instance ѿ)
//    startPos_ : (濴)(instance ѿ)
//    endPos_   : 轪(濴)(instance ѿ)
//    diagonal_ : žгѤ(instance ѿ)
//
//  Return
//    function : ΰ
//
-(PoCoRect *)calcArea:(int)corr
{
    PoCoRect *r;
    PoCoCalcRotationForBox *rot;

    // žɸ򻻽
    rot = [[PoCoCalcRotationForBox alloc] initWithStartPos:self->startPos_
                                                withEndPos:self->endPos_
                                               ifAnyOrient:self->ortPos_];
    [rot calc:self->diagonal_];

    r = [[PoCoRect alloc] initLeft:INT_MAX
                           initTop:INT_MAX
                         initRight:INT_MIN
                        initBottom:INT_MIN];

    // 
    [r setLeft:MIN([r left], [[rot leftTop] x])];
    [r setLeft:MIN([r left], [[rot rightTop] x])];
    [r setLeft:MIN([r left], [[rot leftBottom] x])];
    [r setLeft:MIN([r left], [[rot rightBottom] x])];

    // 
    [r setTop:MIN([r top], [[rot leftTop] y])];
    [r setTop:MIN([r top], [[rot rightTop] y])];
    [r setTop:MIN([r top], [[rot leftBottom] y])];
    [r setTop:MIN([r top], [[rot rightBottom] y])];

    // 
    [r setRight:MAX([r right], [[rot leftTop] x])];
    [r setRight:MAX([r right], [[rot rightTop] x])];
    [r setRight:MAX([r right], [[rot leftBottom] x])];
    [r setRight:MAX([r right], [[rot rightBottom] x])];

    // 
    [r setBottom:MAX([r bottom], [[rot leftTop] y])];
    [r setBottom:MAX([r bottom], [[rot rightTop] y])];
    [r setBottom:MAX([r bottom], [[rot leftBottom] y])];
    [r setBottom:MAX([r bottom], [[rot rightBottom] y])];

    // ̤ȿ
    [r setLeft:([r left] - corr)];
    [r setTop:([r top] - corr)];
    [r setRight:([r right] + corr)];
    [r setBottom:([r bottom] + corr)];

    [rot release];
    return r;
}


//
// Խ
//
//  Call
//    dc : ΰ
//    uc : äΰ
//
//  Return
//    function : 
//
-(BOOL)startEdit:(int)dc
    withUndoCorr:(int)uc
{
    BOOL result;
    PoCoRect *dr;
    PoCoRect *ur;

    // ϰϤ
    dr = [self calcArea:dc];
    ur = [self calcArea:uc];
    [super correctRect:dr];
    [super correctRect:ur];

    // super class ز
    result = [super startEdit:dr
                 withUndoRect:ur];

    // ϰϤ
    [dr release];
    [ur release];
 
    return result;
}


//
// Խλ
//
//  Call
//    name : ä̾ʸ
//
//  Return
//    None
//
-(void)endEdit:(NSString *)name
{
    // super class ز
    [super endEdit:name];

    return;
}

@end




// ============================================================================
@implementation PoCoControllerEllipseEditterBase

// ---------------------------------------------------------- instance - public
//
// initialize
//
//  Call
//    pict   : Խоݲ
//    info   : Խ
//    undo   : ä
//    eraser : äѲ
//    buf    : ݻ
//    pen    : ڥ
//    tile   : ѥ
//    s      : (濴)
//    e      : 轪(濴)
//    o      : 
//    idx    : оݥ쥤䡼ֹ
//
//  Return
//    function  : 
//    startPos_ : (濴)(instance ѿ)
//    endPos_   : 轪(濴)(instance ѿ)
//    ortPos_   : (instance ѿ)
//
-(id)initPict:(PoCoPicture *)pict
     withInfo:(PoCoEditInfo *)info
     withUndo:(NSUndoManager *)undo
   withEraser:(PoCoColorPattern *)eraser
   withBuffer:(PoCoColorBuffer *)buf
      withPen:(PoCoPenStyle *)pen
     withTile:(PoCoTilePattern *)tile
    withStart:(PoCoPoint *)s
      withEnd:(PoCoPoint *)e
  ifAnyOrient:(PoCoPoint *)o
      atIndex:(int)idx;
{
//    DPRINT((@"[PoCoControllerEllipseEditterBase init]\n"));

    // super class 
    self = [super initPict:pict
                  withInfo:info
                  withUndo:undo
                withEraser:eraser
                withBuffer:buf
                   withPen:pen
                  withTile:tile
                   atIndex:idx];

    // Ȥ
    if (self != nil) {
        // ʣ
        self->startPos_ = [[PoCoPoint alloc] initX:[s x]
                                             initY:[s y]];
        self->endPos_ = [[PoCoPoint alloc] initX:[e x]
                                           initY:[e y]];
        if (o != nil) {
            self->ortPos_ = [[PoCoPoint alloc] initX:[o x]
                                               initY:[o y]];
        } else {
            self->ortPos_ = nil;
        }
    }

    return self;
}


//
// deallocate
//
//  Call
//    None
//
//  Return
//    startPos_ : (濴)(instance ѿ)
//    endPos_   : 轪(濴)(instance ѿ)
//    ortPos_   : (instance ѿ)
//
-(void)dealloc
{
//    DPRINT((@"[PoCoControllerEllipseEditterBase dealloc]\n"));

    // 񸻤
    [self->startPos_ release];
    [self->endPos_ release];
    [self->ortPos_ release];
    self->startPos_ = nil;
    self->endPos_ = nil;
    self->ortPos_ = nil;

    // super class 
    [super dealloc];

    return;
}


//
// ͽΰλ
//  ֤ͤ release ƤӽФǹԤ
//
//  Call
//    corr      : 
//    startPos_ : (濴)(instance ѿ)
//    endPos_   : 轪(濴)(instance ѿ)
//    ortPos_   : (instance ѿ)
//
//  Return
//    function : ΰ
//
-(PoCoRect *)calcArea:(int)corr
{
    PoCoRect *r;
    int i;
    PoCoCalcRotationForEllipse *rot;
    int x;
    int y;
    int h;
    int v;
    int d;

    r = [[PoCoRect alloc] init];

    // ʿλ
    i = ([self->startPos_ x] < [self->endPos_ x]) ?
            ([self->endPos_ x] - [self->startPos_ x]) :
            ([self->startPos_ x] - [self->endPos_ x]);
    [r setLeft:([self->startPos_ x] - i)];
    [r setRight:([self->startPos_ x] + i + 1)];

    // ľλ
    i = ([self->startPos_ y] < [self->endPos_ y]) ?
            ([self->endPos_ y] - [self->startPos_ y]) :
            ([self->startPos_ y] - [self->endPos_ y]);
    [r setTop:([self->startPos_ y] - i)];
    [r setBottom:([self->startPos_ y] + i + 1)];

    // ˤžȿ
    rot = [[PoCoCalcRotationForEllipse alloc] initWithCenterPos:self->startPos_
                                                     withEndPos:self->endPos_
                                                    ifAnyOrient:self->ortPos_];
    if ([rot isExec]) {
        h = ([r width] / 2);
        v = ([r height] / 2);
        [r setLeft:INT_MAX];
        [r setTop:INT_MAX];
        [r setRight:INT_MIN];
        [r setBottom:INT_MIN];
        for (d = 0; d < 3600; (d)++) {
            x = ([self->startPos_ x] + [rot calcXAxis:d
                                           horiLength:h
                                           vertLength:v]);
            y = ([self->startPos_ y] + [rot calcYAxis:d
                                           horiLength:h
                                           vertLength:v]);
            if (x < [r left]) {
                [r setLeft:x];
            } else if (x > [r right]) {
                [r setRight:x];
            }
            if (y < [r top]) {
                [r setTop:y];
            } else if (y > [r bottom]) {
                [r setBottom:y];
            }
        }
    }
    [rot release];

    // ̤ȿ
    [r setLeft:([r left] - corr)];
    [r setTop:([r top] - corr)];
    [r setRight:([r right] + corr)];
    [r setBottom:([r bottom] + corr)];

    return r;
}


//
// Խ
//
//  Call
//    dc : ΰ
//    uc : äΰ
//
//  Return
//    function : 
//
-(BOOL)startEdit:(int)dc
    withUndoCorr:(int)uc
{
    BOOL result;
    PoCoRect *dr;
    PoCoRect *ur;

    // ϰϤ
    dr = [self calcArea:dc];
    ur = [self calcArea:uc];
    [super correctRect:dr];
    [super correctRect:ur];

    // super class ز
    result = [super startEdit:dr
                 withUndoRect:ur];

    // ϰϤ
    [dr release];
    [ur release];
 
    return result;
}


//
// Խλ
//
//  Call
//    name : ä̾ʸ
//
//  Return
//    None
//
-(void)endEdit:(NSString *)name
{
    // super class ز
    [super endEdit:name];

    return;
}

@end




// ============================================================================
@implementation PoCoControllerParallelogramEditterBase

// ---------------------------------------------------------- instance - public
//
// initialize
//
//  Call
//    pict   : Խоݲ
//    info   : Խ
//    undo   : ä
//    eraser : äѲ
//    buf    : ݻ
//    pen    : ڥ
//    tile   : ѥ
//    f      : 1(濴)
//    s      : 2(濴)
//    t      : 3(濴)
//    idx    : оݥ쥤䡼ֹ
//
//  Return
//    function   : 
//    firstPos_  : 1(濴)(instance ѿ)
//    secondPos_ : 2(濴)(instance ѿ)
//    thirdPos_  : 3(濴)(instance ѿ)
//    fourthPos_ : 4(濴)(instance ѿ)
//
-(id)initPict:(PoCoPicture *)pict
     withInfo:(PoCoEditInfo *)info
     withUndo:(NSUndoManager *)undo
   withEraser:(PoCoColorPattern *)eraser
   withBuffer:(PoCoColorBuffer *)buf
      withPen:(PoCoPenStyle *)pen
     withTile:(PoCoTilePattern *)tile
    withFirst:(PoCoPoint *)f
   withSecond:(PoCoPoint *)s
    withThird:(PoCoPoint *)t
      atIndex:(int)idx;
{
//    DPRINT((@"[PoCoControllerParallelogramEditterBase init]\n"));

    // super class 
    self = [super initPict:pict
                  withInfo:info
                  withUndo:undo
                withEraser:eraser
                withBuffer:buf
                   withPen:pen
                  withTile:tile
                   atIndex:idx];

    // Ȥ
    if (self != nil) {
        // ʣ
        self->firstPos_ = [[PoCoPoint alloc] initX:[f x]
                                             initY:[f y]];
        self->secondPos_ = [[PoCoPoint alloc] initX:[s x]
                                              initY:[s y]];
        self->thirdPos_ = [[PoCoPoint alloc] initX:[t x]
                                             initY:[t y]];

        // 4ܤ
        self->fourthPos_ = [[PoCoPoint alloc] initX:([f x] + ([t x] - [s x]))
                                              initY:([f y] + ([t y] - [s y]))];
    }

    return self;
}


//
// deallocate
//
//  Call
//    None
//
//  Return
//    firstPos_  : 1(濴)(instance ѿ)
//    secondPos_ : 2(濴)(instance ѿ)
//    thirdPos_  : 3(濴)(instance ѿ)
//    fourthPos_ : 4(濴)(instance ѿ)
//
-(void)dealloc
{
//    DPRINT((@"[PoCoControllerParallelogramEditterBase dealloc]\n"));

    // 񸻤
    [self->firstPos_ release];
    [self->secondPos_ release];
    [self->thirdPos_ release];
    [self->fourthPos_ release];
    self->firstPos_ = nil;
    self->secondPos_ = nil;
    self->thirdPos_ = nil;
    self->fourthPos_ = nil;

    // super class 
    [super dealloc];

    return;
}


//
// ͽΰλ
//  ֤ͤ release ƤӽФǹԤ
//
//  Call
//    corr       : 
//    firstPos_  : 1(濴)(instance ѿ)
//    secondPos_ : 2(濴)(instance ѿ)
//    thirdPos_  : 3(濴)(instance ѿ)
//    fourthPos_ : 4(濴)(instance ѿ)
//
//  Return
//    function : ΰ
//
-(PoCoRect *)calcArea:(int)corr
{
    PoCoRect *r;

    r = [[PoCoRect alloc] initLeft:[self->firstPos_ x]
                           initTop:[self->firstPos_ y]
                         initRight:[self->firstPos_ x]
                        initBottom:[self->firstPos_ y]];

    // 
    if ([self->secondPos_ x] < [r left]) {
        [r setLeft:[self->secondPos_ x]];
    }
    if ([self->thirdPos_ x] < [r left]) {
        [r setLeft:[self->thirdPos_ x]];
    }
    if ([self->fourthPos_ x] < [r left]) {
        [r setLeft:[self->fourthPos_ x]];
    }
    [r setLeft:([r left] - corr)];

    // 
    if ([self->secondPos_ y] < [r top]) {
        [r setTop:[self->secondPos_ y]];
    }
    if ([self->thirdPos_ y] < [r top]) {
        [r setTop:[self->thirdPos_ y]];
    }
    if ([self->fourthPos_ y] < [r top]) {
        [r setTop:[self->fourthPos_ y]];
    }
    [r setTop:([r top] - corr)];

    // 
    if ([r right] < [self->secondPos_ x]) {
        [r setRight:[self->secondPos_ x]];
    }
    if ([r right] < [self->thirdPos_ x]) {
        [r setRight:[self->thirdPos_ x]];
    }
    if ([r right] < [self->fourthPos_ x]) {
        [r setRight:[self->fourthPos_ x]];
    }
    [r setRight:([r right] + corr)];

    // 
    if ([r bottom] < [self->secondPos_ y]) {
        [r setBottom:[self->secondPos_ y]];
    }
    if ([r bottom] < [self->thirdPos_ y]) {
        [r setBottom:[self->thirdPos_ y]];
    }
    if ([r bottom] < [self->fourthPos_ y]) {
        [r setBottom:[self->fourthPos_ y]];
    }
    [r setBottom:([r bottom] + corr)];

    return r;
}


//
// Խ
//
//  Call
//    dc : ΰ
//    uc : äΰ
//
//  Return
//    function : 
//
-(BOOL)startEdit:(int)dc
    withUndoCorr:(int)uc
{
    BOOL result;
    PoCoRect *dr;
    PoCoRect *ur;

    // ϰϤ
    dr = [self calcArea:dc];
    ur = [self calcArea:uc];
    [super correctRect:dr];
    [super correctRect:ur];

    // super class ز
    result = [super startEdit:dr
                 withUndoRect:ur];

    // ϰϤ
    [dr release];
    [ur release];
 
    return result;
}


//
// Խλ
//
//  Call
//    name : ä̾ʸ
//
//  Return
//    None
//
-(void)endEdit:(NSString *)name
{
    // super class ز
    [super endEdit:name];

    return;
}

@end




// ============================================================================
@implementation PoCoControllerCurveEditterBase

// ---------------------------------------------------------- instance - public
//
// initialize
//
//  Call
//    pict   : Խоݲ
//    info   : Խ
//    undo   : ä
//    eraser : äѲ
//    buf    : ݻ
//    pen    : ڥ
//    tile   : ѥ
//    points : 
//    idx    : оݥ쥤䡼ֹ
//    type   : ˡ
//    freq   : 
//
//  Return
//    function : 
//    points_  : (instance ѿ)
//    type_    : ˡ(instance ѿ)
//    freq_    : (instance ѿ)
//
-(id)initPict:(PoCoPicture *)pict
     withInfo:(PoCoEditInfo *)info
     withUndo:(NSUndoManager *)undo
   withEraser:(PoCoColorPattern *)eraser
   withBuffer:(PoCoColorBuffer *)buf
      withPen:(PoCoPenStyle *)pen
     withTile:(PoCoTilePattern *)tile
   withPoints:(NSMutableArray *)points
     withType:(PoCoCurveWithPointsType)type
     withFreq:(unsigned int)freq
      atIndex:(int)idx;
{
//    DPRINT((@"[PoCoControllerCurveEditterBase init]\n"));

    // super class 
    self = [super initPict:pict
                  withInfo:info
                  withUndo:undo
                withEraser:eraser
                withBuffer:buf
                   withPen:pen
                  withTile:tile
                   atIndex:idx];

    // Ȥ
    if (self != nil) {
        self->points_ = points;
        self->type_ = type;
        self->freq_ = freq;
        [self->points_ retain];
    }

    return self;
}


//
// deallocate
//
//  Call
//    None
//
//  Return
//    points_ : (instance ѿ)
//
-(void)dealloc
{
//    DPRINT((@"[PoCoControllerCurveEditterBase dealloc]\n"));

    // 񸻤
    [self->points_ release];
    self->points_ = nil;

    // super class 
    [super dealloc];

    return;
}


//
// ͽΰλ
//  ֤ͤ release ƤӽФǹԤ
//
//  Call
//    corr    : 
//    points_ : (instance ѿ)
//    type_   : ˡ(instance ѿ)
//    freq_   : (instance ѿ)
//
//  Return
//    function : ΰ
//
-(PoCoRect *)calcArea:(int)corr
{
    PoCoRect *r;
    PoCoCalcCurveWithPointsBase *calc;
    NSEnumerator *iter;
    PoCoPoint *p;

    r = [[PoCoRect alloc] initLeft:INT_MAX
                           initTop:INT_MAX
                         initRight:INT_MIN
                        initBottom:INT_MIN];

    switch (self->type_) {
        default:
        case PoCoCurveWithPoints_Lagrange:
            calc = [[PoCoCalcLagrange alloc] initWithPoints:self->points_];
            break;
        case PoCoCurveWithPoints_Spline:
            calc = [[PoCoCalcSpline alloc] initWithPoints:self->points_];
            break;
    }
    iter = [[calc exec:self->freq_] objectEnumerator];
    for (p = [iter nextObject]; p != nil; p = [iter nextObject]) {
        if (([p x] - corr) < [r left]) {
            [r setLeft:([p x] - corr)];
        }
        if (([p y] - corr) < [r top]) {
            [r setTop:([p y] - corr)];
        }
        if (([p x] + corr) > [r right]) {
            [r setRight:([p x] + corr)];
        }
        if (([p y] + corr) > [r bottom]) {
            [r setBottom:([p y] + corr)];
        }
    }
    [calc release];

    return r;
}


//
// Խ
//
//  Call
//    dc : ΰ
//    uc : äΰ
//
//  Return
//    function : 
//
-(BOOL)startEdit:(int)dc
    withUndoCorr:(int)uc
{
    BOOL result;
    PoCoRect *dr;
    PoCoRect *ur;

    // ϰϤ
    dr = [self calcArea:dc];
    ur = [self calcArea:uc];
    [super correctRect:dr];
    [super correctRect:ur];

    // super class ز
    result = [super startEdit:dr
                 withUndoRect:ur];

    // ϰϤ
    [dr release];
    [ur release];
 
    return result;
}


//
// Խλ
//
//  Call
//    name : ä̾ʸ
//
//  Return
//    None
//
-(void)endEdit:(NSString *)name
{
    // super class ز
    [super endEdit:name];

    return;
}

@end
