//
//	Pelistina on Cocoa - PoCo -
//	/
//
//	Copyright (C) 2005-2017 KAENRYUU Koutoku.
//

#import "PoCoControllerFrameEditterBase.h"

#import "PoCoCalcRotation.h"
#import "PoCoCalcEllipse.h"
#import "PoCoCalcCurveWithPoints.h"

#import "PoCoEditNormalDrawPoint.h" 
#import "PoCoEditNormalDrawLine.h"
#import "PoCoEditUniformedDensityDrawLineBase.h"
#import "PoCoEditDensityDrawLineBase.h"
#import "PoCoEditAtomizerDrawLine.h"
#import "PoCoEditGradationDrawLine.h"
#import "PoCoEditRandomDrawLine.h"

// ============================================================================
@implementation PoCoControllerLineEditterBase

// ---------------------------------------------------------- instance - public
//
// initialize
//
//  Call
//    pict   : Խоݲ
//    info   : Խ
//    undo   : ä
//    eraser : äѲ
//    buf    : ݻ
//    pen    : ڥ
//    tile   : ѥ
//    s      : (濴)
//    e      : 轪(濴)
//    prop   : ɮ
//    idx    : оݥ쥤䡼ֹ
//
//  Return
//    function  : 
//    startPos_ : (濴)(instance ѿ)
//    endPos_   : 轪(濴)(instance ѿ)
//    prop_     : ɮ(instance ѿ)
//    cont_     : Ϣ³ʬ(instance ѿ)
//
-(id)initPict:(PoCoPicture *)pict
     withInfo:(PoCoEditInfo *)info
     withUndo:(NSUndoManager *)undo
   withEraser:(PoCoColorPattern *)eraser
   withBuffer:(PoCoColorBuffer *)buf
      withPen:(PoCoPenStyle *)pen
     withTile:(PoCoTilePattern *)tile
    withStart:(PoCoPoint *)s
      withEnd:(PoCoPoint *)e
       isProp:(BOOL)prop
      atIndex:(int)idx
{
//    DPRINT((@"[PoCoControllerLineEditterBase init]\n"));

    // super class 
    self = [super initPict:pict
                  withInfo:info
                  withUndo:undo
                withEraser:eraser
                withBuffer:buf
                   withPen:pen
                  withTile:tile
                   atIndex:idx];

    // Ȥ
    if (self != nil) {
        self->prop_ = prop;
        self->cont_ = NO;

        // ʣ
        self->startPos_ = [[PoCoPoint alloc] initX:[s x] initY:[s y]];
        self->endPos_   = [[PoCoPoint alloc] initX:[e x] initY:[e y]];
    }

    return self;
}


//
// deallocate
//
//  Call
//    None
//
//  Return
//    startPos_ : (濴)(instance ѿ)
//    endPos_   : 轪(濴)(instance ѿ)
//
-(void)dealloc
{
//    DPRINT((@"[PoCoControllerLineEditterBase dealloc]\n"));

    // 񸻤
    [self->startPos_ release];
    [self->endPos_ release];
    self->startPos_ = nil;
    self->endPos_ = nil;

    // super class 
    [super dealloc];

    return;
}


//
// ڥ
//
//  Call
//    editInfo_ : Խ( instance ѿ)
//    prop_     : ɮ(instance ѿ)
//
//  Return
//    function : (1-16)
//
-(int)selectPenSize
{
    int press;

    // ɮ
    if ((self->prop_) &&
        ([self->editInfo_ sizePropType] == PoCoProportionalType_Relation)) {
        press = [self->editInfo_ pressure];
    } else {
        press = 1000;
    }

    // ¤ͻ
    return MIN(
        [self->editInfo_ penSize],
        (([self->editInfo_ penSize] * press) / 1000)
    );
}


//
// ǻ٤
//
//  Call
//    editInfo_ : Խ( instance ѿ)
//    prop_     : ɮ(instance ѿ)
//
//  Return
//    function : ǻ(0.1%ñ)
//
-(int)selectDensity
{
    int press;

    if ((self->prop_) &&
        ([self->editInfo_ densityPropType] == PoCoProportionalType_Relation)) {
        press = [self->editInfo_ pressure];
    } else {
        press = 1000;
    }

    // ¤ͻ
    return MIN(
        [self->editInfo_ density],
        (([self->editInfo_ density] * press) / 1000)
    );
}


//
// ͽΰλ
//
//  Call
//    corr      : 
//    picture_  : Խоݲ( instance ѿ)
//    startPos_ : (濴)(instance ѿ)
//    endPos_   : 轪(濴)(instance ѿ)
//
//  Return
//    function : ΰ
//
-(PoCoRect *)calcArea:(int)corr
{
    PoCoRect *r;

    r = [[PoCoRect alloc] init];

    // ʿλ
    if ([self->startPos_ x] < [self->endPos_ x]) {
        [r setLeft:([self->startPos_ x] - corr)];
        [r setRight:(([self->endPos_ x] + corr) + 1)];
    } else {
        [r setLeft:([self->endPos_ x] - corr)];
        [r setRight:(([self->startPos_ x] + corr) + 1)];
    }

    // ľλ
    if ([self->startPos_ y] < [self->endPos_ y]) {
        [r setTop:([self->startPos_ y] - corr)];
        [r setBottom:(([self->endPos_ y] + corr) + 1)];
    } else {
        [r setTop:([self->endPos_ y] - corr)];
        [r setBottom:(([self->startPos_ y] + corr) + 1)];
    }

    return r;
}


//
// Խ
//
//  Call
//    dc : ΰ
//    uc : äΰ
//
//  Return
//    function : 
//
-(BOOL)startEdit:(int)dc
    withUndoCorr:(int)uc
{
    BOOL result;
    PoCoRect *dr;
    PoCoRect *ur;

    // ϰϤ
    dr = [self calcArea:dc];
    ur = [self calcArea:uc];
    [super correctRect:dr];
    [super correctRect:ur];

    // super class ز
    result = [super startEdit:dr
                 withUndoRect:ur];

    // ϰϤ
    [dr release];
    [ur release];

    return result;
}


//
// Ϣ³ʬ
//
//  Call
//    None
//
//  Return
//    cont_ : Ϣ³ʬ(instance ѿ)
//
-(void)setCont
{
    self->cont_ = YES;

    return;
}


//
// ľ
//
//  Call
//    edit      : Խ
//    editInfo_ : Խ( instance ѿ)
//    startPos_ : (濴)(instance ѿ)
//    endPos_   : 轪(濴)(instance ѿ)
//    cont_     : Ϣ³ʬ(instance ѿ)
//
//  Return
//    None
//
-(void)drawFrame:(id)edit
{
    if ((self->cont_) ||
        (([self->editInfo_ continuationType]) &&
         (![self->startPos_ isEqualPos:self->endPos_]))) {
        // Ϣ³ʬ
        [edit addPoint:self->startPos_];
        [edit addPoint:self->endPos_];
        [edit executeDraw];
    } else {
        // Ϣ³ʬ
        [edit addPoint:self->startPos_];
        [edit addPoint:self->startPos_];
        [edit executeDraw];
        if (!([self->startPos_ isEqualPos:self->endPos_])) {
            [edit clearPoint];
            [edit addPoint:self->endPos_];
            [edit addPoint:self->endPos_];
            [edit executeDraw];
        }
    }

    return;
}


//
// Խλ
//
//  Call
//    name : ä̾ʸ
//
//  Return
//    None
//
-(void)endEdit:(NSString *)name
{
    // super class ز
    [super endEdit:name];

    return;
}

@end




// ============================================================================
@implementation PoCoControllerBoxFrameEditterBase

// ---------------------------------------------------------- instance - public
//
// initialize
//
//  Call
//    pict   : Խоݲ
//    info   : Խ
//    undo   : ä
//    eraser : äѲ
//    buf    : ݻ
//    pen    : ڥ
//    tile   : ѥ
//    s      : (濴)
//    e      : 轪(濴)
//    o      : 
//    idx    : оݥ쥤䡼ֹ
//
//  Return
//    function : 
//    cont_    : Ϣ³ʬ(instance ѿ)
//
-(id)initPict:(PoCoPicture *)pict
     withInfo:(PoCoEditInfo *)info
     withUndo:(NSUndoManager *)undo
   withEraser:(PoCoColorPattern *)eraser
   withBuffer:(PoCoColorBuffer *)buf
      withPen:(PoCoPenStyle *)pen
     withTile:(PoCoTilePattern *)tile
    withStart:(PoCoPoint *)s
      withEnd:(PoCoPoint *)e
  ifAnyOrient:(PoCoPoint *)o
      atIndex:(int)idx
{
//    DPRINT((@"[PoCoControllerBoxFrameEditterBase init]\n"));

    // super class 
    self = [super initPict:pict
                  withInfo:info
                  withUndo:undo
                withEraser:eraser
                withBuffer:buf
                   withPen:pen
                  withTile:tile
                 withStart:s
                   withEnd:e
               ifAnyOrient:o
                   atIndex:idx];

    // Ȥ
    if (self != nil) {
        self->cont_ = NO;
    }

    return self;
}


//
// deallocate
//
//  Call
//    None
//
//  Return
//    None
//
-(void)dealloc
{
//    DPRINT((@"[PoCoControllerBoxFrameEditterBase dealloc]\n"));

    // 񸻤
    ;

    // super class 
    [super dealloc];

    return;
}


//
// Ϣ³ʬ
//
//  Call
//    None
//
//  Return
//    cont_ : Ϣ³ʬ(instance ѿ)
//
-(void)setCont
{
    self->cont_ = YES;

    return;
}


//
// Ȥ
//
//  Call
//    edit      : Խ
//    editInfo_ : Խ( instance ѿ)
//    startPos_ : (濴)( instance ѿ)
//    endPos_   : 轪(濴)( instance ѿ)
//    ortPos_   : ( instance ѿ)
//    diagonal_ : žгѤ( instance ѿ)
//    cont_     : Ϣ³ʬ(instance ѿ)
//
//  Return
//    None
//
-(void)drawFrame:(id)edit
{
    PoCoCalcRotationForBox *rot;

    rot = nil;

    // žɸ򻻽
    rot = [[PoCoCalcRotationForBox alloc] initWithStartPos:self->startPos_
                                                withEndPos:self->endPos_
                                               ifAnyOrient:self->ortPos_];
    [rot calc:self->diagonal_];
    if ((self->cont_) ||
        (([self->editInfo_ continuationType]) &&
         (![self->startPos_ isEqualPos:self->endPos_]))) {
        // Ϣ³ʬ
        [edit addPoint:[rot leftTop]];
        [edit addPoint:[rot leftBottom]];
        [edit addPoint:[rot rightBottom]];
        [edit addPoint:[rot rightTop]];
        [edit addPoint:[rot leftTop]];
        [edit executeDraw];
    } else {
        // Ϣ³ʬ
        [edit addPoint:[rot leftTop]];
        [edit addPoint:[rot leftTop]];
        [edit executeDraw];
        if (!([self->startPos_ isEqualPos:self->endPos_])) {
            [edit clearPoint];
            [edit addPoint:[rot leftBottom]];
            [edit addPoint:[rot leftBottom]];
            [edit executeDraw];

            [edit clearPoint];
            [edit addPoint:[rot rightBottom]];
            [edit addPoint:[rot rightBottom]];
            [edit executeDraw];

            [edit clearPoint];
            [edit addPoint:[rot rightTop]];
            [edit addPoint:[rot rightTop]];
            [edit executeDraw];
        }
    }
    [rot release];

    return;
}

@end




// ============================================================================
@implementation PoCoControllerEllipseFrameEditterBase

// ---------------------------------------------------------- instance - public
//
// initialize
//
//  Call
//    pict   : Խоݲ
//    info   : Խ
//    undo   : ä
//    eraser : äѲ
//    buf    : ݻ
//    pen    : ڥ
//    tile   : ѥ
//    s      : (濴)
//    e      : 轪(濴)
//    o      : 
//    idx    : оݥ쥤䡼ֹ
//
//  Return
//    function : 
//
-(id)initPict:(PoCoPicture *)pict
     withInfo:(PoCoEditInfo *)info
     withUndo:(NSUndoManager *)undo
   withEraser:(PoCoColorPattern *)eraser
   withBuffer:(PoCoColorBuffer *)buf
      withPen:(PoCoPenStyle *)pen
     withTile:(PoCoTilePattern *)tile
    withStart:(PoCoPoint *)s
      withEnd:(PoCoPoint *)e
  ifAnyOrient:(PoCoPoint *)o
      atIndex:(int)idx
{
//    DPRINT((@"[PoCoControllerEllipseFrameEditterBase init]\n"));

    // super class 
    self = [super initPict:pict
                  withInfo:info
                  withUndo:undo
                withEraser:eraser
                withBuffer:buf
                   withPen:pen
                  withTile:tile
                 withStart:s
                   withEnd:e
               ifAnyOrient:o
                   atIndex:idx];

    // Ȥ
    if (self != nil) {
        ;
    }

    return self;
}


//
// deallocate
//
//  Call
//    None
//
//  Return
//    None
//
-(void)dealloc
{
//    DPRINT((@"[PoCoControllerEllipseFrameEditterBase dealloc]\n"));

    // 񸻤
    ;

    // super class 
    [super dealloc];

    return;
}


//
// Ϣ³ʬ
//
//  Call
//    None
//
//  Return
//    None
//
-(void)setCont
{
    // ellipse Ǥ϶̤ʤ
    ;

    return;
}


//
// /ʱȤ
//
//  Call
//    edit      : Խ
//    startPos_ : (濴)( instance ѿ)
//    endPos_   : 轪(濴)( instance ѿ)
//    ortPos_   : ( instance ѿ)
//
//  Return
//    None
//
-(void)drawFrame:(id)edit
{
    PoCoCalcEllipse *calc;              // ʱߴؿ
    PoCoPoint *p;

    // λ
    calc = [[PoCoCalcEllipse alloc] init];
    [calc calc:self->startPos_
        endPos:self->endPos_
   orientation:self->ortPos_];

    // 
    for (p = [calc points]; p != nil; p = [calc points]) {
        [edit addPoint:p];
    }
 
    // 
    [edit executeDraw];
    [calc release];

    return;
}

@end




// ============================================================================
@implementation PoCoControllerParallelogramFrameEditterBase

// ---------------------------------------------------------- instance - public
//
// initialize
//
//  Call
//    pict   : Խоݲ
//    info   : Խ
//    undo   : ä
//    eraser : äѲ
//    buf    : ݻ
//    pen    : ڥ
//    tile   : ѥ
//    f      : 1(濴)
//    s      : 2(濴)
//    t      : 3(濴)
//    idx    : оݥ쥤䡼ֹ
//
//  Return
//    function : 
//    cont_    : Ϣ³ʬ(instance ѿ)
//
-(id)initPict:(PoCoPicture *)pict
     withInfo:(PoCoEditInfo *)info
     withUndo:(NSUndoManager *)undo
   withEraser:(PoCoColorPattern *)eraser
   withBuffer:(PoCoColorBuffer *)buf
      withPen:(PoCoPenStyle *)pen
     withTile:(PoCoTilePattern *)tile
    withFirst:(PoCoPoint *)f
   withSecond:(PoCoPoint *)s
    withThird:(PoCoPoint *)t
      atIndex:(int)idx
{
//    DPRINT((@"[PoCoControllerParallelogramFrameEditterBase init]\n"));

    // super class 
    self = [super initPict:pict
                  withInfo:info
                  withUndo:undo
                withEraser:eraser
                withBuffer:buf
                   withPen:pen
                  withTile:tile
                 withFirst:f
                withSecond:s
                 withThird:t
                   atIndex:idx];

    // Ȥ
    if (self != nil) {
        self->cont_ = NO;
    }

    return self;
}


//
// deallocate
//
//  Call
//    None
//
//  Return
//    None
//
-(void)dealloc
{
//    DPRINT((@"[PoCoControllerParallelogramFrameEditterBase dealloc]\n"));

    // 񸻤
    ;

    // super class 
    [super dealloc];

    return;
}


//
// Ϣ³ʬ
//
//  Call
//    None
//
//  Return
//    cont_ : Ϣ³ʬ(instance ѿ)
//
-(void)setCont
{
    self->cont_ = YES;

    return;
}


//
// ʿԻշȤ
//
//  Call
//    edit       : Խ
//    editInfo_  : Խ( instance ѿ)
//    firstPos_  : 1(濴)( instance ѿ)
//    secondPos_ : 2(濴)( instance ѿ)
//    thirdPos_  : 3(濴)( instance ѿ)
//    fourthPos_ : 4(濴)( instance ѿ)
//    cont_      : Ϣ³ʬ(instance ѿ)
//
//  Return
//    None
//
-(void)drawFrame:(id)edit
{
    if ((self->cont_) ||
        ([self->editInfo_ continuationType])) {
        // Ϣ³ʬ
        [edit addPoint:self->firstPos_];
        [edit addPoint:self->secondPos_];
        [edit addPoint:self->thirdPos_];
        [edit addPoint:self->fourthPos_];
        [edit addPoint:self->firstPos_];
        [edit executeDraw];
    } else {
        // Ϣ³ʬ
        [edit addPoint:self->firstPos_];
        [edit addPoint:self->firstPos_];
        [edit executeDraw];
        [edit clearPoint];
        [edit addPoint:self->secondPos_];
        [edit addPoint:self->secondPos_];
        [edit executeDraw];
        [edit clearPoint];
        [edit addPoint:self->thirdPos_];
        [edit addPoint:self->thirdPos_];
        [edit executeDraw];
        [edit clearPoint];
        [edit addPoint:self->fourthPos_];
        [edit addPoint:self->fourthPos_];
        [edit executeDraw];
    }

    return;
}

@end




// ============================================================================
@implementation PoCoControllerCurveWithPointsEditterBase

// ---------------------------------------------------------- instance - public
//
// initialize
//
//  Call
//    pict   : Խоݲ
//    info   : Խ
//    undo   : ä
//    eraser : äѲ
//    buf    : ݻ
//    pen    : ڥ
//    tile   : ѥ
//    points : 
//    type   : ˡ
//    freq   : 
//    idx    : оݥ쥤䡼ֹ
//
//  Return
//    function : 
//    cont_    : Ϣ³ʬ(instance ѿ)
//
-(id)initPict:(PoCoPicture *)pict
     withInfo:(PoCoEditInfo *)info
     withUndo:(NSUndoManager *)undo
   withEraser:(PoCoColorPattern *)eraser
   withBuffer:(PoCoColorBuffer *)buf
      withPen:(PoCoPenStyle *)pen
     withTile:(PoCoTilePattern *)tile
   withPoints:(NSMutableArray *)points
     withType:(PoCoCurveWithPointsType)type
     withFreq:(unsigned int)freq
      atIndex:(int)idx
{
//    DPRINT((@"[PoCoControllerCurveWithPointsEditterBase init]\n"));

    // super class 
    self = [super initPict:pict
                  withInfo:info
                  withUndo:undo
                withEraser:eraser
                withBuffer:buf
                   withPen:pen
                  withTile:tile
                withPoints:points
                  withType:type
                  withFreq:freq
                   atIndex:idx];

    // Ȥ
    if (self != nil) {
        self->cont_ = NO;
    }

    return self;
}


//
// deallocate
//
//  Call
//    None
//
//  Return
//    None
//
-(void)dealloc
{
//    DPRINT((@"[PoCoControllerCurveWithPointsEditterBase dealloc]\n"));

    // 񸻤
    ;

    // super class 
    [super dealloc];

    return;
}


//
// Ϣ³ʬ
//
//  Call
//    None
//
//  Return
//    cont_ : Ϣ³ʬ(instance ѿ)
//
-(void)setCont
{
    self->cont_ = YES;

    return;
}


//
// ľ() 
//
//  Call
//    edit       : Խ
//    editInfo_  : Խ( instance ѿ)
//    points_    : ( instance ѿ)
//    type_      : ˡ( instance ѿ)
//    freq_      : ( instance ѿ)
//    cont_      : Ϣ³ʬ(instance ѿ)
//
//  Return
//    None
//
-(void)drawFrame:(id)edit
{
    PoCoCalcCurveWithPointsBase *calc;
    NSEnumerator *iter;
    PoCoPoint *p;
    BOOL doing;

    // ֱ黻»
    switch (self->type_) {
        default:
        case PoCoCurveWithPoints_Lagrange:
            calc = [[PoCoCalcLagrange alloc] initWithPoints:self->points_];
            break;
        case PoCoCurveWithPoints_Spline:
            calc = [[PoCoCalcSpline alloc] initWithPoints:self->points_];
            break;
    }
    iter = [[calc exec:self->freq_] objectEnumerator];

    // 
    if ((self->cont_) ||
        ([self->editInfo_ continuationType])) {
        // Ϣ³ʬ
        for (p = [iter nextObject]; p != nil; p = [iter nextObject]) {
           [edit addPoint:p];
        }

        // 
        [edit executeDraw];
    } else {
        // Ϣ³ʬ
        doing = NO;
        for (p = [iter nextObject]; p != nil; p = [iter nextObject]) {
           [edit addPoint:p];
           if (doing) {
               [edit executeDraw];
               [edit clearPoint];
               doing = NO;
           } else {
               doing = YES;
           }
        }
    }
    [calc release];

    return;
}

@end
