//
//	Pelistina on Cocoa - PoCo -
//	
//
//	Copyright (C) 2005-2015 KAENRYUU Koutoku.
//

#import "PoCoControllerPictureColorExchanger.h"

#import "PoCoLayer.h"
#import "PoCoPalette.h"
#import "PoCoControllerFactory.h"
#import "PoCoColorBuffer.h"

// ============================================================================
@implementation PoCoControllerPictureColorExchanger

// ---------------------------------------------------------- instance - public
//
// initialize
//
//  Call
//    pict : Խоݲ
//    info : Խ
//    undo : ä
//    buf  : ݻ
//    act  : Τ̵ͭ
//    s    : 򴹸ֹ
//    d    : 迧ֹ
//
//  Return
//    function     : 
//    active_      : Τ̵ͭ(instance ѿ)
//    srcNum_      : 򴹸ֹ(instance ѿ)
//    dstNum_      : 迧ֹ(instance ѿ)
//    colorBuffer_ : ݻ(instance ѿ)
//
-(id)init:(PoCoPicture *)pict
     info:(PoCoEditInfo *)info
     undo:(NSUndoManager *)undo
   buffer:(PoCoColorBuffer *)buf
   active:(BOOL)act
      src:(unsigned char)s
      dst:(unsigned char)d
{
//    DPRINT((@"[PoCoControllerPictureColorExchanger init]\n"));

    // super class 
    self = [super initPict:pict
                  withInfo:info
                  withUndo:undo];

    // Ȥ
    if (self != nil) {
        self->active_ = act;
        self->srcNum_ = s;
        self->dstNum_ = d;
        self->colorBuffer_ = buf;
        [self->colorBuffer_ retain];
    }

    return self;
}


//
// deallocate
//
//  Call
//    None
//
//  Return
//    colorBuffer_ : ݻ(instance ѿ)
//
-(void)dealloc
{
//    DPRINT((@"[PoCoControllerPictureColorExchanger dealloc]\n"));

    // 񸻤β
    [self->colorBuffer_ release];
    self->colorBuffer_ = nil;

    // super class 
    [super dealloc];

    return;
}


//
// Խ¹
//  layer  bitmap Ƥľܽ񤭴
//  (layer κؤ/bitmap κؤǤϤʤ)
//
//  Call
//    picture_ : Խоݲ( instance ѿ)
//    active_  : Τ̵ͭ(instance ѿ)
//    srcNum_  : 򴹸ֹ(instance ѿ)
//    dstNum_  : 迧ֹ(instance ѿ)
//
//  Return
//    function     : Խ
//    colorBuffer_ : ݻ(instance ѿ)
//
-(BOOL)execute
{
    BOOL  result;
    int x;
    int y;
    int i;
    NSEnumerator *iter;
    PoCoLayerBase *layer;
    unsigned char *pixmap;
    PoCoColor *dcol;
    PoCoPalette *plt = [self->picture_ palette];
    const int xmax = (int)([self->picture_ bitmapRect].size.width);
    const int ymax = (int)([self->picture_ bitmapRect].size.height);
    const int rb = (xmax + (xmax & 1));

    result = NO;
    if (self->srcNum_ == self->dstNum_) {
        goto EXIT;
    }

    // ι
    iter = [[self->picture_ layer] objectEnumerator];
    layer = [iter nextObject];
    for ( ; layer != nil; layer = [iter nextObject]) {
        // Խоݤ
        if ([layer layerType] != PoCoLayerType_Bitmap) {
            continue;
        } else if ([layer drawLock]) {
            continue;
        }

        // ༡õִ
        pixmap = [[layer bitmap] pixelmap];
        for (y = 0; y < ymax; (y)++) {
            i = (y * rb);
            for (x = 0; x < xmax; (x)++, (i)++) {
                if (pixmap[i] == self->srcNum_) {
                    pixmap[i] = self->dstNum_;
                } else if (pixmap[i] == self->dstNum_) {
                    pixmap[i] = self->srcNum_;
                }
            }
        }
    }

    // ѥåȤθ
    dcol = [[plt palette:self->dstNum_] copy];
    [[plt palette:self->dstNum_] setRed:[[plt palette:self->srcNum_] red]
                               setGreen:[[plt palette:self->srcNum_] green]
                                setBlue:[[plt palette:self->srcNum_] blue]];
    [[plt palette:self->dstNum_] setMask:[[plt palette:self->srcNum_] isMask]];
    [[plt palette:self->dstNum_] setDropper:[[plt palette:self->srcNum_] noDropper]];
    [[plt palette:self->dstNum_] setTrans:[[plt palette:self->srcNum_] isTrans]];

    [[plt palette:self->srcNum_] setRed:[dcol red]
                               setGreen:[dcol green]
                                setBlue:[dcol blue]];
    [[plt palette:self->srcNum_] setMask:[dcol isMask]];
    [[plt palette:self->srcNum_] setDropper:[dcol noDropper]];
    [[plt palette:self->srcNum_] setTrans:[dcol isTrans]];
    [dcol release];

    // λ
    result = YES;

    // ä
    [self createUndo];

    // ݻ
    [self->colorBuffer_ reset:nil];

    // Խ
    if (self->active_) {
        [[NSNotificationCenter defaultCenter]
            postNotificationName:PoCoChangePalette
                          object:nil];
    }

EXIT:
    return result;
}


//
// ä
//
//  Call
//    undoManager_ : ä( instance ѿ)
//    factory_     : Խ( instance ѿ)
//    active_      : Τ̵ͭ(instance ѿ)
//    srcNum_      : 򴹸ֹ(instance ѿ)
//    dstNum_      : 迧ֹ(instance ѿ)
//
//  Return
//    None
//
-(void)createUndo
{
    if (self->active_) {
        [[self->undoManager_ prepareWithInvocationTarget:self->factory_]
            createPictureColorExchanger:YES
                                    src:self->srcNum_
                                    dst:self->dstNum_];
    } else {
        [[self->undoManager_ prepareWithInvocationTarget:self->factory_]
            createPictureColorExchangerPassive:YES
                                           src:self->srcNum_
                                           dst:self->dstNum_];
    }

    [super setUndoName:[[NSBundle mainBundle]
                           localizedStringForKey:@"PaletteExchange"
                                           value:@"exchange palette"
                                           table:nil]];

    return;
}

@end
