//
//	Pelistina on Cocoa - PoCo -
//	PNG 饤֥
//
//	Copyright (C) 2005-2015 KAENRYUU Koutoku.
//

#import "PoCoPNG.h"

#import <zlib.h>

// 
static const unsigned int BUFFER_SIZE = 4096;

// ============================================================================
@implementation PoCoPNGEncoder

// ------------------------------------------------------------- class - public
//
// CRC32 ׻
//
//  Call
//    crc   : CRC
//    bytes : CRC ׻Х
//    len   : ХĹ
//
//  Return
//    crc : CRC
//
+(unsigned int)calcCRC:(unsigned int)crc
                 bytes:(const void *)bytes
                length:(unsigned int)len
{
    return (unsigned int)(crc32(crc, bytes, len));
}


// ---------------------------------------------------------- instance - public
//
// initialize
//
//  Call
//    None
//
//  Return
//    function     : 
//    chunkName_[] : CHUNK ̾(instance ѿ)
//    chunkBody_   : CHUNK (instance ѿ)
//    bufferLen_   : chunkBody_ (instance ѿ)
//    bufferSize_  : chunkBody_ (instance ѿ)
//
-(id)init
{
    DPRINT((@"[PoCoPNGEncoder init]\n"));

    // super class ν
    self = [super init];

    // Ȥν
    if (self != nil) {
        self->chunkBody_ = NULL;
        self->bufferLen_ = 0;
        self->bufferSize_ = 0;
        memset(self->chunkName_, 0x00, 4);
    }

    return self;
}


//
// deallocate
//
//  Call
//    None
//
//  Return
//    chunkBody_ : CHUNK (instance ѿ)
//
-(void)dealloc
{
    DPRINT((@"[PoCoPNGEncoder dealloc]\n"));

    if (self->chunkBody_ != NULL) {
        free(self->chunkBody_);
    }
    self->chunkBody_ = NULL;

    // super class β
    [super dealloc];

    return;
}


//
// CHUNK 
//
//  Call
//    name : CHUNK ̾
//
//  Return
//    function     : error code
//    chunkName_[] : CHUNK ̾(instance ѿ)
//    bufferLen_   : chunkBody_ (instance ѿ)
//
-(PoCoErr)createChunk:(const unsigned char *)name
{
    // ̾ݻ
    memcpy(self->chunkName_, name, 4);

    // body λ̤
    self->bufferLen_ = 0;

    return ER_OK;
}


//
// CHUNK 񤭹(1Х)
//
//  Call
//    val : 
//
//  Return
//    function : error code
//
-(PoCoErr)writeByteChunk:(unsigned char)val
{
    return [self writeChunk:(const void *)(&(val))
                     length:sizeof(unsigned char)];
}


//
// CHUNK 񤭹(int)(4Х)
//
//  Call
//    val : 
//
//  Return
//    function : error code
//
-(PoCoErr)writeIntChunk:(unsigned int)val
{
#ifdef  __LITTLE_ENDIAN__
    val = PoCoSwapUW(val);
#endif  // __LITTLE_ENDIAN__
    return [self writeChunk:(const void *)(&(val))
                     length:sizeof(unsigned int)];
}


//
// CHUNK 񤭹(ǤĹ)
//
//  Call
//    bytes       : Х
//    len         : ХĹ
//    chunkBody_  : CHUNK (instance ѿ)
//    bufferLen_  : chunkBody_ (instance ѿ)
//    bufferSize_ : chunkBody_ (instance ѿ)
//
//  Return
//    chunkBody_  : CHUNK (instance ѿ)
//    bufferLen_  : chunkBody_ (instance ѿ)
//    bufferSize_ : chunkBody_ (instance ѿ)
//
-(PoCoErr)writeChunk:(const void *)bytes
              length:(unsigned int)len
{
    PoCoErr er;
    unsigned char *ptr;

    er = ER_OK;

    // Хåեĥ
    while (self->bufferSize_ < (self->bufferLen_ + len)) {
        // Ƴ
        ptr = (unsigned char *)(realloc(self->chunkBody_, self->bufferSize_ + BUFFER_SIZE));
        if (ptr == NULL) {
            DPRINT((@"memory allocation error.\n"));
            er = ER_NOMEM;
            goto EXIT;
        }

        // 
        self->chunkBody_ = ptr;
        self->bufferSize_ += BUFFER_SIZE;
    }

    // ʣ
    memcpy(&(self->chunkBody_[self->bufferLen_]), bytes, len);
    self->bufferLen_ += len;

EXIT:
    return er;
}


//
// CHUNK Ĥ
//
//  Call
//    None
//
//  Return
//    function : error code
//
-(PoCoErr)closeChunk
{
    // ⤷ʤ
    ;

    return ER_OK;
}


//
// NSData 
//
//  Call
//    chunkName_[] : CHUNK ̾(instance ѿ)
//    chunkBody_   : CHUNK (instance ѿ)
//    bufferLen_   : chunkBody_ (instance ѿ)
//
//  Return
//    function : data
//
-(NSData *)createNSData
{
    NSMutableData *data;
    unsigned int crc;
#ifdef  __LITTLE_ENDIAN__
    unsigned int i;
#endif  // __LITTLE_ENDIAN__

    data = [NSMutableData data];

    // crc ׻
    crc = [PoCoPNGEncoder calcCRC:0
                            bytes:NULL
                           length:0];
    crc = [PoCoPNGEncoder calcCRC:crc
                            bytes:self->chunkName_
                           length:4];
    if (self->bufferLen_ > 0) {
        crc = [PoCoPNGEncoder calcCRC:crc
                                bytes:self->chunkBody_
                               length:self->bufferLen_];
    }

    // Ĺ
#ifdef  __LITTLE_ENDIAN__
    i = PoCoSwapUW(self->bufferLen_);
    [data appendBytes:&(i)
               length:sizeof(unsigned int)];
#else   // __LITTLE_ENDIAN__
    [data appendBytes:&(self->bufferLen_)
               length:sizeof(unsigned int)];
#endif  // __LITTLE_ENDIAN__

    // CHUNK ̾
    [data appendBytes:self->chunkName_
               length:4];

    // body
    if (self->bufferLen_ > 0) {
        [data appendBytes:self->chunkBody_
                   length:self->bufferLen_];
    }

    // CRC
#ifdef  __LITTLE_ENDIAN__
    i = PoCoSwapUW(crc);
    [data appendBytes:&(i)
               length:sizeof(unsigned int)];
#else   // __LITTLE_ENDIAN__
    [data appendBytes:&(crc)
               length:sizeof(unsigned int)];
#endif  // __LITTLE_ENDIAN__

    return data;
}

@end




// ============================================================================
@implementation PoCoPNGDecoder

// ---------------------------------------------------------- instance - public
//
// initialize
//
//  Call
//    None
//
//  Return
//    function      : 
//    chunkName_[]  : CHUNK ̾(instance ѿ)
//    chunkBody_    : CHUNK (instance ѿ)
//    bufferLen_    : chunkBody_ (instance ѿ)
//    bufferSize_   : chunkBody_ (instance ѿ)
//    bufferRefer_  : chunkBody_ Ȱ(instance ѿ)
//    filterBuffer_ : ե륿ѤΥХåե(instance ѿ)
//    filterSize_   : ե륿(instance ѿ)
//
-(id)init
{
    DPRINT((@"[PoCoPNGDecoder init]\n"));

    // super class ν
    self = [super init];

    // Ȥν
    if (self != nil) {
        self->chunkBody_ = NULL;
        self->bufferLen_ = 0;
        self->bufferSize_ = 0;
        self->bufferRefer_ = 0;
        memset(self->chunkName_, 0x00, 4);
        self->filterBuffer_ = NULL;
        self->filterSize_ = 0;
    }

    return self;
}


//
// deallocate
//
//  Call
//    None
//
//  Return
//    chunkBody_    : CHUNK (instance ѿ)
//    filterBuffer_ : ե륿ѤΥХåե(instance ѿ)
//
-(void)dealloc
{
    DPRINT((@"[PoCoPNGDecoder dealloc]\n"));

    // Хåե
    if (self->chunkBody_ != NULL) {
        free(self->chunkBody_);
    }
    if (self->filterBuffer_ != NULL) {
        free(self->filterBuffer_);
    }
    self->chunkBody_ = NULL;
    self->filterBuffer_ = NULL;

    // super class β
    [super dealloc];

    return;
}


//
// CHUNK 򳫤
//  CHUNK ̾ξȹȤ
//
//  Call
//    name         : ȹ礹 CHUNK ̾
//    chunkName_[] : CHUNK ̾(instance ѿ)
//
//  Return
//    function     : YES : ̾
//                   NO  : ̾԰
//    bufferRefer_ : chunkBody_ Ȱ(instance ѿ)
//
-(BOOL)openChunk:(const unsigned char *)name
{
    self->bufferRefer_ = 0;

    return (memcmp(name, self->chunkName_, 4) == 0);
}


//
// ɤ߹(1Х)
//
//  Call
//    val : ɤ߹֤ͤ
//
//  Return
//    function : YES : 
//               NO  : 
//    val      : ɤ߹
//
-(BOOL)readByteChunk:(unsigned char *)val
{
    return [self readChunk:(void *)(val)
                    length:sizeof(unsigned char)
                    readed:NULL];
}


//
// ɤ߹(int)(4Х)
//
//  Call
//    val : ɤ߹֤ͤ
//
//  Return
//    function : YES : 
//               NO  : 
//    val      : ɤ߹
//
-(BOOL)readIntChunk:(unsigned int *)val
#ifdef  __LITTLE_ENDIAN__
{
    BOOL result = [self readChunk:(void *)(val)
                           length:sizeof(unsigned int)
                           readed:NULL];

    *val = PoCoSwapUW(*val);

    return result;
}
#else   // __LITTLE_ENDIAN__
{
    return [self readChunk:(void *)(val)
                    length:sizeof(unsigned int)
                    readed:NULL];
}
#endif  // __LITTLE_ENDIAN__


//
// ɤ߹(ǤĹ)
//
//  Call
//    bytes        : ɤ߹Ƥ֤
//    len          : ɤ߹Ĺ
//    read         : ɤ߹
//    chunkBody_   : CHUNK       (instance ѿ)
//    bufferLen_   : chunkBody_   (instance ѿ)
//    bufferRefer_ : chunkBody_ Ȱ(instance ѿ)
//
//  Return
//    function     : YES : 
//                   NO  : 
//    read         : ɤ߹(NULL )
//    bufferRefer_ : chunkBody_ Ȱ(instance ѿ)
//
-(BOOL)readChunk:(void *)bytes
          length:(unsigned int)len
          readed:(unsigned int *)read
{
    BOOL result;

    result = NO;

    // ɤ߹̤򻻽
    if ((self->bufferLen_ - self->bufferRefer_) < len) {
        len = (self->bufferLen_ - self->bufferRefer_);
    }

    // 1ХȰʾʤ
    if (len > 0) {
        // Хåեʣ
        if (bytes != NULL) {
            memcpy(bytes, &(self->chunkBody_[self->bufferRefer_]), len);
        }

        // 
        self->bufferRefer_ += len;

        // 
        result = YES;
        if (read != NULL) {
            *(read) = len;
        }
    }

    return result;
}


//
// CHUNK ̾
//
//  Call
//    chunkName_[] : CHUNK ̾(instance ѿ)
//
//  Return
//    function : CHUNK ̾
//
-(const unsigned char *)chunkName
{
    return self->chunkName_;
}


//
// CHUNK Ƥ
//
//  Call
//    chunkBody_   : CHUNK (instance ѿ)
//    bufferRefer_ : chunkBody_ Ȱ(instance ѿ)
//
//  Return
//    function : CHUNK 
//
-(const void *)bytes
{
    return (const void *)(&(self->chunkBody_[self->bufferRefer_]));
}


//
// CHUNK Ĺ
//
//  Call
//    bufferLen_   : chunkBody_ (instance ѿ)
//    bufferRefer_ : chunkBody_ Ȱ(instance ѿ)
//
//  Return
//    function : CHUNK Ĺ
//
-(unsigned int)length
{
    return (self->bufferLen_ - self->bufferRefer_);
}


//
// CHUNK Ĥ
//
//  Call
//    None
//
//  Return
//    None
//
-(void)closeChunk
{
    // ⤷ʤ
    ;

    return;
}


// NSData ɤ߹
//
//  Call
//    data  : ɤ߹
//    range : ɤ߹ϰ
//
//  Return
//    function     : YES : 
//                   NO  : 
//    range        : ɤ߹ϰ
//    chunkName_[] : CHUNK ̾(instance ѿ)
//    chunkBody_   : CHUNK (instance ѿ)
//    bufferLen_   : chunkBody_ (instance ѿ)
//    bufferSize_  : chunkBody_ (instance ѿ)
//    bufferRefer_ : chunkBody_ Ȱ(instance ѿ)
//
-(BOOL)loadNSData:(NSData *)data
        withRange:(NSRange *)range
{
    BOOL result;
    unsigned char *ptr;
    unsigned int crc;

    result = NO;
    self->bufferLen_ = 0;
    self->bufferRefer_ = 0;

    // CHUNK Ĺ
    range->length = 4;
    if ([data length] < (range->location + range->length)) {
        goto EXIT;
    }
    [data getBytes:&(self->bufferLen_) range:*(range)];
#ifdef  __LITTLE_ENDIAN__
    self->bufferLen_ = PoCoSwapUW(self->bufferLen_);
#endif  // __LITTLE_ENDIAN__
    range->location += range->length;

    // CHUNK ̾
    if ([data length] < (range->location + range->length)) {
        goto EXIT;
    }
    [data getBytes:self->chunkName_ range:*(range)];
    range->location += range->length;

    // body ΰ
    if (self->bufferSize_ < self->bufferLen_) {
        // Ƴ
        ptr = (unsigned char *)(realloc(self->chunkBody_, self->bufferLen_));
        if (ptr == NULL) {
            DPRINT((@"memory allocation error.\n"));
            goto EXIT;
        }

        // 
        self->chunkBody_ = ptr;
        self->bufferSize_ = self->bufferLen_;
    }

    // CHUNK ɤ߹
    range->length = self->bufferLen_;
    if ([data length] < (range->location + range->length)) {
        goto EXIT;
    }
    [data getBytes:self->chunkBody_ range:*(range)];
    range->location += range->length;

    // CRC ɤ߹(ɤ߼Τ)
    range->length = 4;
    if ([data length] < (range->location + range->length)) {
        goto EXIT;
    }
    [data getBytes:&(crc) range:*(range)];
    range->location += range->length;

    // 
    result = YES;

EXIT:
    return result;
}


//
// ե륿
//
//  Call
//    w             : (ե륿ʬϴޤޤʤ)
//    filterBuffer_ : ե륿ѤΥХåե(instance ѿ)
//
//  Return
//    filterBuffer_ : ե륿ѤΥХåե(instance ѿ)
//    filterSize_   : ե륿(instance ѿ)
//
-(void)initFilter:(unsigned int)w
{
    // ΥХåե
    if (self->filterBuffer_ != NULL) {
        free(self->filterBuffer_);
    }
    self->filterBuffer_ = NULL;

    // (+1 ϺǺüκ)
    self->filterSize_ = w;
    self->filterBuffer_ = (unsigned char *)(calloc(w + 1, sizeof(unsigned char)));

    return;
}


//
// ե륿򤫤
//
//  Call
//    type          : ե륿
//    data          : ե륿򤫤ǡ
//    filterBuffer_ : ե륿ѤΥХåե(instance ѿ)
//    filterSize_   : ե륿(instance ѿ)
//
//  Return
//    data          : ե륿
//    filterBuffer_ : ե륿ѤΥХåե(instance ѿ)
//
-(void)filtering:(unsigned int)type
            data:(unsigned char *)data
{
    unsigned int p;                     // paeth filter 
    unsigned int pa;                    // paeth filter 
    unsigned int pb;                    // paeth filter 
    unsigned int pc;                    // paeth filter 
    unsigned int x = self->filterSize_;
    unsigned char left = 0x00;
    unsigned char *flt = &(self->filterBuffer_[1]);
    unsigned char *ptr = data;

    // ե륿򤫤
    switch (type) {
        default:
        case 0:
            // none filter
            ;
            break;
        case 1:
            // sub filter
            do {
                // ե륿򤫤
                *(ptr) += left;

                // 
                left = *(ptr);
                (x)--;
                (ptr)++;
            } while (x != 0);
            break;
        case 2:
            // up filter
            do {
                // ե륿򤫤
                *(ptr) += *(flt);

                // 
                (x)--;
                (ptr)++;
                (flt)++;
            } while (x != 0);
            break;
        case 3:
            // average filter
            do {
                // ե륿򤫤
                *(ptr) += (unsigned char)(((unsigned int)(left) + (unsigned int)(*(flt))) >> 1);

                // 
                left = *(ptr);
                (x)--;
                (ptr)++;
                (flt)++;
            } while (x != 0);
            break;
        case 4:
            // paeth filter
            do {
                // ե륿򤫤
                p = left + *(flt) - *(flt - 1);
                pa = (unsigned int)(abs((int)(p - left)));
                pb = (unsigned int)(abs((int)(p - *(flt))));
                pc = (unsigned int)(abs((int)(p - *(flt - 1))));
                if ((pa <= pb) && (pa <= pc)) {
                    *(ptr) += left;
                } else if (pb <= pc) {
                    *(ptr) += *(flt);
                } else {
                    *(ptr) += *(flt - 1);
                }

                // 
                left = *(ptr);
                (x)--;
                (ptr)++;
                (flt)++;
            } while (x != 0);
            break;
    }

    // ե륿򤫤̤Хåե᤹
    memcpy(&(self->filterBuffer_[1]), data, self->filterSize_);

    return;
}

@end
