//
//	Pelistina on Cocoa - PoCo -
//	ϰϴ
//
//	Copyright (C) 2005-2017 KAENRYUU Koutoku.
//

#import "PoCoSelectionShape.h"

#import "PoCoMyDocument.h"
#import "PoCoAppController.h"
#import "PoCoPicture.h"
#import "PoCoSelLayer.h"
#import "PoCoEditInfo.h"

#import "PoCoCreateStyleMask.h"
#import "PoCoCalcRotation.h"
#import "PoCoEditResizeBitmap.h"
#import "PoCoEditRotateBitmap.h"
#import "PoCoEditRegionFill.h"
#import "PoCoEditUniformedDensityFill.h"
#import "PoCoEditAtomizerFill.h"
#import "PoCoEditGradationFill.h"
#import "PoCoEditRandomFill.h"
#import "PoCoEditWaterDropFill.h"
#import "PoCoEditFlipHoriBitmap.h"
#import "PoCoEditFlipVertBitmap.h"
#import "PoCoEditAutoGradation.h"
#import "PoCoEditAutoGradationAdjacent.h"
#import "PoCoEditBitmapColorReplace.h"

// ============================================================================
@implementation PoCoSelectionShape

// ------------------------------------------------------------ class - private
//
// Ƚ(ϥɥΤ)
//
//  Call
//    p      : ڤ
//    gap    : ϥɥγĥ
//    hndl[] : ϥɥ(ѷ)(instance ѿ)
//
//  Return
//    function : 
//
+(PoCoHandleType)hitTest:(PoCoPoint *)p
               handleGap:(int)gap
             withHanldes:(PoCoRect **)hndl
{
    PoCoHandleType type;
    int l;

    type = PoCoHandleType_unknown;

    // ϥɥ夫Ƚ
    for (l = 1; l < (int)(PoCoHandleType_effective); (l)++) {
        if (([p x] >= ([hndl[l]   left] - gap)) &&
            ([p y] >= ([hndl[l]    top] - gap)) &&
            ([p x] <  ([hndl[l]  right] + gap)) &&
            ([p y] <  ([hndl[l] bottom] + gap))) {
            // Ĥ
            type = (PoCoHandleType)(l);
            break;
        }
    }

    return type;
}


//
// Ƚ(ʤĹΤ)
//
//  Call
//    p     : ڤ
//    r     : Ĺ(ѷ)(instance ѿ)
//    shape : (ѷ)(instance ѿ)
//            != nil : ȳĹȽꤹ
//            == nil : ĹΤߤȽꤹ
//
//  Return
//    function : 
//
+(PoCoHandleType)hitTest:(PoCoPoint *)p
                withRect:(PoCoRect *)r
               withShape:(PoCoBitmap *)shape
{
    PoCoHandleType type;

    // ϰ⤫Ƚ
    if ([r isPointInRect:p]) {
        if (shape != nil) {
            // Ƚ
            if ([shape pixelmap][((([p y] - [r top]) * ([r width] + ([r width] & 0x01))) + ([p x] - [r left]))] != 0x00) {
                type = PoCoHandleType_in_shape;
            } else {
                type = PoCoHandleType_out_shape;
            }
        } else {
            type = PoCoHandleType_in_rect;
        }
    } else {
        type = ((shape != nil) ? PoCoHandleType_out_shape :
                                 PoCoHandleType_out_rect);
    }

    return type;
}


//
// Ƚ
//
//  Call
//    p      : ڤ
//    gap    : ϥɥγĥ
//    r      : Ĺ(ѷ)(instance ѿ)
//    hndl[] : ϥɥ(ѷ)(instance ѿ)
//    shape  : (ѷ)(instance ѿ)
//             != nil : ȳĹȽꤹ
//             == nil : ĹΤߤȽꤹ
//
//  Return
//    function : 
//
+(PoCoHandleType)hitTest:(PoCoPoint *)p
               handleGap:(int)gap
                withRect:(PoCoRect *)r
             withHanldes:(PoCoRect **)hndl
               withShape:(PoCoBitmap *)shape
{
    PoCoHandleType type;

    // ϥɥ夫Ƚ
    type = [PoCoSelectionShape hitTest:p
                             handleGap:gap
                           withHanldes:hndl];
    if (type == PoCoHandleType_unknown) {
        // ϰ⤫Ƚ
        type = [PoCoSelectionShape hitTest:p
                                  withRect:r
                                 withShape:shape];
    }

    return type;
}


//
// Ʊ
//
//  Call
//    r      : Ĺ
//    points : 
//
//  Return
//    function : Ƚ
//
+(BOOL)isSameStyle:(PoCoRect *)r
         testShape:(NSMutableArray *)points
{
    BOOL result;
    PoCoCreateStyleMask *mask;
    PoCoBitmap *rect;
    PoCoBitmap *region;

    result = NO;
    rect = nil;
    region = nil;

    // ޥ
    mask = [[PoCoCreateStyleMask alloc] init];
    if ([mask box:[r lefttop]
              end:[r rightbot]
      orientation:nil
       isDiagonal:NO]) {
        rect = [[mask mask] copy];
    }
    if ([mask polygon:points]) {
        region = [[mask mask] copy];
    }
    [mask release];

    // Ƚ
    if ((rect != nil) && (region != nil)) {
        if (([rect width] == [region width]) &&
            ([rect height] == [region height])) {
            result = (memcmp([rect pixelmap], [region pixelmap], ([rect width] *
 [rect height])) == 0);
        }
    }
    [rect release];
    [region release];

    return result;
}


//
// ɲ
//
//  Call
//    p : Ͽ
//
//  Return
//    points : Ͽ
//    r      : Ĺ
//
+(void)addControlPoint:(PoCoPoint *)p
           targetArray:(NSMutableArray *)points
            targetRect:(PoCoRect *)r
{
    PoCoPoint *tmp;

    // ɲ
    tmp = [[PoCoPoint alloc] initX:[p x]
                             initY:[p y]];
    [points addObject:tmp];
    [tmp release];

    // Ĺĥ
    [r   setLeft:MIN([r   left], [p x])];
    [r    setTop:MIN([r    top], [p y])];
    [r  setRight:MAX([r  right], [p x])];
    [r setBottom:MAX([r bottom], [p y])];

    return;
}


//
// ϥɥ()
//
//  Call
//    r : Ĺ
//
//  Return
//    hndl[] : ϥɥ
//
+(void)createHandleRect:(PoCoRect *)r
          targetHandles:(PoCoRect **)hndl
{
    const int w = [r width];
    const int w2 = (w >> 1);
    const int h = [r height];
    const int h2 = (h >> 1);

    // 濴(濴ȤΤ lefttop == rightbot)
    hndl[PoCoHandleType_center] = [[PoCoRect alloc]
          initLeft:([r left] + w2)
           initTop:([r  top] + h2)
         initRight:([r left] + w2)
        initBottom:([r  top] + h2)];

    // 
    hndl[PoCoHandleType_corner_lt] = [[PoCoRect alloc]
          initLeft:([r left]      - HANDLE_GAP)
           initTop:([r  top]      - HANDLE_GAP)
         initRight:([r left]      + HANDLE_GAP + 1)
        initBottom:([r  top]      + HANDLE_GAP + 1)];

    // 
    hndl[PoCoHandleType_edge_t] = [[PoCoRect alloc]
          initLeft:([r left] + w2 - HANDLE_GAP)
           initTop:([r  top]      - HANDLE_GAP)
         initRight:([r left] + w2 + HANDLE_GAP + 1)
        initBottom:([r  top]      + HANDLE_GAP + 1)];

    // 
    hndl[PoCoHandleType_corner_rt] = [[PoCoRect alloc]
          initLeft:([r left] + w  - HANDLE_GAP)
           initTop:([r  top]      - HANDLE_GAP)
         initRight:([r left] + w  + HANDLE_GAP + 1)
        initBottom:([r  top]      + HANDLE_GAP + 1)];

    // 
    hndl[PoCoHandleType_edge_r] = [[PoCoRect alloc]
          initLeft:([r left] + w  - HANDLE_GAP)
           initTop:([r  top] + h2 - HANDLE_GAP)
         initRight:([r left] + w  + HANDLE_GAP + 1)
        initBottom:([r  top] + h2 + HANDLE_GAP + 1)];

    // 
    hndl[PoCoHandleType_corner_rb] = [[PoCoRect alloc]
          initLeft:([r left] + w  - HANDLE_GAP)
           initTop:([r  top] + h  - HANDLE_GAP)
         initRight:([r left] + w  + HANDLE_GAP + 1)
        initBottom:([r  top] + h  + HANDLE_GAP + 1)];

    // 
    hndl[PoCoHandleType_edge_b] = [[PoCoRect alloc]
          initLeft:([r left] + w2 - HANDLE_GAP)
           initTop:([r  top] + h  - HANDLE_GAP)
         initRight:([r left] + w2 + HANDLE_GAP + 1)
        initBottom:([r  top] + h  + HANDLE_GAP + 1)];

    // 
    hndl[PoCoHandleType_corner_lb] = [[PoCoRect alloc]
          initLeft:([r left]      - HANDLE_GAP)
           initTop:([r  top] + h  - HANDLE_GAP)
         initRight:([r left]      + HANDLE_GAP + 1)
        initBottom:([r  top] + h  + HANDLE_GAP + 1)];

    // 
    hndl[PoCoHandleType_edge_l] = [[PoCoRect alloc]
          initLeft:([r left]      - HANDLE_GAP)
           initTop:([r  top] + h2 - HANDLE_GAP)
         initRight:([r left]      + HANDLE_GAP + 1)
        initBottom:([r  top] + h2 + HANDLE_GAP + 1)];

    return;
}


//
// 
//
//  Call
//    points : ɸ
//    r      : Ĺ
//
//  Return
//    hndl[] : ϥɥ
//    shape  : 
//
+(void)createShape:(NSMutableArray *)points
          withRect:(PoCoRect *)r
     targetHandles:(PoCoRect **)hndl
       targetShape:(PoCoBitmap **)shape
{
    PoCoCreateStyleMask *mask;

    // ޥ
    mask = [[PoCoCreateStyleMask alloc] init];
    if ([mask polygon:points]) {
        [(*(shape)) release];
        *(shape) = [[mask mask] copy];
    }
    [mask release];

    // ϥɥ֤
    [PoCoSelectionShape createHandleRect:r
                           targetHandles:hndl];

    return;
}


// --------------------------------------------------------- instance - private
//
// ѷη˺
//
//  Call
//    None
//
//  Return
//    resultPoints_   : ɸ(ѷ)(instance ѿ)
//    resultRect_     : Ĺ(ѷ)(instance ѿ)
//    resultHandle_[] : ϥɥ(ѷ)(instance ѿ)
//    resultShape_    : (ѷ)(instance ѿ)
//
-(void)clearResult
{
    int l;

    [self->resultPoints_ removeAllObjects];
    [self->resultRect_   setLeft:INT_MAX];
    [self->resultRect_    setTop:INT_MAX];
    [self->resultRect_  setRight:INT_MIN];
    [self->resultRect_ setBottom:INT_MIN];
    [self->resultShape_ release];
    for (l = 0; l < (int)(PoCoHandleType_effective); (l)++) {
        [self->resultHandle_[l] release];
        self->resultHandle_[l] = nil;
    }
    self->resultShape_ = nil;

    return;
}


//
// ѷηѷηʣ
//
//  Call
//    originalPoints_   : ɸ(ѷ)(instance ѿ)
//    originalRect_     : Ĺ(ѷ)(instance ѿ)
//    originalHandle_[] : ϥɥ(ѷ)(instance ѿ)
//    originalShape_    : (ѷ)(instance ѿ)
//
//  Return
//    resultPoints_   : ɸ(ѷ)(instance ѿ)
//    resultRect_     : Ĺ(ѷ)(instance ѿ)
//    resultHandle_[] : ϥɥ(ѷ)(instance ѿ)
//    resultShape_    : (ѷ)(instance ѿ)
//
-(void)copyOriginalToResult
{
    NSEnumerator* iter;
    PoCoPoint *p;
    PoCoPoint *tmp;
    int l;

    // Ƥ˺
    [self clearResult];

    // ɸ
    [self->resultPoints_ release];
    self->resultPoints_ = [[NSMutableArray alloc] init];
    iter = [self->originalPoints_ objectEnumerator];
    for (p = [iter nextObject]; p != nil; p = [iter nextObject]) {
        tmp = [[PoCoPoint alloc] initX:[p x]
                                 initY:[p y]];
        [self->resultPoints_ addObject:tmp];
        [tmp release];
    }

    // Ĺ
    [self->resultRect_ release];
    self->resultRect_ = [[PoCoRect alloc]
                              initLeft:[self->originalRect_ left]
                               initTop:[self->originalRect_ top]
                             initRight:[self->originalRect_ right]
                            initBottom:[self->originalRect_ bottom]];

    // ޥ
    self->resultShape_ = [self->originalShape_ copy];

    // ϥɥ
    for (l = 0; l < (int)(PoCoHandleType_effective); (l)++) {
        self->resultHandle_[l] = [[PoCoRect alloc]
              initLeft:[self->originalHandle_[l] left]
               initTop:[self->originalHandle_[l] top]
             initRight:[self->originalHandle_[l] right]
            initBottom:[self->originalHandle_[l] bottom]];
    }

    return;
}


//
// ѷκɸ
//
//  Call
//    p                 : ΰư
//    originalHandle_[] : ϥɥ(ѷ)(instance ѿ)
//    point_            : ()(instance ѿ)
//    modify_           : ѷν(instance ѿ)
//
//  Return
//    p       : ΰư
//    modify_ : ѷν(instance ѿ)
//
-(void)similerPoint:(PoCoPoint *)p
{
    float ccw;                          // 濴
    float cch;                          // 濴ι⤵
    float tcw;                          // ư
    float tch;                          // ưι⤵

    ccw = (float)([self->point_ x] - [[self->originalHandle_[PoCoHandleType_center] lefttop] x]);
    cch = (float)([self->point_ y] - [[self->originalHandle_[PoCoHandleType_center] lefttop] y]);
    tcw = (float)([p x] - [self->point_ x]);
    tch = (float)([p y] - [self->point_ y]);

    if ((tcw == 0.0) && (tch == 0.0)) {
        /* ưƤʤ */
        ;
    } else if (self->modify_ == PoCoModifierType_resize_similer_hori) {
        /* ʿ̰¸ݻ */
        [p setY:([self->point_ y] + (int)(floor((cch * (tcw / ccw)) + 0.5)))];
    } else if (self->modify_ == PoCoModifierType_resize_similer_vert) {
        /* ľ̰¸ݻ */
        [p setX:([self->point_ x] + (int)(floor((ccw * (tch / cch)) + 0.5)))];
    } else if (fabs(tcw) < fabs(tch)) {
        /* ʿ̰¸򳫻 */
        [p setY:([self->point_ y] + (int)(floor((cch * (tcw / ccw)) + 0.5)))];
        self->modify_ = PoCoModifierType_resize_similer_hori;
    } else {
        /* ľ̰¸򳫻 */
        [p setX:([self->point_ x] + (int)(floor((ccw * (tch / cch)) + 0.5)))];
        self->modify_ = PoCoModifierType_resize_similer_vert;
    }

    return;
}


//
// ѷηžγĹ򻻽
//  ˴ޤह٤Ƥ
//
//  Call
//    originalRect_  : Ĺ(ѷ)(instance ѿ)
//    originalShape_ : (ѷ)(instance ѿ)
//    rotate_        : žؿ(instance ѿ)
//
//  Return
//    resultRect_ : Ĺ(ѷ)(instance ѿ)
//
-(void)calcRotateRectFromOriginalShape
{
    PoCoPoint *p;
    const int skip = ([self->originalShape_ width] & 0x01);
    int x;
    int y;
    unsigned char *bmp;

    // 
    bmp = [self->originalShape_ pixelmap];
    for (y = [self->originalRect_ top];
         y < [self->originalRect_ bottom];
         (y)++, bmp += skip) {
        for (x = [self->originalRect_ left];
             x < [self->originalRect_ right];
             (x)++, (bmp)++) {
            // оݤˤ
            if (*(bmp) != 0x00) {
                // 
                p = [[PoCoPoint alloc] initX:x
                                       initY:y];

                // žΰ֤򻻽
                [self->rotate_ calcPoint:p];

                // Ĺȿ
                [self->resultRect_   setLeft:MIN([self->resultRect_   left], [p x])];
                [self->resultRect_    setTop:MIN([self->resultRect_    top], [p y])];
                [self->resultRect_  setRight:MAX([self->resultRect_  right], [p x])];
                [self->resultRect_ setBottom:MAX([self->resultRect_ bottom], [p y])];

                // 
                [p release];
            }
        }
    }

#if 0
    // ʤǤ餺 half-open property ǤϤʤΤ -1
    [self->resultRect_  setRight:([self->resultRect_  right] - 1)];
    [self->resultRect_ setBottom:([self->resultRect_ bottom] - 1)];
#endif  // 0

    return;
}


// ---------------------------------------------- instance - private - 
//
// ư
//
//  Call
//    org : ѷ
//
//  Return
//    bmp : ѷ
//
-(void)createMoveBitmap:(PoCoBitmap *)org
           targetBitmap:(PoCoBitmap **)bmp
{
    [(*(bmp)) release];
    *(bmp) = nil;
    *(bmp) = [org copy];

    return;
}


//
// ư(ξ)
//
//  Call
//    originalShape_ : (ѷ)(instance ѿ)
//    originalImage_ : (ѷ)(instance ѿ)
//
//  Return
//    resultShape_ : (ѷ)(instance ѿ)
//    resultImage_ : (ѷ)(instance ѿ)
//
-(void)createMoveBitmapTwin
{
    // η̤˺
    [self->resultShape_ release];
    [self->resultImage_ release];
    self->resultShape_ = nil;
    self->resultImage_ = nil;

    // ʣ
    self->resultShape_ = [self->originalShape_ copy];
    self->resultImage_ = [self->originalImage_ copy];

    return;
}


//
// 
//
//  Call
//    org             : ѷ
//    originalPoints_ : ɸ(ѷ)(instance ѿ)
//    resultRect_     : Ĺ(ѷ)(instance ѿ)
//
//  Return
//    bmp : ѷ
//
-(void)createResizeBitmap:(PoCoBitmap *)org
             targetBitmap:(PoCoBitmap **)bmp
{
    PoCoEditResizeBitmap *edit;
    int w;
    int h;

    [(*(bmp)) release];
    *(bmp) = nil;
    if (!([self->resultRect_ empty])) {
        w = ([self->resultRect_ width] + 1);
        h = ([self->resultRect_ height] + 1);
        *(bmp) = [[PoCoBitmap alloc] initWidth:w
                                    initHeight:h
                                  defaultColor:0];
        edit = [[PoCoEditResizeBitmap alloc] initDst:*(bmp)
                                             withSrc:org];
        [edit executeResize];
        [edit release];
    }

    return;
}


//
// (ξ)
//
//  Call
//    originalPoints_ : ɸ(ѷ)(instance ѿ)
//    originalShape_  : (ѷ)(instance ѿ)
//    originalImage_  : (ѷ)(instance ѿ)
//    resultRect_     : Ĺ(ѷ)(instance ѿ)
//
//  Return
//    resultShape_ : (ѷ)(instance ѿ)
//    resultImage_ : (ѷ)(instance ѿ)
//
-(void)createResizeBitmapTwin
{
    PoCoEditResizeBitmapDouble *edit;
    int w;
    int h;

    // η̤˺
    [self->resultShape_ release];
    [self->resultImage_ release];
    self->resultShape_ = nil;
    self->resultImage_ = nil;
    if (!([self->resultRect_ empty])) {
        // ̤γǼ
        w = ([self->resultRect_ width] + 1);
        h = ([self->resultRect_ height] + 1);
        self->resultShape_ = [[PoCoBitmap alloc] initWidth:w
                                                initHeight:h
                                              defaultColor:0];
        self->resultImage_ = [[PoCoBitmap alloc] initWidth:w
                                                initHeight:h
                                              defaultColor:0];

        // 
        edit = [[PoCoEditResizeBitmapDouble alloc] initDst1:self->resultShape_
                                                   withDst2:self->resultImage_
                                                   withSrc1:self->originalShape_
                                                   withSrc2:self->originalImage_];
        [edit executeResize];
        [edit release];
    }

    return;
}


//
// ž
//
//  Call
//    org             : ѷ
//    originalPoints_ : ɸ(ѷ)(instance ѿ)
//    originalRect_   : Ĺ(ѷ)(instance ѿ)
//    resultRect_     : Ĺ(ѷ)(instance ѿ)
//    rotate_         : žؿ(instance ѿ)
//
//  Return
//    bmp : ѷ
//
-(void)createRotateBitmap:(PoCoBitmap *)org
             targetBitmap:(PoCoBitmap **)bmp
{
    PoCoEditRotateBitmap *edit;
    int w;
    int h;

    [*(bmp) release];
    *(bmp) = nil;
    if (!([self->resultRect_ empty])) {
        w = ([self->resultRect_ width] + 1);
        h = ([self->resultRect_ height] + 1);
        *(bmp) = [[PoCoBitmap alloc] initWidth:w
                                    initHeight:h
                                  defaultColor:0];
        edit = [[PoCoEditRotateBitmap alloc] initDst:*(bmp)
                                             withSrc:org
                                             dstRect:self->resultRect_
                                             srcRect:self->originalRect_
                                         withRotater:self->rotate_];
        [edit executeRotate];
        [edit release];
    }

    return;
}


//
// ž(ξ)
//
//  Call
//    originalPoints_ : ɸ(ѷ)(instance ѿ)
//    originalRect_   : Ĺ(ѷ)(instance ѿ)
//    originalShape_  : (ѷ)(instance ѿ)
//    originalImage_  : (ѷ)(instance ѿ)
//    resultRect_     : Ĺ(ѷ)(instance ѿ)
//    rotate_         : žؿ(instance ѿ)
//
//  Return
//    resultShape_  : (ѷ)(instance ѿ)
//    resultImage_  : (ѷ)(instance ѿ)
//
-(void)createRotateBitmapTwin
{
    PoCoEditRotateBitmapDouble *edit;
    int w;
    int h;

    // η̤˺
    [self->resultShape_ release];
    [self->resultImage_ release];
    self->resultShape_ = nil;
    self->resultImage_ = nil;
    if (!([self->resultRect_ empty])) {
        // ̤γǼ
        w = ([self->resultRect_ width] + 1);
        h = ([self->resultRect_ height] + 1);
        self->resultShape_ = [[PoCoBitmap alloc] initWidth:w
                                                initHeight:h
                                              defaultColor:0];
        self->resultImage_ = [[PoCoBitmap alloc] initWidth:w
                                                initHeight:h
                                              defaultColor:0];

        // ž
        edit = [[PoCoEditRotateBitmapDouble alloc] initDst1:self->resultShape_
                                                   withDst2:self->resultImage_
                                                   withSrc1:self->originalShape_
                                                   withSrc2:self->originalImage_
                                                    dstRect:self->resultRect_
                                                    srcRect:self->originalRect_
                                                withRotater:self->rotate_];
        [edit executeRotate];
        [edit release];
    }

    return;
}


//
// ȿž(ʿ)
//
//  Call
//    org             : ѷ
//    originalPoints_ : ɸ(ѷ)(instance ѿ)
//    resultRect_     : Ĺ(ѷ)(instance ѿ)
//
//  Return
//    bmp : ѷ
//
-(void)createFlipHoriBitmap:(PoCoBitmap *)org
               targetBitmap:(PoCoBitmap **)bmp
{
    PoCoEditFlipHoriBitmap *edit;
    int w;
    int h;

    [*(bmp) release];
    *(bmp) = nil;
    if (!([self->resultRect_ empty])) {
        w = ([self->resultRect_ width] + 1);
        h = ([self->resultRect_ height] + 1);
        *(bmp) = [[PoCoBitmap alloc] initWidth:w
                                    initHeight:h
                                  defaultColor:0];
        edit = [[PoCoEditFlipHoriBitmap alloc] initDst:*(bmp)
                                               withSrc:org];
        [edit executeFlip];
        [edit release];
    }

    return;
}


//
// ȿž(ľ)
//
//  Call
//    org             : ѷ
//    originalPoints_ : ɸ(ѷ)(instance ѿ)
//    resultRect_     : Ĺ(ѷ)(instance ѿ)
//
//  Return
//    bmp : ѷ
//
-(void)createFlipVertBitmap:(PoCoBitmap *)org
               targetBitmap:(PoCoBitmap **)bmp
{
    PoCoEditFlipVertBitmap *edit;
    int w;
    int h;

    [*(bmp) release];
    *(bmp) = nil;
    if (!([self->resultRect_ empty])) {
        w = ([self->resultRect_ width] + 1);
        h = ([self->resultRect_ height] + 1);
        *(bmp) = [[PoCoBitmap alloc] initWidth:w
                                    initHeight:h
                                  defaultColor:0];
        edit = [[PoCoEditFlipVertBitmap alloc] initDst:*(bmp)
                                               withSrc:org];
        [edit executeFlip];
        [edit release];
    }

    return;
}


// ------------------------------------------------ instance - private - ѷ
//
// ư
//
//  Call
//    p               : ΰư
//    evt             : ٥
//    liveShape       : ༡
//    liveResult      : ̤༡
//    originalPoints_ : ɸ(ѷ)(instance ѿ)
//    originalShape_  : (ѷ)(instance ѿ)
//    originalImage_  : (ѷ)(instance ѿ)
//    point_          : ()(instance ѿ)
//    modify_         : ѷν(instance ѿ)
//
//  Return
//    resultPoints_ : ɸ(ѷ)(instance ѿ)
//    resultRect_   : Ĺ(ѷ)(instance ѿ)
//    resultShape_  : (ѷ)(instance ѿ)
//    resultImage_  : (ѷ)(instance ѿ)
//    modify_       : ѷν(instance ѿ)
//
-(void)moveTrans:(PoCoPoint *)p
       withEvent:(NSEvent *)evt
     updateShape:(BOOL)liveShape
    updateResult:(BOOL)liveResult
{
    int dx;
    int dy;
    NSEnumerator* iter;
    PoCoPoint *tp1;
    PoCoPoint *tp2;

    if ([self isInnerHandle]) {
        // η̤˺
        [self clearResult];

        // «
        if (([evt modifierFlags] & NSShiftKeyMask) != 0x00) {
            // ư
            if (self->modify_ == PoCoModifierType_move_hori) {
                // ʿưݻ
                [p setY:[self->point_ y]];
            } else if (self->modify_ == PoCoModifierType_move_vert) {
                // ľưݻ
                [p setX:[self->point_ x]];
            } else if (abs([self->point_ y] - [p y]) < abs([self->point_ x] - [p x])) {
                // ʿư򳫻
                self->modify_ = PoCoModifierType_move_hori;
                [p setY:[self->point_ y]];
            } else {
                // ľư򳫻
                self->modify_ = PoCoModifierType_move_vert;
                [p setX:[self->point_ x]];
            }
        } else {
            // ̵
            self->modify_ = PoCoModifierType_none;
        }

        // ư̤򻻽
        dx = ([p x] - [self->point_ x]);
        dy = ([p y] - [self->point_ y]);

        // ɸư
        if ((self->originalPoints_ == nil) ||
            ([self->originalPoints_ count] == 0)) {
            // ʤ(ɤ)ϳĹѷ
            [self->resultRect_   setLeft:([self->originalRect_   left] + dx)];
            [self->resultRect_    setTop:([self->originalRect_    top] + dy)];
            [self->resultRect_  setRight:([self->originalRect_  right] + dx)];
            [self->resultRect_ setBottom:([self->originalRect_ bottom] + dy)];
        } else {
            // ѷ
            iter = [self->originalPoints_ objectEnumerator];
            for (tp1 = [iter nextObject]; tp1 != nil; tp1 = [iter nextObject]) {
                tp2 = [[PoCoPoint alloc] initX:([tp1 x] + dx)
                                         initY:([tp1 y] + dy)];
                [PoCoSelectionShape addControlPoint:tp2
                                        targetArray:self->resultPoints_
                                         targetRect:self->resultRect_];
                [tp2 release];
            }
        }

        // ѷη
        if (liveResult) {
            [self createMoveBitmapTwin];
        } else if (liveShape) {
            [self createMoveBitmap:self->originalShape_
                      targetBitmap:&(self->resultShape_)];
        }
    }

    return;
}


//
// 
//
//  Call
//    p               : ΰư
//    evt             : ٥
//    liveShape       : ༡
//    liveResult      : ̤༡
//    originalPoints_ : ɸ(ѷ)(instance ѿ)
//    originalRect_   : Ĺ(ѷ)(instance ѿ)
//    originalShape_  : (ѷ)(instance ѿ)
//    originalImage_  : (ѷ)(instance ѿ)
//    handle_         : ѷΥϥɥ(instance ѿ)
//    point_          : ()(instance ѿ)
//    modify_         : ѷν(instance ѿ)
//
//  Return
//    resultPoints_ : ɸ(ѷ)(instance ѿ)
//    resultRect_   : Ĺ(ѷ)(instance ѿ)
//    resultShape_  : (ѷ)(instance ѿ)
//    resultImage_  : (ѷ)(instance ѿ)
//    modify_       : ѷν(instance ѿ)
//
-(void)resizeTrans:(PoCoPoint *)p
         withEvent:(NSEvent *)evt
       updateShape:(BOOL)liveShape
      updateResult:(BOOL)liveResult
{
    float h;
    float v;
    float x;
    float y;
    PoCoRect *r;
    NSEnumerator* iter;
    PoCoPoint *tp1;
    PoCoPoint *tp2;

    if ([self isCornerHandle]) {
        // «
        if (([evt modifierFlags] & NSShiftKeyMask) != 0x00) {
            if ((self->modify_ == PoCoModifierType_resize_similer_hori) ||
                (self->modify_ == PoCoModifierType_resize_similer_vert)) {
                // ѷݻ
                [self similerPoint:p];
            } else if (self->modify_ == PoCoModifierType_resize_hori) {
                // ʿѷݻ
                [p setY:[self->point_ y]];
            } else if (self->modify_ == PoCoModifierType_resize_vert) {
                // ľѷݻ
                [p setX:[self->point_ x]];
            } else if (self->handle_ == PoCoHandleType_corner_lt) {
                // 
                if ((([self->point_ x] <= [p x])  &&
                     ([self->point_ y] <= [p y])) ||
                    (([self->point_ x] >  [p x])  &&
                     ([self->point_ y] >  [p y]))) {
                    // ѷ򳫻
                    [self similerPoint:p];
                } else if (([self->point_ x] <= [p x]) &&
                           ([self->point_ y] >  [p y])) {
                    // ľѷ򳫻
                    self->modify_ = PoCoModifierType_resize_vert;
                    [p setX:[self->point_ x]];
                } else /* if (([self->point_ x] >  [p x]) &&
                              ([self->point_ y] <= [p y])) */ {
                    // ʿѷ򳫻
                    self->modify_ = PoCoModifierType_resize_hori;
                    [p setY:[self->point_ y]];
                }
            } else if (self->handle_ == PoCoHandleType_corner_rt) {
                // 
                if ((([self->point_ x] <= [p x])  &&
                     ([self->point_ y] >  [p y])) ||
                    (([self->point_ x] >  [p x])  &&
                     ([self->point_ y] <= [p y]))) {
                    // ѷ򳫻
                    [self similerPoint:p];
                } else if (([self->point_ x] <= [p x]) &&
                           ([self->point_ y] <= [p y])) {
                    // ʿѷ򳫻
                    self->modify_ = PoCoModifierType_resize_hori;
                    [p setY:[self->point_ y]];
                } else /* if (([self->point_ x] >  [p x]) &&
                              ([self->point_ y] >  [p y])) */ {
                    // ľѷ򳫻
                    self->modify_ = PoCoModifierType_resize_vert;
                    [p setX:[self->point_ x]];
                }
            } else if (self->handle_ == PoCoHandleType_corner_rb) {
                // 
                if ((([self->point_ x] <= [p x])  &&
                     ([self->point_ y] <= [p y])) ||
                    (([self->point_ x] >  [p x])  &&
                     ([self->point_ y] >  [p y]))) {
                    // ѷ򳫻
                    [self similerPoint:p];
                } else if (([self->point_ x] <= [p x]) &&
                           ([self->point_ y] >  [p y])) {
                    // ʿѷ򳫻
                    self->modify_ = PoCoModifierType_resize_hori;
                    [p setY:[self->point_ y]];
                } else /* if (([self->point_ x] >  [p x]) &&
                              ([self->point_ y] <= [p y])) */ {
                    // ľѷ򳫻
                    self->modify_ = PoCoModifierType_resize_vert;
                    [p setX:[self->point_ x]];
                }
            } else /* if (self->handle_ == PoCoHandleType_corner_lb) */ {
                // 
                if ((([self->point_ x] <= [p x])  &&
                     ([self->point_ y] >  [p y])) ||
                    (([self->point_ x] >  [p x])  &&
                     ([self->point_ y] <= [p y]))) {
                    // ѷ򳫻
                    [self similerPoint:p];
                } else if (([self->point_ x] <= [p x]) &&
                           ([self->point_ y] <= [p y])) {
                    // ľѷ򳫻
                    self->modify_ = PoCoModifierType_resize_vert;
                    [p setX:[self->point_ x]];
                } else /* if (([self->point_ x] >  [p x]) &&
                              ([self->point_ y] >  [p y])) */ {
                    // ʿѷ򳫻
                    self->modify_ = PoCoModifierType_resize_hori;
                    [p setY:[self->point_ y]];
                }
            }
        } else {
            // ̵
            self->modify_ = PoCoModifierType_none;
        }

        // ܷ̤򻻽
        r = [[PoCoRect alloc] initPoCoRect:self->originalRect_];
        if (self->handle_ == PoCoHandleType_corner_lt) {
            [r setLeft:[p x]];
            [r setTop:[p y]];
        } else if (self->handle_ == PoCoHandleType_corner_rt) {
            [r setRight:[p x]];
            [r setTop:[p y]];
        } else if (self->handle_ == PoCoHandleType_corner_rb) {
            [r setRight:[p x]];
            [r setBottom:[p y]];
        } else /* if (self->handle_ == PoCoHandleType_corner_lb) */ {
            [r setLeft:[p x]];
            [r setBottom:[p y]];
        }
        if (!([r empty])) {
            // η̤˺
            [self clearResult];

            // ܷ̤򹹿
            [self->resultRect_ release];
            self->resultRect_ = [[PoCoRect alloc] initPoCoRect:r];

            // Ψ򻻽
            h = ((float)([self->resultRect_ width]) / (float)([self->originalRect_ width]));
            v = ((float)([self->resultRect_ height]) / (float)([self->originalRect_ height]));

            // ɸ
            if ((self->originalPoints_ == nil) ||
                ([self->originalPoints_ count] == 0)) {
                // ʤ(ɤ)ϳĹѷ
                [self->resultRect_   setLeft:[self->originalRect_   left]];
                [self->resultRect_    setTop:[self->originalRect_    top]];
                [self->resultRect_  setRight:[self->originalRect_  right]];
                [self->resultRect_ setBottom:[self->originalRect_ bottom]];
                if (self->handle_ == PoCoHandleType_corner_lt) {
                    [self->resultRect_ setLeft:(int)(floor(((float)([self->originalRect_ right]) - ((float)([self->originalRect_ right] - [self->originalRect_ left]) * h)) + 0.5))];
                    [self->resultRect_ setTop:(int)(floor(((float)([self->originalRect_ bottom]) - ((float)([self->originalRect_ bottom] - [self->originalRect_ top]) * v)) + 0.5))];
                } else if (self->handle_ == PoCoHandleType_corner_rt) {
                    [self->resultRect_ setRight:(int)(floor(((float)([self->originalRect_ left]) + ((float)([self->originalRect_ right] - [self->originalRect_ left]) * h)) + 0.5))];
                    [self->resultRect_ setTop:(int)(floor(((float)([self->originalRect_ bottom]) - ((float)([self->originalRect_ bottom] - [self->originalRect_ top]) * v)) + 0.5))];
                } else if (self->handle_ == PoCoHandleType_corner_rb) {
                    [self->resultRect_ setRight:(int)(floor(((float)([self->originalRect_ left]) + ((float)([self->originalRect_ right] - [self->originalRect_ left]) * h)) + 0.5))];
                    [self->resultRect_ setBottom:(int)(floor(((float)([self->originalRect_ top]) + ((float)([self->originalRect_ bottom] - [self->originalRect_ top]) * v)) + 0.5))];
                } else /* if (self->handle_ == PoCoHandleType_corner_lb) */ {
                    [self->resultRect_ setLeft:(int)(floor(((float)([self->originalRect_ right]) - ((float)([self->originalRect_ right] - [self->originalRect_ left]) * h)) + 0.5))];
                    [self->resultRect_ setBottom:(int)(floor(((float)([self->originalRect_ top]) + ((float)([self->originalRect_ bottom] - [self->originalRect_ top]) * v)) + 0.5))];
                }
            } else {
                // ѷ
                iter = [self->originalPoints_ objectEnumerator];
                for (tp1 = [iter nextObject]; tp1 != nil; tp1 = [iter nextObject]) {
                    if (self->handle_ == PoCoHandleType_corner_lt) {
                        x = ((float)([self->originalRect_ right]) - ((float)([self->originalRect_ right] - [tp1 x]) * h));
                        y = ((float)([self->originalRect_ bottom]) - ((float)([self->originalRect_ bottom] - [tp1 y]) * v));
                    } else if (self->handle_ == PoCoHandleType_corner_rt) {
                        x = ((float)([self->originalRect_ left]) + ((float)([tp1 x] - [self->originalRect_ left]) * h));
                        y = ((float)([self->originalRect_ bottom]) - ((float)([self->originalRect_ bottom] - [tp1 y]) * v));
                    } else if (self->handle_ == PoCoHandleType_corner_rb) {
                        x = ((float)([self->originalRect_ left]) + ((float)([tp1 x] - [self->originalRect_ left]) * h));
                        y = ((float)([self->originalRect_ top]) + ((float)([tp1 y] - [self->originalRect_ top]) * v));
                    } else /* if (self->handle_ == PoCoHandleType_corner_lb) */ {
                        x = ((float)([self->originalRect_ right]) - ((float)([self->originalRect_ right] - [tp1 x]) * h));
                        y = ((float)([self->originalRect_ top]) + ((float)([tp1 y] - [self->originalRect_ top]) * v));
                    }
                    tp2 = [[PoCoPoint alloc] initX:(int)(floor(x + 0.5))
                                             initY:(int)(floor(y + 0.5))];
                    [PoCoSelectionShape addControlPoint:tp2
                                            targetArray:self->resultPoints_
                                             targetRect:self->resultRect_];
                    [tp2 release];
                }
            }

            // ѷη
            if (liveResult) {
                [self createResizeBitmapTwin];
            } else if (liveShape) {
                [self createResizeBitmap:self->originalShape_
                            targetBitmap:&(self->resultShape_)];
            }
        }
        [r release];
    }

    return;
}


//
// ž
//
//  Call
//    p               : ΰư
//    evt             : ٥
//    liveShape       : ༡
//    liveResult      : ̤༡
//    originalPoints_ : ɸ(ѷ)(instance ѿ)
//    originalShape_  : (ѷ)(instance ѿ)
//    originalImage_  : (ѷ)(instance ѿ)
//    modify_         : ѷν(instance ѿ)
//
//  Return
//    resultPoints_ : ɸ(ѷ)(instance ѿ)
//    resultRect_   : Ĺ(ѷ)(instance ѿ)
//    rotate_       : žؿ(instance ѿ)
//    resultShape_  : (ѷ)(instance ѿ)
//    resultImage_  : (ѷ)(instance ѿ)
//    modify_       : ѷν(instance ѿ)
//
-(void)rotateTrans:(PoCoPoint *)p
         withEvent:(NSEvent *)evt
       updateShape:(BOOL)liveShape
      updateResult:(BOOL)liveResult
{
    NSEnumerator* iter;
    PoCoPoint *tp1;
    PoCoPoint *tp2;

    if ([self isEdgeHandle]) {
        // «
        if (([evt modifierFlags] & NSShiftKeyMask) != 0x00) {
            // 
            self->modify_ = PoCoModifierType_rotate;
        } else {
            // ̵
            self->modify_ = PoCoModifierType_none;
        }

        // ٤򻻽
        if ([self->rotate_ moveControlPoint:p
                                  isCorrect:(self->modify_ == PoCoModifierType_rotate)]) {
            // η̤˺
            [self clearResult];

            // ɸž
            if ((self->originalPoints_ == nil) ||
                ([self->originalPoints_ count] == 0)) {
                // ʤ(ɤ)ѷηΤ٤Ƥ
                [self calcRotateRectFromOriginalShape];
            } else {
                // ѷ
                iter = [self->originalPoints_ objectEnumerator];
                for (tp1 = [iter nextObject]; tp1 != nil; tp1 = [iter nextObject]) {
                    tp2 = [[PoCoPoint alloc] initX:[tp1 x]
                                             initY:[tp1 y]];
                    [self->rotate_ calcPoint:tp2];
                    [PoCoSelectionShape addControlPoint:tp2
                                            targetArray:self->resultPoints_
                                             targetRect:self->resultRect_];
                    [tp2 release];
                }
            }

            // ѷη
            if (liveResult) {
                [self createRotateBitmapTwin];
            } else if (liveShape) {
                [self createRotateBitmap:self->originalShape_
                            targetBitmap:&(self->resultShape_)];
            }
        }
    }

    return;
}


//
// ȿž(ʿ)
//
//  Call
//    originalPoints_ : ɸ(ѷ)(instance ѿ)
//    originalRect_   : Ĺ(ѷ)(instance ѿ)
//
//  Return
//    resultPoints_ : ɸ(ѷ)(instance ѿ)
//    resultRect_   : Ĺ(ѷ)(instance ѿ)
//
-(void)flipHori
{
    int left;
    int right;
    NSEnumerator* iter;
    PoCoPoint *tp1;
    PoCoPoint *tp2;

    left = [self->originalRect_ left];
    right = [self->originalRect_ right];

    // η̤˺
    [self->resultPoints_ removeAllObjects];

    // ɸư
    iter = [self->originalPoints_ objectEnumerator];
    for (tp1 = [iter nextObject]; tp1 != nil; tp1 = [iter nextObject]) {
        tp2 = [[PoCoPoint alloc] initX:(right - ([tp1 x] - left))
                                 initY:[tp1 y]];
        [self->resultPoints_ addObject:tp2];
        [tp2 release];
    }

    return;
}


//
// ȿž(ľ)
//
//  Call
//    originalPoints_ : ɸ(ѷ)(instance ѿ)
//    originalRect_   : Ĺ(ѷ)(instance ѿ)
//
//  Return
//    resultPoints_ : ɸ(ѷ)(instance ѿ)
//    resultRect_   : Ĺ(ѷ)(instance ѿ)
//
-(void)flipVert
{
    int top;
    int bottom;
    NSEnumerator* iter;
    PoCoPoint *tp1;
    PoCoPoint *tp2;

    top = [self->originalRect_ top];
    bottom = [self->originalRect_ bottom];

    // Ƥ˺
    [self->resultPoints_ removeAllObjects];

    // ɸư
    iter = [self->originalPoints_ objectEnumerator];
    for (tp1 = [iter nextObject]; tp1 != nil; tp1 = [iter nextObject]) {
        tp2 = [[PoCoPoint alloc] initX:[tp1 x]
                                 initY:(bottom - ([tp1 y] - top))];
        [self->resultPoints_ addObject:tp2];
        [tp2 release];
    }

    return;
}


// ------------------------------------------------------------- class - public
//
// Ͷ
//
//  Call
//    type : 
//
//  Return
//    function : Ƚ
//
+(BOOL)isCorner:(PoCoHandleType)type
{
    return ((type == PoCoHandleType_corner_lt) ||
            (type == PoCoHandleType_corner_rt) ||
            (type == PoCoHandleType_corner_rb) ||
            (type == PoCoHandleType_corner_lb));
}


//
// դ
//
//  Call
//    type : 
//
//  Return
//    function : Ƚ
//
+(BOOL)isEdge:(PoCoHandleType)type
{
    return ((type == PoCoHandleType_edge_t) ||
            (type == PoCoHandleType_edge_r) ||
            (type == PoCoHandleType_edge_b) ||
            (type == PoCoHandleType_edge_l));
}


//
// ΰ⤫
//
//  Call
//    type : 
//
//  Return
//    function : Ƚ
//
+(BOOL)isInner:(PoCoHandleType)type
{
    return ((type == PoCoHandleType_center) ||
            (type == PoCoHandleType_in_rect) ||
            (type == PoCoHandleType_in_shape));
}


//
// ΰ賰
//
//  Call
//    type : 
//
//  Return
//    function : Ƚ
//
+(BOOL)isOuter:(PoCoHandleType)type
{
    return ((type == PoCoHandleType_effective) ||
            (type == PoCoHandleType_out_rect) ||
            (type == PoCoHandleType_out_shape) ||
            (type == PoCoHandleType_unknown));
}


// ---------------------------------------------------------- instance - public
//
// initialize
//
//  Call
//    None
//
//  Return
//    originalPoints_   : ɸ(ѷ)(instance ѿ)
//    originalRect_     : Ĺ(ѷ)(instance ѿ)
//    originalHandle_[] : ϥɥ(ѷ)(instance ѿ)
//    originalShape_    : (ѷ)(instance ѿ)
//    originalImage_    : (ѷ)(instance ѿ)
//    resultPoints_     : ɸ(ѷ)(instance ѿ)
//    resultRect_       : Ĺ(ѷ)(instance ѿ)
//    resultHandle_[]   : ϥɥ(ѷ)(instance ѿ)
//    resultShape_      : (ѷ)(instance ѿ)
//    resultImage_      : (ѷ)(instance ѿ)
//    handle_           : ѷΥϥɥ(instance ѿ)
//    point_            : ()(instance ѿ)
//    rotate_           : žؿ(instance ѿ)
//    modify_           : ѷν(instance ѿ)
//
-(id)init
{
    int l;

    DPRINT((@"[PoCoSelectionShape init]\n"));

    // super class ν
    self = [super init];

    // Ȥν
    if (self != nil) {
        for (l = 0; l < (int)(PoCoHandleType_effective); (l)++) {
            self->originalHandle_[l] = nil;
            self->resultHandle_[l] = nil;
        }
        self->originalPoints_ = nil;
        self->originalRect_ = nil;
        self->originalShape_ = nil;
        self->originalImage_ = nil;
        self->resultPoints_ = nil;
        self->resultRect_ = nil;
        self->resultShape_ = nil;
        self->resultImage_ = nil;
        self->handle_ = PoCoHandleType_unknown;
        self->point_ = nil;
        self->rotate_ = nil;
        self->modify_ = PoCoModifierType_none;

        // 񸻤
        self->originalPoints_ = [[NSMutableArray alloc] init];
        self->resultPoints_ = [[NSMutableArray alloc] init];
        self->originalRect_ = [[PoCoRect alloc] initLeft:INT_MAX
                                                 initTop:INT_MAX
                                               initRight:INT_MIN
                                              initBottom:INT_MIN];
        self->resultRect_ = [[PoCoRect alloc] initLeft:INT_MAX
                                               initTop:INT_MAX
                                             initRight:INT_MIN
                                            initBottom:INT_MIN];
        if ((self->originalPoints_ == nil) ||
            (self->originalRect_ == nil) ||
            (self->resultPoints_ == nil) ||
            (self->resultRect_ == nil)) {
            DPRINT((@"can't allocation.\n"));
            [self release];
            self = nil;
        }
    }

    return self;
}


//
// deallocate
//
//  Call
//    None
//
//  Return
//    originalPoints_   : ɸ(ѷ)(instance ѿ)
//    originalRect_     : Ĺ(ѷ)(instance ѿ)
//    originalHandle_[] : ϥɥ(ѷ)(instance ѿ)
//    originalShape_    : (ѷ)(instance ѿ)
//    originalImage_    : (ѷ)(instance ѿ)
//    resultPoints_     : ɸ(ѷ)(instance ѿ)
//    resultRect_       : Ĺ(ѷ)(instance ѿ)
//    resultHandle_[]   : ϥɥ(ѷ)(instance ѿ)
//    resultShape_      : (ѷ)(instance ѿ)
//    resultImage_      : (ѷ)(instance ѿ)
//    point_            : ()(instance ѿ)
//    rotate_           : žؿ(instance ѿ)
//
-(void)dealloc
{
    int l;

    DPRINT((@"[PoCoSelectionShape dealloc]\n"));

    // 񸻤
    for (l = 0; l < (int)(PoCoHandleType_effective); (l)++) {
        [self->originalHandle_[l] release];
        [self->resultHandle_[l] release];
        self->originalHandle_[l] = nil;
        self->resultHandle_[l] = nil;
    }
    [self->originalPoints_ release];
    [self->originalRect_ release];
    [self->originalShape_ release];
    [self->originalImage_ release];
    [self->resultPoints_ release];
    [self->resultRect_ release];
    [self->resultShape_ release];
    [self->resultImage_ release];
    [self->point_ release];
    [self->rotate_ release];
    self->originalPoints_ = nil;
    self->originalRect_ = nil;
    self->originalShape_ = nil;
    self->originalImage_ = nil;
    self->resultPoints_ = nil;
    self->resultRect_ = nil;
    self->resultShape_ = nil;
    self->resultImage_ = nil;
    self->point_ = nil;
    self->rotate_ = nil;

    // super class β
    [super dealloc];

    return;
}


//
// ѷη˺
//
//  Call
//    None
//
//  Return
//    originalPoints_   : ɸ(ѷ)(instance ѿ)
//    originalRect_     : Ĺ(ѷ)(instance ѿ)
//    originalHandle_[] : ϥɥ(ѷ)(instance ѿ)
//    originalShape_    : (ѷ)(instance ѿ)
//
-(void)clearOriginal
{
    int l;

    [self->originalPoints_ removeAllObjects];
    [self->originalRect_   setLeft:INT_MAX];
    [self->originalRect_    setTop:INT_MAX];
    [self->originalRect_  setRight:INT_MIN];
    [self->originalRect_ setBottom:INT_MIN];
    [self->originalShape_ release];
    for (l = 0; l < (int)(PoCoHandleType_effective); (l)++) {
        [self->originalHandle_[l] release];
        self->originalHandle_[l] = nil;
    }
    self->originalShape_ = nil;

    return;
}


//
// ѷλϿ
//
//  Call
//    p : Ͽ
//        == nil : 
//        != nil : ɲ
//
//  Return
//    originalPoints_   : ɸ(ѷ)(instance ѿ)
//    originalRect_     : Ĺ(ѷ)(instance ѿ)
//    originalHandle_[] : ϥɥ(ѷ)(instance ѿ)
//    originalShape_    : (ѷ)(instance ѿ)
//
-(void)addPoint:(PoCoPoint *)p
{
    if (p != nil) {
        // ɲ
        [PoCoSelectionShape addControlPoint:p
                                targetArray:self->originalPoints_
                                 targetRect:self->originalRect_];
    } else {
        // 
        [PoCoSelectionShape createShape:self->originalPoints_
                               withRect:self->originalRect_
                          targetHandles:(PoCoRect **)(&(self->originalHandle_))
                            targetShape:&(self->originalShape_)];
        [self copyOriginalToResult];
    }

    return;
}


//
// ѷβϿ
//  ѷȤƳФƤñ˼ΤؤΥݥʣ̤ʤΤǡ
//  ƤӽФ bmp  release (ƤӽФ bmp  alloc ɬ)
//
//  Call
//    bmp : 
//
//  Return
//    originalImage_ : (ѷ)(instance ѿ)
//    resultImage_   : (ѷ)(instance ѿ)
//
-(void)setImage:(PoCoBitmap *)bmp
{
    // ʬ˺
    [self->originalImage_ release];
    [self->resultImage_ release];

    // ѷñ˼ΤؤΥݥʣ
    self->originalImage_ = bmp;

    // ѷʣ
    self->resultImage_ = [bmp copy];

    return;
}


//
// ѷηѷηʣ
//
//  Call
//    resultPoints_   : ɸ(ѷ)(instance ѿ)
//    resultRect_     : Ĺ(ѷ)(instance ѿ)
//    resultHandle_[] : ϥɥ(ѷ)(instance ѿ)
//    resultShape_    : (ѷ)(instance ѿ)
//
//  Return
//    originalPoints_   : ɸ(ѷ)(instance ѿ)
//    originalRect_     : Ĺ(ѷ)(instance ѿ)
//    originalHandle_[] : ϥɥ(ѷ)(instance ѿ)
//    originalShape_    : (ѷ)(instance ѿ)
//
-(void)copyResultToOriginal
{
    NSEnumerator* iter;
    PoCoPoint *p;
    PoCoPoint *tmp;
    int l;

    // Ƥ˺
    [self clearOriginal];

    // ɸ
    [self->originalPoints_ release];
    self->originalPoints_ = [[NSMutableArray alloc] init];
    iter = [self->resultPoints_ objectEnumerator];
    for (p = [iter nextObject]; p != nil; p = [iter nextObject]) {
        tmp = [[PoCoPoint alloc] initX:[p x]
                                 initY:[p y]];
        [self->originalPoints_ addObject:tmp];
        [tmp release];
    }

    // Ĺ
    [self->originalRect_ release];
    self->originalRect_ = [[PoCoRect alloc]
                              initLeft:[self->resultRect_ left]
                               initTop:[self->resultRect_ top]
                             initRight:[self->resultRect_ right]
                            initBottom:[self->resultRect_ bottom]];

    // ޥ
    self->originalShape_ = [self->resultShape_ copy];

    // ϥɥ
    for (l = 0; l < (int)(PoCoHandleType_effective); (l)++) {
        self->originalHandle_[l] = [[PoCoRect alloc]
              initLeft:[self->resultHandle_[l] left]
               initTop:[self->resultHandle_[l] top]
             initRight:[self->resultHandle_[l] right]
            initBottom:[self->resultHandle_[l] bottom]];
    }

    return;
}


// ----------------------------------------------- instance - public - 
//
// ѷΥϥɥ
//
//  Call
//    handle_ : ѷΥϥɥ(instance ѿ)
//
//  Return
//    function : ѷΥϥɥ
//
-(PoCoHandleType)controlHandle
{
    return self->handle_;
}


//
// 
//
//  Call
//    point_ : ()(instance ѿ)
//
//  Return
//    function : ()
//
-(PoCoPoint *)controlPoint
{
    return self->point_;
}


//
// žؿ
//
//  Call
//    rotate_ : žؿ(instance ѿ)
//
//  Return
//    rotate_ : žؿ
//
-(PoCoCalcRotation *)calcRotater
{
    return self->rotate_;
}


//
// Ͷ
//
//  Call
//    handle_ : ѷΥϥɥ(instance ѿ)
//
//  Return
//    function : Ƚ
//
-(BOOL)isCornerHandle
{
    return [PoCoSelectionShape isCorner:self->handle_];
}


//
// դ
//
//  Call
//    handle_ : ѷΥϥɥ(instance ѿ)
//
//  Return
//    function : Ƚ
//
-(BOOL)isEdgeHandle
{
    return [PoCoSelectionShape isEdge:self->handle_];
}


//
// ΰ⤫
//
//  Call
//    handle_ : ѷΥϥɥ(instance ѿ)
//
//  Return
//    function : Ƚ
//
-(BOOL)isInnerHandle
{
    return [PoCoSelectionShape isInner:self->handle_];
}


//
// ΰ賰
//
//  Call
//    handle_ : ѷΥϥɥ(instance ѿ)
//
//  Return
//    function : Ƚ
//
-(BOOL)isOuterHandle
{
    return [PoCoSelectionShape isOuter:self->handle_];
}


// ----------------------------------------- instance - public - ѷ
//
// ɸ
//
//  Call
//    originalPoints_ : ɸ(ѷ)(instance ѿ)
//
//  Return
//    function : ɸ(ѷ)
//
-(NSMutableArray *)originalPoints
{
    return self->originalPoints_;
}


//
// Ĺ
//
//  Call
//    originalRect_ : Ĺ(ѷ)(instance ѿ)
//
//  Return
//    function : Ĺ(ѷ)
//
-(PoCoRect *)originalRect
{
    return self->originalRect_;
}


//
// ϥɥ
//
//  Call
//    type              : 
//    originalHandle_[] : ϥɥ(ѷ)(instance ѿ)
//
//  Return
//    function : ϥɥ(ѷ)
//
-(PoCoRect *)originalHandle:(PoCoHandleType)type
{
    int i = (int)(type);

    return ((i < (int)(PoCoHandleType_effective)) ? self->originalHandle_[i] : nil);
}


//
// 
//
//  Call
//    originalShape_ : (ѷ)(instance ѿ)
//
//  Return
//    function : (ѷ)
//
-(PoCoBitmap *)originalShape
{
    return self->originalShape_;
}


//
// 
//
//  Call
//    originalImage_ : (ѷ)(instance ѿ)
//
//  Return
//    function : (ѷ)
//
-(PoCoBitmap *)originalImage
{
    return self->originalImage_;
}


//
// Ƚ
//
//  Call
//    p                 : ڤ
//    gap               : ϥɥγĥ
//    shape             : оݤȤ뤫
//                        YES : 鳰ĹȽꤹ
//                        NO  : ĹΤߤȽꤹ
//    originalRect_     : Ĺ(ѷ)(instance ѿ)
//    originalHandle_[] : ϥɥ(ѷ)(instance ѿ)
//    originalShape_    : (ѷ)(instance ѿ)
//
//  Return
//    function : 
//
-(PoCoHandleType)originalHitTest:(PoCoPoint *)p
                       handleGap:(int)gap
                       withShape:(BOOL)shape
{
    return [PoCoSelectionShape hitTest:p
                             handleGap:gap
                              withRect:self->originalRect_
                           withHanldes:(PoCoRect **)(&(self->originalHandle_))
                             withShape:((shape) ? self->originalShape_ : nil)];
}


//
// ĹȻƱΰ褫
//
//  Call
//    originalPoints_ : ɸ(ѷ)(instance ѿ)
//    originalRect_   : Ĺ(ѷ)(instance ѿ)
//
//  Return
//    function : Ƚ
//
-(BOOL)isSameStyleOriginalRectShape
{
    return [PoCoSelectionShape isSameStyle:self->originalRect_
                                 testShape:self->originalPoints_];
}


// ----------------------------------------- instance - public - ѷ
//
// ɸ
//
//  Call
//    resultPoints_ : ɸ(ѷ)(instance ѿ)
//
//  Return
//    function : ɸ(ѷ)
//
-(NSMutableArray *)resultPoints
{
    return self->resultPoints_;
}


//
// Ĺ
//
//  Call
//    resultRect_ : Ĺ(ѷ)(instance ѿ)
//
//  Return
//    function : Ĺ(ѷ)
//
-(PoCoRect *)resultRect
{
    return self->resultRect_;
}


//
// ϥɥ
//
//  Call
//    type            : 
//    resultHandle_[] : ϥɥ(ѷ)(instance ѿ)
//
//  Return
//    function : ϥɥ(ѷ)
//
-(PoCoRect *)resultHandle:(PoCoHandleType)type
{
    int i = (int)(type);

    return ((i < (int)(PoCoHandleType_effective)) ? self->resultHandle_[i] : nil);
}


//
// 
//
//  Call
//    resultShape_ : (ѷ)(instance ѿ)
//
//  Return
//    function : (ѷ)
//
-(PoCoBitmap *)resultShape
{
    return self->resultShape_;
}


//
// 
//
//  Call
//    resultImage_ : (ѷ)(instance ѿ)
//
//  Return
//    function : (ѷ)
//
-(PoCoBitmap *)resultImage
{
    return self->resultImage_;
}


//
// Ƚ
//
//  Call
//    p               : ڤ
//    gap             : ϥɥγĥ
//    shape           : оݤȤ뤫
//                      YES : 鳰ĹȽꤹ
//                      NO  : ĹΤߤȽꤹ
//    resultRect_     : Ĺ(ѷ)(instance ѿ)
//    resultHandle_[] : ϥɥ(ѷ)(instance ѿ)
//    resultShape_    : (ѷ)(instance ѿ)
//
//  Return
//    function : 
//
-(PoCoHandleType)resultHitTest:(PoCoPoint *)p
                     handleGap:(int)gap
                     withShape:(BOOL)shape
{
    return [PoCoSelectionShape hitTest:p
                             handleGap:gap
                              withRect:self->resultRect_
                           withHanldes:(PoCoRect **)(&(self->resultHandle_))
                             withShape:((shape) ? self->resultShape_ : nil)];
}


//
// ĹȻƱΰ褫
//
//  Call
//    resultPoints_ : ɸ(ѷ)(instance ѿ)
//    resultRect_   : Ĺ(ѷ)(instance ѿ)
//
//  Return
//    function : Ƚ
//
-(BOOL)isSameStyleResultRectShape
{
    return [PoCoSelectionShape isSameStyle:self->resultRect_
                                 testShape:self->resultPoints_];
}


// --------------------------------------------------- instance - public - ѷ
//
// 
//
//  Call
//    p     : ()
//    gap   : ϥɥγĥ
//    shape : оݤȤ뤫
//             YES : 鳰ĹȽꤹ
//             NO  : ĹΤߤȽꤹ
//
//  Return
//    function : Ϥ
//               YES : ѹ
//               NO  : ϤƤʤ(ϰϳ)
//    handle_  : ѷΥϥɥ(instance ѿ)
//    point_   : ()(instance ѿ)
//    rotate_  : žؿ(instance ѿ)
//    modify_  : ѷν(instance ѿ)
//
-(BOOL)startTrans:(PoCoPoint *)p
        handleGap:(int)gap
        withShape:(BOOL)shape
{
    BOOL result;
    BOOL v;
    BOOL g;

    result = YES;

    // ѷ˺
    [self->point_ release];
    [self->rotate_ release];
    self->point_ = nil;
    self->rotate_ = nil;
    self->handle_ = PoCoHandleType_unknown;
    self->modify_ = PoCoModifierType_none;

    // Ф
    self->point_ = [[PoCoPoint alloc] initX:[p x]
                                      initY:[p y]];

    // ѷ̤Ф hit test
    self->handle_ = [self resultHitTest:p
                              handleGap:gap
                              withShape:shape];
    if ([self isInnerHandle]) {
        // ư
        ;
    } else if ([self isCornerHandle]) {
        // ܷ
        ;
    } else if ([self isEdgeHandle]) {
        // ž
        switch (self->handle_) {
            default:
            case PoCoHandleType_edge_t:
                // 
                v = NO;
                g = NO;
                break;
            case PoCoHandleType_edge_b:
                // 
                v = NO;
                g = YES;
                break;
            case PoCoHandleType_edge_r:
                // 
                v = YES;
                g = YES;
                break;
            case PoCoHandleType_edge_l:
                // 
                v = YES;
                g = NO;
                break;
        }
        self->rotate_ = [[PoCoCalcRotation alloc]
                            initWithControlPoint:[self->resultHandle_[PoCoHandleType_center] lefttop]
                               isVerticalControl:v
                                isGreaterControl:g];
    } else /* if ([self isOterHandle]) */ {
        // 
        [self->point_ release];
        self->point_ = nil;
        self->handle_ = PoCoHandleType_unknown;
        self->modify_ = PoCoModifierType_none;
        result = NO;
    }

    return result;
}


//
// ѷ
//
//  Call
//    p          : ΰư
//    evt        : ٥
//    liveShape  : ༡
//    liveResult : ̤༡
//    handle_    : ѷΥϥɥ(instance ѿ)
//    point_     : ()(instance ѿ)
//
//  Return
//    None
//
-(void)runningTrans:(PoCoPoint *)p
          withEvent:(NSEvent *)evt
  isLiveUpdateShape:(BOOL)liveShape
 isLiveUpdateResult:(BOOL)liveResult
{
    PoCoPoint *tmp;

    // ѷν񤭤Τʣ
    tmp = [[PoCoPoint alloc] initX:[p x]
                             initY:[p y]];
    if ([self isInnerHandle]) {
        // ư
        [self moveTrans:tmp
              withEvent:evt
            updateShape:liveShape
           updateResult:liveResult];
    } else if ([self isCornerHandle]) {
        // ܷ
        [self resizeTrans:tmp
                withEvent:evt
              updateShape:liveShape
             updateResult:liveResult];
    } else if ([self isEdgeHandle]) {
        // ž
        [self rotateTrans:tmp
                withEvent:evt
              updateShape:liveShape
             updateResult:liveResult];
    } else /* if ([self isOterHandle]) */ {
        // 
        ;
    }
    [tmp release];

    return;
}


//
// λ
//
//  Call
//    originalShape_ : (ѷ)(instance ѿ)
//    originalImage_ : (ѷ)(instance ѿ)
//    resultPoints_  : ɸ(ѷ)(instance ѿ)
//    resultRect_    : Ĺ(ѷ)(instance ѿ)
//    handle_        : ѷΥϥɥ(instance ѿ)
//
//  Return
//    resultHandle_[] : ϥɥ(ѷ)(instance ѿ)
//    resultShape_    : (ѷ)(instance ѿ)
//    resultImage_    : (ѷ)(instance ѿ)
//
-(void)endTrans
{
    if ([self isOuterHandle]) {
        // 
        ;
    } else {
        // 
        [PoCoSelectionShape createShape:self->resultPoints_
                               withRect:self->resultRect_
                          targetHandles:(PoCoRect **)(&(self->resultHandle_))
                            targetShape:&(self->resultShape_)];

        if ([self isInnerHandle]) {
            // ư
            [self createMoveBitmapTwin];
            [self copyResultToOriginal];
        } else if ([self isCornerHandle]) {
            // ܷ
            [self createResizeBitmapTwin];
        } else if ([self isEdgeHandle]) {
            // ž
            [self createRotateBitmapTwin];
        }
    }

    return;
}


// ------------------------------------------- instance - public - ¾βù
//
// 
//
//  Call
//    doc           : document
//    info          : Խ
//    resultPoints_ : ɸ(ѷ)(instance ѿ)
//    resultRect_   : Ĺ(ѷ)(instance ѿ)
//    resultShape_  : (ѷ)(instance ѿ)
//
//  Return
//    resultImage_   : (ѷ)(instance ѿ)
//    originalImage_ : (ѷ)(instance ѿ)
//
-(void)delete:(MyDocument *)doc
 withEditInfo:(PoCoEditInfo *)info
#if 0   // (򿧤ǤɤĤ֤Τ)
{
    PoCoEditRegionFill *edit;
    PoCoMonochromePattern *tile;
    PoCoColorPattern *pat;
    PoCoPoint *sp;
    PoCoPoint *ep;

    edit = nil;
    tile = nil;
    pat = nil;

    // η̤˺
    [self->resultImage_ release];
    self->resultImage_ = nil;

    // ϰ֤ʤ
    if (!([self->resultRect_ empty])) {
        // ѥѥ
        tile = [[(PoCoAppController *)([NSApp delegate]) tilePattern] pattern:[info tileNumber]];  

        // ѥ
        if (([info eraserType]) && ([doc eraser] != nil)) {
            // β(alloc ϤʤΤ retain Ƥ)
            pat = [doc eraser];
            [pat retain];
        } else if (([[info selColor] isUnder]) && ([[doc selLayer] sel] > 0)) {
            // ̥쥤䡼
            pat = [[PoCoColorPattern alloc] initWithBitmap:[[[doc picture] layer:([[doc selLayer] sel] - 1)] bitmap]]; 
        } else if ([[info selColor] isPattern]) {
            // 顼ѥ(alloc ϤʤΤ retain Ƥ)
            pat = [[doc picture] colpat:[[info selColor] num]];
            [pat retain];
        } else {
            // ̾ο
            pat = [[PoCoColorPattern alloc] initWidth:1
                                           initHeight:1
                                         defaultColor:[[info selColor] num]];
        }

        // Խ¹
        self->resultImage_ = [[PoCoBitmap alloc]
                                     initWidth:([self->resultRect_ width] + 1)
                                    initHeight:([self->resultRect_ height] + 1)
                                  defaultColor:0];
        ep = [[PoCoPoint alloc] init];
        sp = [[PoCoPoint alloc] initX:[self->resultImage_ width]
                                initY:[self->resultImage_ height]];
        edit = [[PoCoEditRegionFill alloc]
                   initWithPattern:self->resultImage_
                           palette:[[doc picture] palette]
                              tile:tile
                           pattern:pat
                         checkDist:NO];
        [edit executeDraw:self->resultShape_
             withStartPos:sp
               withEndPos:ep];
        [edit release];
        [ep release];
        [sp release];

        // ѥ
        [pat release];

        // ̤ʣ
        [self copyResultToOriginal];
        [self->originalImage_ release];
        self->originalImage_ = [self->resultImage_ copy];
    }

    return;
}
#else   // 0  // ڥ˰¸Ȥ
{
    id edit;
    PoCoMonochromePattern *tile;
    PoCoColorPattern *pat;
    PoCoBitmap *behind;
    PoCoPoint *sp;
    PoCoPoint *ep;
    PoCoRect *r;
    int ratio;
    PoCoBitmap *shape;

    edit = nil;
    tile = nil;
    pat = nil;
    sp = nil;
    ep = nil;
    r = nil;
    shape = nil;
    behind = nil;

    // η̤˺
    [self->resultImage_ release];
    self->resultImage_ = nil;

    // ϰ̵ǤϲǤʤ
    if ([self->resultRect_ empty]) {
        goto EXIT;
    }

    // ѥѥ
    tile = [[(PoCoAppController *)([NSApp delegate]) tilePattern] pattern:[info tileNumber]];  

    // ѥ
    if (([info eraserType]) && ([doc eraser] != nil)) {
        // β(alloc ϤʤΤ retain Ƥ)
        pat = [doc eraser];
        [pat retain];
    } else if (([[info selColor] isUnder]) && ([[doc selLayer] sel] > 0)) {
        // ̥쥤䡼
        behind = [[[[doc picture] layer:([[doc selLayer] sel] - 1)] bitmap] getBitmap:self->resultRect_];
        pat = [[PoCoColorPattern alloc] initWithBitmap:behind]; 
        [behind release];
        behind = nil;
    } else if ([[info selColor] isPattern]) {
        // 顼ѥ(alloc ϤʤΤ retain Ƥ)
        pat = [[doc picture] colpat:[[info selColor] num]];
        [pat retain];
    } else {
        // ̾ο
        pat = [[PoCoColorPattern alloc] initWidth:1
                                       initHeight:1
                                     defaultColor:[[info selColor] num]];
    }

    // Խ¹
    if ([info penStyleType] == PoCoPenStyleType_Normal) {
        // ̾
        self->resultImage_ = [[PoCoBitmap alloc]
                                     initWidth:([self->resultRect_ width] + 1)
                                    initHeight:([self->resultRect_ height] + 1)
                                  defaultColor:0];
        ep = [[PoCoPoint alloc] init];
        sp = [[PoCoPoint alloc] initX:[self->resultImage_ width]
                                initY:[self->resultImage_ height]];
        edit = [[PoCoEditRegionFill alloc]
                   initWithPattern:self->resultImage_
                           palette:[[doc picture] palette]
                              tile:tile
                           pattern:pat
                         checkDist:NO];
        [edit executeDraw:self->resultShape_
             withStartPos:sp
               withEndPos:ep];
        [edit release];
        [ep release];
        [sp release];
    } else {
        self->resultImage_ = [self->originalImage_ copy];
        shape = [self->resultShape_ copy];
        r = [[PoCoRect alloc] initLeft:0
                               initTop:0
                             initRight:[self->resultImage_ width]
                            initBottom:[self->resultImage_ height]];
        switch ([info penStyleType]) {
            case PoCoPenStyleType_Normal:
            default:
                // ̾
                ;
                break;
            case PoCoPenStyleType_UniformedDensity:
            case PoCoPenStyleType_Density:
                // ñǻ
                // ǻ
                edit = [[PoCoEditUniformedDensityFill alloc]
                              init:self->resultImage_
                           colMode:[info colorMode]
                           palette:[[doc picture] palette]
                           pattern:pat
                            buffer:[doc colorBuffer]
                           density:[info density]];
                break;
            case PoCoPenStyleType_Atomizer:
                // ̸᤭
                ratio = [info density];
                if (ratio < 250) {
                    ratio /= 75;
                } else if (ratio < 750) {
                    ratio /= 2;
                    ratio /= 50;
                } else if (ratio < 1000) {
                    ratio /= 5;
                    ratio /= 25;
                }
                if (ratio > 0) {
                    edit = [[PoCoEditAtomizerFill alloc]
                                  init:self->resultImage_
                               palette:[[doc picture] palette]
                               pattern:pat
                                 ratio:ratio];
                }
                break;
            case PoCoPenStyleType_Gradation:
                // ǡ
                ratio = [info density];
                if (ratio < 250) {
                    ratio /= 75;
                } else if (ratio < 750) {
                    ratio /= 50;
                } else if (ratio < 1000) {
                    ratio /= 25;
                }
                if (ratio > 0) {
                    edit = [[PoCoEditGradationFill alloc]
                                  init:self->resultImage_
                               colMode:[info colorMode]
                               palette:[[doc picture] palette]
                               pattern:pat
                                buffer:[doc colorBuffer]
                                 ratio:ratio];
                    [edit replaceDensity:(1000 / ([info penSize] + 7))
                              withBitmap:self->resultImage_];
                }
                break;
            case PoCoPenStyleType_Random:
                // Ȼ
                edit = [[PoCoEditRandomFill alloc]
                              init:self->resultImage_
                           palette:[[doc picture] palette]
                             ratio:[info density]
                             range:([info penSize] * 8)];
                break;
            case PoCoPenStyleType_WaterDrop:
                // ܤ
                edit = [[PoCoEditWaterDropFill alloc]
                              init:self->resultImage_
                           colMode:[info colorMode]
                           palette:[[doc picture] palette]
                            buffer:[doc colorBuffer]];
                break;
        }
        [edit executeDraw:shape
                 withTile:tile
             withTrueRect:r
             withDrawRect:r];
        [edit release];
        [shape release];
        [r release];
    }

    // ѥ
    [pat release];

    // ̤ʣ
    [self copyResultToOriginal];
    [self->originalImage_ release];
    self->originalImage_ = [self->resultImage_ copy];

EXIT:
    return;
}
#endif  // 0


//
// ȿž
//
//  Call
//    hori           : ʿ
//                     YES : ʿ˲ȿž
//                     NO  : ľ˲ȿž
//    originalShape_ : (ѷ)(instance ѿ)
//    originalImage_ : (ѷ)(instance ѿ)
//
//  Return
//    resultShape_ : (ѷ)(instance ѿ)
//    resultImage_ : (ѷ)(instance ѿ)
//
-(void)flip:(BOOL)hori
{
    if (hori) {
        // ʿȿž
        [self flipHori];
        [self createFlipHoriBitmap:self->originalShape_
                      targetBitmap:&(self->resultShape_)];
        [self createFlipHoriBitmap:self->originalImage_
                      targetBitmap:&(self->resultImage_)];
    } else {
        // ľȿž
        [self flipVert];
        [self createFlipVertBitmap:self->originalShape_
                      targetBitmap:&(self->resultShape_)];
        [self createFlipVertBitmap:self->originalImage_
                      targetBitmap:&(self->resultImage_)];
    }

    // ̤ʣ
    [self copyResultToOriginal];
    [self->originalImage_ release];
    self->originalImage_ = [self->resultImage_ copy];

    return;
}


//
// ưǡ
//
//  Call
//    doc            : document
//    size           : 礭(0: Ǥա1-128 ϰ)
//    adj            : (YES: ٤礦ֹΤ)
//    mtx            : оݿ(256 ĹYES: о)
//    originalImage_ : (ѷ)(instance ѿ)
//    originalShape_ : (ѷ)(instance ѿ)
//    resultRect_    : Ĺ(ѷ)(instance ѿ)
//
//  Return
//    originalImage_ : (ѷ)(instance ѿ)
//    resultImage_   : (ѷ)(instance ѿ)
//
-(void)autoGrad:(MyDocument *)doc
        penSize:(int)size
     isAdjacent:(BOOL)adj
         matrix:(const BOOL *)mtx
   withSizePair:(NSDictionary *)sizePair
{
    id edit;

    edit = nil;

    // η̤˺
    [self->resultImage_ release];
    self->resultImage_ = nil;

    // ϰ֤ʤ
    if (!([self->resultRect_ empty])) {
        // Խ¹
        self->resultImage_ = [[PoCoBitmap alloc]
                                     initWidth:([self->resultRect_ width] + 1)
                                    initHeight:([self->resultRect_ height] + 1)
                                  defaultColor:0];
        if (adj) {
            // 
            edit = [[PoCoEditAutoGradationAdjacent alloc]
                       initDst:self->resultImage_
                       withSrc:self->originalImage_
                      withMask:self->originalShape_
                       palette:[[doc picture] palette]
                       penSize:size
                        matrix:mtx
                          rect:self->resultRect_
                  withSizePair:sizePair];
        } else {
            // Ǥ
            edit = [[PoCoEditAutoGradation alloc]
                       initDst:self->resultImage_
                       withSrc:self->originalImage_
                      withMask:self->originalShape_
                       palette:[[doc picture] palette]
                       penSize:size
                        matrix:mtx
                          rect:self->resultRect_];
        }
        [edit executeDraw];
        [edit release];

        // ̤ʣ
        [self copyResultToOriginal];
        [self->originalImage_ release];
        self->originalImage_ = [self->resultImage_ copy];
    }

    return;
}


//
// ִ
//
//  Call
//    doc            : document
//    mtx            : ִɽ(256 Ĺ)
//    originalImage_ : (ѷ)(instance ѿ)
//    originalShape_ : (ѷ)(instance ѿ)
//    resultRect_    : Ĺ(ѷ)(instance ѿ)
//
//  Return
//    originalImage_ : (ѷ)(instance ѿ)
//    resultImage_   : (ѷ)(instance ѿ)
//
-(void)colorReplace:(MyDocument *)doc
             matrix:(const unsigned char *)mtx
{
    PoCoEditBitmapColorReplace *edit;

    edit = nil;

    // η̤˺
    [self->resultImage_ release];
    self->resultImage_ = nil;

    // ϰ֤ʤ
    if (!([self->resultRect_ empty])) {
        // Խ¹
        self->resultImage_ = [[PoCoBitmap alloc]
                                     initWidth:([self->resultRect_ width] + 1)
                                    initHeight:([self->resultRect_ height] + 1)
                                  defaultColor:0];
        edit = [[PoCoEditBitmapColorReplace alloc]
                   initTargetBitmap:self->resultImage_
                   withSourceBitmap:self->originalImage_
                           withMask:self->originalShape_
                         withMatrix:mtx];
        [edit execute];
        [edit release];

        // ̤ʣ
        [self copyResultToOriginal];
        [self->originalImage_ release];
        self->originalImage_ = [self->resultImage_ copy];
    }

    return;
}


//
// ƥ
//
//  Call
//    doc           : document
//    info          : Խ
//    base          : ܿ(NSArray<unsigned int>)
//    grad          : ǻø(NSArray<unsinged int, NSArray<unsigned int> >)
//    resultPoints_ : ɸ(ѷ)(instance ѿ)
//    resultRect_   : Ĺ(ѷ)(instance ѿ)
//    resultShape_  : (ѷ)(instance ѿ)
//
//  Return
//    originalImage_ : (ѷ)(instance ѿ)
//    resultImage_   : (ѷ)(instance ѿ)
//
-(void)texture:(MyDocument *)doc
  withEditInfo:(PoCoEditInfo *)info
    baseColors:(NSArray *)base
gradientColors:(NSArray *)grad
{
    PoCoPalette *pal;
    PoCoEditRegionFill *fill;
    PoCoEditBitmapColorReplace *rep;
    PoCoMonochromePattern *tile;
    PoCoBitmap *behind;
    PoCoColorPattern *pat;
    PoCoPoint *sp;
    PoCoPoint *ep;
    NSEnumerator *iter;
    PoCoBaseGradientPair *pair;
    NSNumber *bnum;
    NSNumber *gnum;
    PoCoBaseGradientPair *bpair;
    int idx;
    unsigned char mtx[COLOR_MAX];
    BOOL mask[COLOR_MAX];
    BOOL drop[COLOR_MAX];

    fill = nil;
    rep = nil;
    tile = nil;
    pat = nil;
    behind = nil;
    sp = nil;
    ep = nil;
    pal = [[doc picture] palette];
    bpair = nil;

    // ϰ̵ǤϲǤʤ
    if ([self->resultRect_ empty]) {
        goto EXIT;
    }

    // ѥå°
    for (idx = 0; idx < COLOR_MAX; (idx)++) {
        mask[idx] = [[pal palette:idx] isMask];
        drop[idx] = [[pal palette:idx] noDropper];
    }

    // ѥѥ
    tile = [[(PoCoAppController *)([NSApp delegate]) tilePattern] pattern:[info tileNumber]];  

    // ̥쥤䡼
    behind = [[[[doc picture] layer:([[doc selLayer] sel] - 1)] bitmap] getBitmap:self->resultRect_];
    pat = [[PoCoColorPattern alloc] initWithBitmap:behind]; 
    [behind release];
    behind = nil;

    // Խϰ
    ep = [[PoCoPoint alloc] init];
    sp = [[PoCoPoint alloc] initX:[self->resultImage_ width]
                            initY:[self->resultImage_ height]];

    // ǻøȿư
    iter = [grad objectEnumerator];
    for (pair = [iter nextObject]; pair != nil; pair = [iter nextObject]) {
        // ȿưǤϴܿФƤϲ⤷ʤ
        for (idx = 0; idx < [base count]; (idx)++) {
            bnum = (NSNumber *)([base objectAtIndex:idx]);
            gnum = (NSNumber *)([[pair colors] objectAtIndex:0]);
            if ([bnum intValue] == [gnum intValue]) {
                bpair = pair;
                goto NEXT;
            }
        }

        // ǻø¦ǾɤػߤʤΤǲ⤷ʤ
        if (mask[[pair base]]) {
            goto NEXT;
        }

        // ѥå°ڤؤ
        for (idx = 0; idx < COLOR_MAX; (idx)++) {
            [[pal palette:idx] setMask:YES];
            [[pal palette:idx] setDropper:YES];
        }
        for (idx = 0; idx < [base count]; (idx)++) {
            bnum = (NSNumber *)([base objectAtIndex:idx]);
            gnum = (NSNumber *)([[pair colors] objectAtIndex:idx]);
            if (drop[[bnum intValue]]) {
                // ƥ¦ǵۤػ
                ;
            } else if (drop[[gnum intValue]]) {
                // ִǵۤػ
                ;
            } else {
                // оݤοǽȤ
                [[pal palette:[bnum intValue]] setMask:NO];
                [[pal palette:[bnum intValue]] setDropper:NO];
                [[pal palette:[gnum intValue]] setMask:NO];
                [[pal palette:[gnum intValue]] setDropper:NO];
            }
        }
        [[pal palette:[pair base]] setMask:NO];
        [[pal palette:[pair base]] setDropper:NO];

        // ̥쥤䡼Ф
        [self->resultImage_ release];   // ̤˺
        self->resultImage_ = [self->originalImage_ copy];
        fill = [[PoCoEditRegionFill alloc]
                   initWithPattern:self->resultImage_
                           palette:pal
                              tile:tile
                           pattern:pat
                         checkDist:YES];
        [fill executeDraw:self->resultShape_
             withStartPos:sp
               withEndPos:ep];
        [fill release];
        [self copyResultToOriginal];    // Ф̤ʣ
        [self->originalImage_ release];
        self->originalImage_ = [self->resultImage_ copy];

        // ǻøȿ
        for (idx = 0; idx < COLOR_MAX; (idx)++) {
            mtx[idx] = idx;
        }
        for (idx = 0; idx < [base count]; (idx)++) {
            bnum = (NSNumber *)([base objectAtIndex:idx]);
            gnum = (NSNumber *)([[pair colors] objectAtIndex:idx]);
            mtx[[bnum intValue]] = [gnum intValue];
        }
        [self->resultImage_ release];   // ̤˺
        self->resultImage_ = [self->originalImage_ copy];
        rep = [[PoCoEditBitmapColorReplace alloc]
                   initTargetBitmap:self->resultImage_
                   withSourceBitmap:self->originalImage_
                           withMask:self->originalShape_
                         withMatrix:mtx];
        [rep execute];
        [rep release];
        [self copyResultToOriginal];    // ǻøη̤ʣ
        [self->originalImage_ release];
        self->originalImage_ = [self->resultImage_ copy];

        // 
NEXT:
        ;
    }

    // Ǹ˴ܿȿ
    if ((bpair == nil) ||
        (mask[[bpair base]])) {
        // ܿ뤤Ͼɤػ
        ;
    } else {
        for (idx = 0; idx < COLOR_MAX; (idx)++) {
            [[pal palette:idx] setMask:YES];
            [[pal palette:idx] setDropper:YES];
        }
        iter = [[bpair colors] objectEnumerator];
        for (gnum = [iter nextObject]; gnum != nil; gnum = [iter nextObject]) {
            if (drop[[gnum intValue]]) {
                // ۤػ
                ;
            } else {
                // оݤοǽȤ
                [[pal palette:[gnum intValue]] setMask:NO];
                [[pal palette:[gnum intValue]] setDropper:NO];
            }
        }
        [[pal palette:[bpair base]] setMask:NO];
        [[pal palette:[bpair base]] setDropper:NO];

        // ̥쥤䡼Ф
        [self->resultImage_ release];   // ̤˺
        self->resultImage_ = [self->originalImage_ copy];
        fill = [[PoCoEditRegionFill alloc]
                   initWithPattern:self->resultImage_
                           palette:pal
                              tile:tile
                           pattern:pat
                         checkDist:YES];
        [fill executeDraw:self->resultShape_
             withStartPos:sp
               withEndPos:ep];
        [fill release];
        [self copyResultToOriginal];    // Ф̤ʣ
        [self->originalImage_ release];
        self->originalImage_ = [self->resultImage_ copy];
    }

    // ѥå°ʽ
    for (idx = 0; idx < COLOR_MAX; (idx)++) {
        [[pal palette:idx] setMask:mask[idx]];
        [[pal palette:idx] setDropper:drop[idx]];
    }

EXIT:
    [pat release];
    [ep release];
    [sp release];
    return;
}


// ----------------------------------------------- instance - public - ɤ
//
// 
//
//  Call
//    p      : о(seed)
//    bmp    : оݲ
//    border : ꤫
//             YES : ˤɤ
//             NO  : ɸοˤɤ
//    range  : ϰ
//
//  Return
//    originalRect_     : Ĺ(ѷ)(instance ѿ)
//    originalHandle_[] : ϥɥ(ѷ)(instance ѿ)
//    originalShape_    : (ѷ)(instance ѿ)
//
-(void)seedNew:(PoCoPoint *)p
        bitmap:(PoCoBitmap *)bmp
      isBorder:(BOOL)border
    colorRange:(int)range
{
    PoCoCreateStyleMask *mask;
    PoCoRect *tr;

    // ΤȤ
    tr = [[PoCoRect alloc] initLeft:0
                            initTop:0
                          initRight:[bmp width]
                         initBottom:[bmp height]];

    // ޥ
    mask = [[PoCoCreateStyleMask alloc] init];
    if ([mask paint:p
        imageBitmap:bmp
          imageRect:tr
           isBorder:border
         colorRange:range
         resultRect:self->originalRect_]) {
        [self->originalShape_ release];
        self->originalShape_ = [[mask mask] copy];
    }
    [mask release];

    // ϥɥ
    [PoCoSelectionShape createHandleRect:self->originalRect_
                           targetHandles:(PoCoRect **)(&(self->originalHandle_))];

    // ΤȤ
    [tr release];

    // ѷηѷηʣ
    [self copyOriginalToResult];

    return;
}


//
// 
//
//  Call
//    p              : о(seed)
//    bmp            : оݲ
//    border         : ꤫
//                     YES : ˤɤ
//                     NO  : ɸοˤɤ
//    range          : ϰ
//    originalRect_  : Ĺ(ѷ)(instance ѿ)
//    originalShape_ : (ѷ)(instance ѿ)
//    resultRect_    : Ĺ(ѷ)(instance ѿ)
//
//  Return
//    originalRect_     : Ĺ(ѷ)(instance ѿ)
//    originalHandle_[] : ϥɥ(ѷ)(instance ѿ)
//    originalShape_    : (ѷ)(instance ѿ)
//
-(void)seedJoin:(PoCoPoint *)p
         bitmap:(PoCoBitmap *)bmp
       isBorder:(BOOL)border
     colorRange:(int)range
{
    PoCoCreateStyleMask *mask;
    PoCoRect *tr;

    // ΤȤ
    tr = [[PoCoRect alloc] initLeft:0
                            initTop:0
                          initRight:[bmp width]
                         initBottom:[bmp height]];

    // ޥ
    mask = [[PoCoCreateStyleMask alloc] init];
    if ([mask paintJoin:p
            imageBitmap:bmp
              imageRect:tr
            styleBitmap:self->originalShape_
              styleRect:self->resultRect_
               isBorder:border
             colorRange:range
             resultRect:self->originalRect_]) {
        [self->originalShape_ release];
        self->originalShape_ = [[mask mask] copy];
    }
    [mask release];

    // ϥɥ
    [PoCoSelectionShape createHandleRect:self->originalRect_
                           targetHandles:(PoCoRect **)(&(self->originalHandle_))];

    // ΤȤ
    [tr release];

    // ѷηѷηʣ
    [self copyOriginalToResult];

    return;
}


//
// ʬΥ
//
//  Call
//    p              : о(seed)
//    bmp            : оݲ
//    border         : ꤫
//                     YES : ˤɤ
//                     NO  : ɸοˤɤ
//    range          : ϰ
//    originalRect_  : Ĺ(ѷ)(instance ѿ)
//    originalShape_ : (ѷ)(instance ѿ)
//    resultRect_    : Ĺ(ѷ)(instance ѿ)
//
//  Return
//    originalRect_     : Ĺ(ѷ)(instance ѿ)
//    originalHandle_[] : ϥɥ(ѷ)(instance ѿ)
//    originalShape_    : (ѷ)(instance ѿ)
//
-(void)seedSeparate:(PoCoPoint *)p
             bitmap:(PoCoBitmap *)bmp
           isBorder:(BOOL)border
         colorRange:(int)range
{
    PoCoCreateStyleMask *mask;
    PoCoRect *tr;

    // ϰʤ¹
    if ([PoCoSelectionShape hitTest:p
                           withRect:self->originalRect_
                          withShape:self->originalShape_] == PoCoHandleType_in_shape) {
        // ΤȤ
        tr = [[PoCoRect alloc] initLeft:0
                                initTop:0
                              initRight:[bmp width]
                             initBottom:[bmp height]];

        // ޥ
        mask = [[PoCoCreateStyleMask alloc] init];
        if ([mask paintSeparate:p
                    imageBitmap:bmp
                      imageRect:tr
                    styleBitmap:self->originalShape_
                      styleRect:self->resultRect_
                       isBorder:border
                     colorRange:range
                     resultRect:self->originalRect_]) {
            [self->originalShape_ release];
            self->originalShape_ = [[mask mask] copy];
        }
        [mask release];

        // ϥɥ
        [PoCoSelectionShape createHandleRect:self->originalRect_
                               targetHandles:(PoCoRect **)(&(self->originalHandle_))];

        // ΤȤ
        [tr release];

        // ѷηѷηʣ
        [self copyOriginalToResult];
    }

    return;
}

@end
