//
//	Pelistina on Cocoa - PoCo -
//	layer 
//
//	Copyright (C) 2005-2016 KAENRYUU Koutoku.
//

#import "PoCoLayer.h"

#import "PoCoAppController.h"
#import "PoCoEditInfo.h"

#import "PoCoPalette.h"
#import "PoCoPNG.h"
#import "PoCoZlib.h"

// 
#if !(MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_5)
static const unsigned int SAMPLE_SIZE = 48;
static const unsigned int SAMPLE_SIZE_4X = (48 * 4);
#endif  // !(MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_5)

// ============================================================================
@implementation PoCoBitmap

// --------------------------------------------------------- instance - private
//
// ĥ
//
//  Call
//    buffer : ĥ
//    png    : PNG Decoder
//    size_  :  (instance ѿ)
//
//  Return
//    function : YES : 
//               NO  : 
//    buffer   : ĥ
//
-(BOOL)inflateChunkData:(unsigned char *)buffer
         withPNGDecoder:(PoCoPNGDecoder *)png
{
    BOOL result;
    unsigned int idx;                   // DI-reg.
    PoCoWErr rv;
    PoCoZlibInflate *zlib;

    DPRINT((@"[PoCoBitmap inflateChunkData]\n"));

    result = NO;
    idx = 0;
    zlib = nil;

    // ĥ
    zlib = [[PoCoZlibInflate alloc] init];
    if (zlib == nil) {
        goto EXIT;
    }
    rv = [zlib appendBytes:[png bytes]
                    length:[png length]];
    if (rv < ER_OK) {
        DPRINT((@"[zlib appendBytes: length:] : %d(0x%08x)\n", rv, rv));
        goto EXIT;
    } else if (self->size_ < [zlib length]) {
        DPRINT((@"data overflow\n"));
        goto EXIT;
    }
    [png readChunk:NULL
            length:[png length]
            readed:NULL];
    memcpy(&(buffer[idx]), [zlib bytes], [zlib length]);
    idx += [zlib length];
    [zlib clearBuffer:[zlib length]];
    while (rv == 1) {
        rv = [zlib appendBytes:NULL
                        length:0];
        if (rv < ER_OK) {
            DPRINT((@"[zlib appendBytes: length:] : %d(0x%08x)\n", rv, rv));
            goto EXIT;
        } else if (self->size_ < (idx + [zlib length])) {
            DPRINT((@"data overflow\n"));
            goto EXIT;
        }
        memcpy(&(buffer[idx]), [zlib bytes], [zlib length]);
        idx += [zlib length];
        [zlib clearBuffer:[zlib length]];
    }
    do {
        rv = [zlib finishData];
        if (rv < ER_OK) {
            DPRINT((@"[zlib finishData] : %d(0x%08x)\n", rv, rv));
            goto EXIT;
        } else if (self->size_ < (idx + [zlib length])) {
            DPRINT((@"data overflow\n"));
            goto EXIT;
        }
        memcpy(&(buffer[idx]), [zlib bytes], [zlib length]);
        idx += [zlib length];
        [zlib clearBuffer:[zlib length]];
    } while (rv == 1);

    // ｪλ
    result = YES;

EXIT:
    [zlib release];
    return result;
}


// ---------------------------------------------------------- instance - public
//
// initialize
//
//  Call
//    None
//
//  Return
//    function : 
//
-(id)init
{
    return [self initWidth:0
                initHeight:0
              defaultColor:0];
}


//
// initialize(ꥤ˥饤)
//  ΰĤ
//
//  Call
//    w : 
//    h : ⤵
//    c : 
//
//  Return
//    function : 
//    width_   : (instance ѿ)
//    height_  : ⤵(instance ѿ)
//    size_    : (instance ѿ)
//    bitmap_  : pixel map(instance ѿ)
//
-(id)initWidth:(int)w
    initHeight:(int)h
  defaultColor:(unsigned char)c
{
//    DPRINT((@"[PoCoBitmap initWidth:%d initHeight:%d defaultColor:%d]\n", w, h, c));

    // super class ν
    self = [super init];

    // Ȥν
    if (self != nil) {
        self->width_ = w;
        self->height_ = h;
        self->size_ = ((w + (w & 1)) * h);  // ˶
        self->bitmap_ = NULL;
        if (self->size_ > 0) {
            self->bitmap_ = (unsigned char *)(malloc(self->size_));
            if (self->bitmap_ == NULL) {
                DPRINT((@"memory allocation error.\n"));
                [self release];
                self = nil;
                goto EXIT;
            }

            // οɤĤ֤Ƥ
            memset(self->bitmap_, c, self->size_);
        }
    }

EXIT:
    return self;
}


//
// initialize(ʣ)
//
//  Call
//    zone    : zone(api ѿ)
//    width_  : (instance ѿ)
//    height_ : ⤵(instance ѿ)
//    bitmap_ : pixel map(instance ѿ)
//
//  Return
//    function : (ʣ)
//
-(id)copyWithZone:(NSZone *)zone
{
    PoCoBitmap *copy;

//    DPRINT((@"[PoCoBitmap copyWithZone:0x%x]\n", (unsigned int)(zone)));

    // ʣ
    copy = [[[self class] allocWithZone:zone] initWidth:self->width_
                                             initHeight:self->height_
                                           defaultColor:0];

    // ʣƤ
    if (copy != nil) {
        memcpy(copy->bitmap_, self->bitmap_, copy->size_);
    }

    return copy;
}


//
// deallocate
//
//  Call
//    None
//
//  Return
//    width_  : (instance ѿ)
//    height_ : ⤵(instance ѿ)
//    size_   : (instance ѿ)
//    bitmap_ : pixel map(instance ѿ)
//
-(void)dealloc
{
//    DPRINT((@"[PoCoBitmap dealloc]\n"));

    if (self->bitmap_ != NULL) {
        free(self->bitmap_);
    }
    self->width_ = 0;
    self->height_ = 0;
    self->size_ = 0;
    self->bitmap_ = NULL;

    // super class 
    [super dealloc];

    return;
}


//
// μ
//
//  Call
//    width_ : (instance ѿ)
//
//  Return
//    function : 
//
-(int)width
{
    return self->width_;
}


//
// ⤵μ
//
//  Call
//    height_ : ⤵(instance ѿ)
//
//  Return
//    function : ⤵
//
-(int)height
{
    return self->height_;
}


//
// bitmap λȼ
//  줿 pointer  free() ʤ
//
//  Call
//    None
//
//  Return
//    function : pixel map
//
-(unsigned char *)pixelmap
{
    return self->bitmap_;
}


//
// bitmap μ
//  
//
//  Call
//    size_   : (instance ѿ)
//    bitmap_ : pixel map(instance ѿ)
//
//  Return
//    function : pixel map
//
-(unsigned char *)getBitmap
{
    unsigned char *ptr;

    ptr = NULL;

    // ΰ
    ptr = (unsigned char *)(malloc(self->size_));
    if (ptr == NULL) {
        DPRINT((@"memory allocation error.\n"));
        ;
    } else {
        // ñʣ
        memcpy(ptr, self->bitmap_, self->size_);
    }

    return ptr;
}


//
// 
//  ʬ
//
//  Call
//    r       : ϰ
//    width_  : (instance ѿ)
//    height_ : ⤵(instance ѿ)
//    bitmap_ : pixel map(instance ѿ)
//
//  Return
//    function : != nil : 
//               == nil : ʤ
//
-(PoCoBitmap *)getBitmap:(const PoCoRect *)r
{
    PoCoBitmap *bmp;
    int x;
    int y;
    int xe;
    int ye;
    unsigned char *sp;                  // SI-reg.(bitmap_ )
    unsigned char *dp;                  // DI-reg.(bmp )
    const size_t sstep = (self->width_ + (self->width_ & 1)); // SI-reg. ι
    size_t dstep;                       // DI-reg. ι(byte ñ)
    const int rl = [r left];
    const int rt = [r top];
    const int rr = [r right];
    const int rb = [r bottom];

    bmp = nil;

    // ϰϤλ
    if ((rl >= self->width_) || (rt >= self->height_)) {
        // lefttop ϰϳ
        ;
    } else if ((rr < 0) || (rb < 0)) {
        // rightbot ϰϳ
        ;
    } else {
        // λ
        x = 0;
        y = 0;
        sp = self->bitmap_;
        if (rl > 0) {
            x = rl;
            sp += rl;
        }
        if (rt > 0) {
            y = rt;
            sp += (sstep * y);
        }

        // žüλ
        xe = ((rr > self->width_)  ? self->width_  : rr);
        ye = ((rb > self->height_) ? self->height_ : rb);

        // PoCoBitmap 
        bmp = [[PoCoBitmap alloc] initWidth:(xe - x)
                                 initHeight:(ye - y)
                               defaultColor:0];
        if (bmp != nil) {
            // bmp->bitmap_ Ƥ
            ye -= y;
            dp = bmp->bitmap_;
            dstep = (bmp->width_ + (bmp->width_ & 1));
            do {
                memcpy(dp, sp, bmp->width_);

                //  index-reg ι
                sp += sstep;
                dp += dstep;
                (ye)--;
            } while (ye != 0);
        }
    }

    return bmp;
}


//
// 
//  ΰ褬Ϥ߽ФƤƤĥȤϤʤ clip 
//
//  Call
//    bmp     : 
//    r       : bmp ΰ(lefttopʣ踶)
//    width_  : (instance ѿ)
//    height_ : ⤵(instance ѿ)
//
//  Return
//    bitmap_ : pixel map(instance ѿ)
//
-(void)setBitmap:(const unsigned char *)bmp
        withRect:(const PoCoRect *)r
{
    int y;                              // ž
    const unsigned char *sp;            // SI-reg.(bmp ¦)
    unsigned char *dp;                  // DI-reg.(bitmap_ ¦)
    size_t csize;                       // 1Ԥʣ(byte ñ)
    size_t sstep;                       // SI-reg. ι(byte ñ)
    const size_t dstep = (self->width_ + (self->width_ & 1)); // DI-reg. ι
    const int rx = [r left];
    const int ry = [r top];
    const int rxe = [r right];
    const int rye = [r bottom];
    const int rw = [r width];

    if ((rx >= self->width_) || (ry >= self->height_)) {
        // lefttop Ǥϰϳ
        ;
    } else if ((rxe < 0) || (rye < 0)) {
        // rightbot ϰϳ
        ;
    } else {
        // ̤λ
        sstep = (rw + (rw & 1));

        // ʣ̸λ
        sp = bmp;
        dp = self->bitmap_;
        if (rx < 0) {
            sp -= rx;
        } else if (rx > 0) {
            dp += rx;
        }
        if (ry < 0) {
            sp += (sstep * -(ry));
        } else if (ry > 0) {
            dp += (dstep * ry);
        }

        // žλ
        csize = rw;
        if (rx < 0) {
            csize += rx;
        } else if (csize > (self->width_ - rx)) {
            csize = (self->width_ - rx);
        }

        // žϰϤλ
        y = ((rye > self->height_) ? self->height_ : rye);
        if (ry > 0) {
            y -= ry;
        }

        // ñ̤ʣ
        do {
            memcpy(dp, sp, csize);

            //  index-reg ι
            sp += sstep;
            dp += dstep;
            (y)--;
        } while (y != 0);
    }

    return;
}


//
// ѷ
//  resize ˤȤʤ bitmap  clip ȤϤʤ
//
//  Call
//    w     : (dot ñ)
//    h     : ⤵(dot ñ)
//    size_ : (instance ѿ)
//
//  Return
//    function : == YES : ѷ(bitmap ϺƳݺѤ)
//               == NO  : ѷ(bitmap ϰΤޤ)
//    width_   : (instance ѿ)
//    height_  : ⤵(instance ѿ)
//    size_    : (instance ѿ)
//    bitmap_  : pixel map(instance ѿ)
//
-(BOOL)resizeWidth:(int)w
      resizeHeight:(int)h
{
    BOOL flg;
    int s;
    unsigned char *buf;

    flg = YES;                          // ͤ
    buf = NULL;

    // ɬ̤λ
    if (w <= 0) {
        w = self->width_;
    }
    if (h <= 0) {
        h = self->height_;
    }
    s = ((w + (w & 1)) * h);

    // bitmap ѷ
    if (self->size_ < s) {
        buf = (unsigned char *)(realloc(self->bitmap_, s));
        if (buf == NULL) {
           DPRINT((@"memory allocation error.\n"));
           flg = NO;
        } else {
            // bitmap ι
            self->bitmap_ = buf;
            self->size_ = s;
        }
    }

    // 礭ι
    if (flg) {
        self->width_ = w;
        self->height_ = h;
    }

    return flg;
}


// ----------------------------------------- instance - public - ե
//
// ¸
//
//  Call
//    width_  : (instance ѿ)
//    height_ : ⤵(instance ѿ)
//    size_   : (instance ѿ)
//    bitmap_ : pixel map(instance ѿ)
//
//  Return
//    function : data
//
-(NSData *)createFileData
{
    NSMutableData *data;
    PoCoWErr rv;
    PoCoZlibDeflate *zlib;
#ifdef  __LITTLE_ENDIAN__
    int i;
#endif  // __LITTLE_ENDIAN__

    DPRINT((@"[PoCoBitmap createFileData]\n"));

    data = [NSMutableData data];

    // 񤭹
#ifdef  __LITTLE_ENDIAN__
    i = PoCoSwapUW(self->width_);
    [data appendBytes:&(i)
               length:sizeof(int)];
#else   // __LITTLE_ENDIAN__
    [data appendBytes:&(self->width_)
               length:sizeof(int)];
#endif  // __LITTLE_ENDIAN__

    // ⤵񤭹
#ifdef  __LITTLE_ENDIAN__
    i = PoCoSwapUW(self->height_);
    [data appendBytes:&(i)
               length:sizeof(int)];
#else   // __LITTLE_ENDIAN__
    [data appendBytes:&(self->height_)
               length:sizeof(int)];
#endif  // __LITTLE_ENDIAN__

    // 
    zlib = [[PoCoZlibDeflate alloc] init];
    rv = [zlib appendBytes:self->bitmap_ length:self->size_];
    if (rv < ER_OK) {
        DPRINT((@"[zlib appendBytes: length:] : %d(0x%08x)\n", rv, rv));
        data = nil;
    } else {
        [data appendBytes:[zlib bytes]
                   length:[zlib length]];
        [zlib clearBuffer:[zlib length]];
        do {
            rv = [zlib finishData];
            if (rv < ER_OK) {
                DPRINT((@"[zlib finishData] : %d(0x%08x)\n", rv, rv));
                data = nil;
                break;
            }
            [data appendBytes:[zlib bytes]
                       length:[zlib length]];
            [zlib clearBuffer:[zlib length]];
        } while (rv == 1);
    }
    [zlib release];

    return data;
}


//
// ɤ߹(̾)
//
//  Call
//    png : PNG Decoder
//
//  Return
//    function : 
//    width_   : (instance ѿ)
//    height_  : ⤵(instance ѿ)
//    bitmap_  : pixel map(instance ѿ)
//
-(id)initWithPNGDecoder:(PoCoPNGDecoder *)png
{
    int w;
    int h;

    DPRINT((@"[PoCoBitmap initWithPNGDecoder]\n"));

    [png readIntChunk:(unsigned int *)(&(w))];
    [png readIntChunk:(unsigned int *)(&(h))];

    // ꥤ˥饤ƤӽФ
    self = [self initWidth:w
                initHeight:h
              defaultColor:0];

    // 
    if (self != nil) {
        if ([self inflateChunkData:self->bitmap_
                    withPNGDecoder:png]) {
            // 
            ;
        } else {
            // 
            [self release];
            self = nil;
        }
    }

    return self;
}


//
// ɤ߹(IDAT CHUNK)
//  ʣ IDAT CHUNK ˤ̤б
//  PNG ؤδбʤΤǡIDAT CHUNK бϤޤܵǼʤ
//
//  Call
//    png : PNG Decoder
//    w   : 
//    h   : ⤵
//
//  Return
//    function : 
//    width_   : (instance ѿ)
//    height_  : ⤵(instance ѿ)
//    size_    : (instance ѿ)
//    bitmap_  : pixel map(instance ѿ)
//
-(id)initWithPNGDecoderIDAT:(PoCoPNGDecoder *)png
                  sizeWidth:(unsigned int)w
                 sizeHeight:(unsigned int)h
{
    unsigned int l;
    unsigned char *ptr;
    unsigned char *ptr2;

    DPRINT((@"[PoCoBitmap initWithPNGDecoderIDAT sizeWidth sizeHeigth]\n"));

    // (+1  filter )
    self = [self initWidth:(w + 1)
                initHeight:h
              defaultColor:0];
    if (self == nil) {
        goto EXIT;
    }

    // 
    if ([self inflateChunkData:self->bitmap_
                withPNGDecoder:png]) {
        // ե륿»
        ptr = self->bitmap_;
        [png initFilter:w];
        l = h;
        do {
            // 1ԤΥե륿»
            [png filtering:ptr[0]
                      data:&(ptr[1])];

            // 
            (l)--;
            ptr += self->width_;
        } while (l != 0);

        // ե륿ʬä
        ptr = self->bitmap_;
        l = h;
        if (w & 1) {
            do {
                memmove(&(ptr[0]), &(ptr[1]), w);
                ptr[w] = 0x00;

                // 
                (l)--;
                ptr += self->width_;
            } while (l != 0);
            self->width_ = w;
        } else {
            ptr2 = (self->bitmap_ + 1);
            do {
                memmove(ptr, ptr2, w);

                // 
                (l)--;
                ptr += w;
                ptr2 += self->width_;
            } while (l != 0);

            // Ƴ(ʤΤǡ­Ϥʤ...)
            self->width_ = w;
            self->size_ = (self->width_ * self->height_);
            ptr = (unsigned char *)(realloc(self->bitmap_, self->size_));
            if (ptr != NULL) {
                self->bitmap_ = ptr;
            } else {
                [self release];
                self = nil;
            }
        }
    } else {
        // 
        [self release];
        self = nil;
    }

EXIT:
    return self;
}

@end




// ============================================================================
@implementation PoCoLayerBase

// ---------------------------------------------------------- instance - public
//
// initialize
//
//  Call
//    None
//
//  Return
//    function : 
//
-(id)init
{
    return [self initWidth:0
                initHeight:0
              defaultColor:0];
}


//
// initialize(ꥤ˥饤)
//  쥤䡼ˤʤ
//
//  Call
//    w : (dot ñ)
//    h : ⤵(dot ñ)
//    c : 
//
//  Return
//    function    : 
//    bitmap_     : Ƥμ(instance ѿ)
//    preview_    : ܤɽ(instance ѿ)
//    sampleRect_ : ܤ礭(instance ѿ)
//    drawLock_   : ػߤ̵ͭ(instance ѿ)
//    isDisplay_  : ɽػߤ̵ͭ(instance ѿ)
//    name_       : 쥤䡼̾(instance ѿ)
//
-(id)initWidth:(int)w
    initHeight:(int)h
  defaultColor:(unsigned char)c
{
    DPRINT((@"[PoCoLayerBase initWidth:%d initHeight:%d defaultColor:%d]\n", w, h, c));

    // super class ν
    self = [super init];

    // Ȥν
    if (self != nil) {
        // ͤν
        self->bitmap_ = nil;
        self->preview_ = nil;
        self->sampleRect_ = NSMakeRect(0.0, 0.0, 0.0, 0.0);
        self->drawLock_ = NO;
        self->isDisplay_ = YES;
        self->name_ = [[NSString alloc] initWithString:@""];

        // ԽоݲΤ
        self->bitmap_ = [[PoCoBitmap alloc] initWidth:w
                                           initHeight:h
                                         defaultColor:c];
        if (self->bitmap_ == nil) {
            DPRINT((@"can't alloc bitmap.\n"));
            [self release];
            self = nil;
        }
    }

    return self;
}


//
// initialize(ʣ)
//
//  Call
//    zone        : zone(api ѿ)
//    bitmap_     : Ƥμ(instance ѿ)
//    preview_    : ܤɽ(instance ѿ)
//    sampleRect_ : ܤ礭(instance ѿ)
//    drawLock_   : ػߤ̵ͭ(instance ѿ)
//    isDisplay_  : ɽػߤ̵ͭ(instance ѿ)
//    name_       : 쥤䡼̾(instance ѿ)
//
//  Return
//    function : (ʣ)
//
-(id)copyWithZone:(NSZone *)zone
{
    PoCoLayerBase *copy;

    DPRINT((@"[PoCoLayerBase copyWithZone:0x%p]\n", zone));

    // ʣ
    copy = [[[self class] allocWithZone:zone] init];

    // ʣƤ
    if (copy != nil) {
        // Ѱդ줿Τ
        [copy->bitmap_ release];
        [copy->preview_ release];
        [copy->name_ release];
        copy->bitmap_ = nil;
        copy->preview_ = nil;
        copy->name_ = nil;

        // copy 
        copy->bitmap_ = [self->bitmap_ copy];
        copy->preview_ = [self->preview_ copy];
        copy->sampleRect_ = self->sampleRect_;
        if ([self drawLock]) {
            [copy setDrawLock];
        }
        if (!([self isDisplay])) {
            [copy clearIsDisplay];
        }
        copy->name_ = [self->name_ copy];
    }

    return copy;
}


//
// deallocate
//
//  Call
//    None
//
//  Return
//    bitmap_  : Ƥμ(instance ѿ)
//    preview_ : ܤɽ(instance ѿ)
//    name_    : 쥤䡼̾(instance ѿ)
//
-(void)dealloc
{
    DPRINT((@"[PoCoLayer dealloc]\n"));

    // ƻ񸻤
    [self->preview_ release];
    [self->bitmap_ release];
    [self->name_ release];
    self->bitmap_ = nil;
    self->preview_ = nil;
    self->name_ = nil;

    // super class 
    [super dealloc];

    return;
}


//
// 
//
//  Call
//    None
//
//  Return
//    function : ֹ
//
-(PoCoLayerType)layerType
{
    return PoCoLayerType_Base;
}


//
// ̾
//
//  Call
//    None
//
//  Return
//    function : ̾ʸ
//
-(NSString *)typeName
{
    return @"";
}


//
// bitmap λ
//
//  Call
//    bitmap_ : Ƥμ(instance ѿ)
//
//  Return
//    function : Խоݤ bitmap
//
-(PoCoBitmap *)bitmap
{
    return self->bitmap_;
}


//
// preview λ
//
//  Call
//    preview_    : ܤɽ(instance ѿ)
//    sampleRect_ : ܤ礭(instance ѿ)
//
//  Return
//    function : ɽѤ bitmap
//
-(NSImage *)preview
{
    NSImage *image;

#if (MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_5)
    image = [[NSImage alloc] initWithCGImage:[self->preview_ CGImage]
                                        size:self->sampleRect_.size];
#else   // (MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_5)
    image = [[NSImage alloc] init];
    [image addRepresentation:self->preview_];
#endif  // (MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_5)

    // 쥤䡼ξ identifier ǸƤФ줿ޤޤˤʤΤ autorelease
    [image autorelease];

    return image;
}


//
// ػߤγǧ
//
//  Call
//    drawLock_ : ػߤ̵ͭ(instance ѿ)
//
//  Return
//    function : YES : ػ߾
//               NO  : ľ
//
-(BOOL)drawLock
{
    return self->drawLock_;
}


//
// ɽػߤγǧ
//
//  Call
//    isDisplay_ : ɽػߤ̵ͭ(instance ѿ)
//
//  Return
//    function : YES : ɽľ
//               NO  : ɽػ߾
//
-(BOOL)isDisplay
{
    return self->isDisplay_;
}


//
// ̾Τμ
//
//  Call
//    name_ : 쥤䡼̾(instance ѿ)
//
//  Return
//    function : ̾
//
-(NSString *)name
{
    return self->name_;
}


//
// ػߤˤ
//
//  Call
//    None
//
//  Return
//    drawLock_ : ػߤ̵ͭ(instance ѿ)
//
-(void)setDrawLock
{
    self->drawLock_ = YES;

    return;
}


//
// ǽˤ
//
//  Call
//    None
//
//  Return
//    drawLock_ : ػߤ̵ͭ(instance ѿ)
//
-(void)clearDrawLock
{
    self->drawLock_ = NO;

    return;
}


//
// ɽĤˤ
//
//  Call
//    None
//
//  Return
//    isDisplay_ : ɽػߤ̵ͭ(instance ѿ)
//
-(void)setIsDisplay
{
    self->isDisplay_ = YES;

    return;
}


//
// ɽػߤˤ
//
//  Call
//    None
//
//  Return
//    isDisplay_ : ɽػߤ̵ͭ(instance ѿ)
//
-(void)clearIsDisplay
{
    self->isDisplay_ = NO;

    return;
}


//
// 쥤䡼̾
//
//  Call
//    str : ̾ʸ
//
//  Return
//    name_ : 쥤䡼̾(instance ѿ)
//
-(void)setName:(NSString *)str
{
    [self->name_ release];
    self->name_ = [str copy];

    return;
}


//
// preview ι
//
//  Call
//    palette : ѥå
//    bitmap_ : Խоݤβ(instance ѿ)
//
//  Return
//    preview_    : ܤɽ(instance ѿ)
//    sampleRect_ : ܤ礭(instance ѿ)
//
-(void)updatePreview:(PoCoPalette *)palette
#if (MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_5)
{
    unsigned int sx;
    unsigned int sy;
    unsigned int dy;
    unsigned int si;
    unsigned int di;
    unsigned int sourceXStep;
    unsigned int sourceYStep;
    unsigned int sourceRow;
    unsigned int sampleWidth;
    unsigned int sampleHeight;
    unsigned int sampleRow;
    unsigned char *source;
    unsigned char *sample;
    PoCoColor *color;
    NSBitmapImageRep *buffer;
    const unsigned int sourceWidth = [self->bitmap_ width];
    const unsigned int sourceHeight = [self->bitmap_ height];
    PoCoEditInfo *info;
    unsigned int size;

//    DPRINT((@"[PoCoLayerBase updatePreview]\n"));

    info = [(PoCoAppController *)([NSApp delegate]) editInfo];
    size = [info previewSize];

    // 礭λ(ɽʬ)
    if ((size == 0) ||
        ((sourceWidth <= size) && (sourceHeight <= size))) {
        self->sampleRect_.size.width = (CGFloat)(sourceWidth);
        self->sampleRect_.size.height = (CGFloat)(sourceHeight);
    } else {
        if (sourceWidth < sourceHeight) {
            self->sampleRect_.size.width = (CGFloat)(MAX(1, ((size * sourceWidth) / sourceHeight)));
            self->sampleRect_.size.height = (CGFloat)(size);
        } else {
            self->sampleRect_.size.width = (CGFloat)(size);
            self->sampleRect_.size.height = (CGFloat)(MAX(1, ((size * sourceHeight) / sourceWidth)));
        }
    }

    // 礭λ()
    size *= [info previewQuality];
    if ((size == 0) ||
        ((sourceWidth <= size) && (sourceHeight <= size))) {
        sampleWidth = sourceWidth;
        sampleHeight = sourceHeight;
        sourceXStep = 1;
        sourceYStep = 1;
    } else {
        if (sourceWidth < sourceHeight) {
            sampleWidth = MAX(1, ((size * sourceWidth) / sourceHeight));
            sampleHeight = size;
        } else {
            sampleWidth = size;
            sampleHeight = MAX(1, ((size * sourceHeight) / sourceWidth));
        }
        sourceXStep = (sourceWidth / sampleWidth);
        sourceYStep = (sourceHeight / sampleHeight);
    }

    // NSBitmapImageRep 
    buffer = [[NSBitmapImageRep alloc]
                 initWithBitmapDataPlanes:NULL
                               pixelsWide:sampleWidth
                               pixelsHigh:sampleHeight
                            bitsPerSample:8
                          samplesPerPixel:3
                                 hasAlpha:NO
                                 isPlanar:NO
                           colorSpaceName:NSCalibratedRGBColorSpace
                              bytesPerRow:0
                             bitsPerPixel:0];
    if (buffer == nil) {
        DPRINT((@"can't alloc NSBitmapImageRep\n"));
    } else {
        // ؤ
        [self->preview_ release];
        self->preview_ = buffer;

        // ܲ
        source = [self->bitmap_ pixelmap];
        sample = [self->preview_ bitmapData];
        sourceRow = (sourceWidth + (sourceWidth & 1));
        sampleRow = (unsigned int)([self->preview_ bytesPerRow]);
        for (sy = 0, dy = 0;
             (sy < sourceHeight) && (dy < sampleHeight);
             sy += sourceYStep, (dy)++) {
            si = (sy * sourceRow);
            di = (dy * sampleRow);
            for (sx = 0; sx < sourceWidth; sx += sourceXStep) {
                color = [palette palette:source[si]];
                sample[di + 0] = [color red];
                sample[di + 1] = [color green];
                sample[di + 2] = [color blue];
                si += sourceXStep;
                di += 3;
            }
        }
    }

    return;
}
#else   // (MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_5)
{
    unsigned int sx;
    unsigned int sy;
    unsigned int dy;
    unsigned int si;
    unsigned int di;
    unsigned int sourceXStep;
    unsigned int sourceYStep;
    unsigned int sourceRow;
    unsigned int sampleWidth;
    unsigned int sampleHeight;
    unsigned int sampleRow;
    unsigned char *source;
    unsigned char *sample;
    PoCoColor *color;
    NSBitmapImageRep *buffer;
    const unsigned int sourceWidth = [self->bitmap_ width];
    const unsigned int sourceHeight = [self->bitmap_ height];

//    DPRINT((@"[PoCoLayerBase updatePreview]\n"));

    // 礭λ
    if ((sourceWidth <= SAMPLE_SIZE) && (sourceHeight <= SAMPLE_SIZE)) {
        sampleWidth = sourceWidth;
        sampleHeight = sourceHeight;
        sourceXStep = 1;
        sourceYStep = 1;
    } else {
        if (sourceWidth < sourceHeight) {
            sampleWidth = MAX(1, ((SAMPLE_SIZE * sourceWidth) / sourceHeight));
            sampleHeight = SAMPLE_SIZE;
        } else {
            sampleWidth = SAMPLE_SIZE;
            sampleHeight = MAX(1, ((SAMPLE_SIZE * sourceHeight) / sourceWidth));
        }
        sourceXStep = (sourceWidth / sampleWidth);
        sourceYStep = (sourceHeight / sampleHeight);
    }

    // NSBitmapImageRep 
    buffer = [[NSBitmapImageRep alloc]
                 initWithBitmapDataPlanes:NULL
                               pixelsWide:sampleWidth
                               pixelsHigh:sampleHeight
                            bitsPerSample:8
                          samplesPerPixel:3
                                 hasAlpha:NO
                                 isPlanar:NO
                           colorSpaceName:NSCalibratedRGBColorSpace
                              bytesPerRow:0
                             bitsPerPixel:0];
    if (buffer == nil) {
        DPRINT((@"can't alloc NSBitmapImageRep\n"));
    } else {
        // ؤ
        [self->preview_ release];
        self->preview_ = buffer;

        // ܲ
        source = [self->bitmap_ pixelmap];
        sample = [self->preview_ bitmapData];
        sourceRow = (sourceWidth + (sourceWidth & 1));
        sampleRow = (unsigned int)([self->preview_ bytesPerRow]);
        for (sy = 0, dy = 0;
             (sy < sourceHeight) && (dy < sampleHeight);
             sy += sourceYStep, (dy)++) {
            si = (sy * sourceRow);
            di = (dy * sampleRow);
            for (sx = 0; sx < sourceWidth; sx += sourceXStep) {
                color = [palette palette:source[si]];
                sample[di + 0] = [color red];
                sample[di + 1] = [color green];
                sample[di + 2] = [color blue];
                si += sourceXStep;
                di += 3;
            }
        }
    }

    return;
}
#endif  // (MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_5)
#if 0   // (ŤΤǻȤʤ)
{
    unsigned int x;
    unsigned int y;
    unsigned int sskip;
    unsigned int dskip;
    unsigned char *sbmp;
    unsigned char *dbmp;
    PoCoColor *c;
    NSBitmapImageRep *buffer;
    const unsigned int w = [self->bitmap_ width];
    const unsigned int h = [self->bitmap_ height];

//    DPRINT((@"[PoCoLayerBase updatePreview]\n"));

    // 礭λ
    if ((w <= SAMPLE_SIZE) && (h <= SAMPLE_SIZE)) {
        self->sampleRect_.size.width = (CGFloat)(w);
        self->sampleRect_.size.height = (CGFloat)(h);
    } else {
        if (w < h) {
            self->sampleRect_.size.width = (CGFloat)(MAX(1, ((SAMPLE_SIZE * w) / h)));
            self->sampleRect_.size.height = (CGFloat)(SAMPLE_SIZE);
        } else {
            self->sampleRect_.size.width = (CGFloat)(SAMPLE_SIZE);
            self->sampleRect_.size.height = (CGFloat)(MAX(1, ((SAMPLE_SIZE * h) / w)));
        }
    }

    // NSBitmapImageRep 
    buffer = [[NSBitmapImageRep alloc]
                 initWithBitmapDataPlanes:NULL
                               pixelsWide:w
                               pixelsHigh:h
                            bitsPerSample:8
                          samplesPerPixel:3
                                 hasAlpha:NO
                                 isPlanar:NO
                           colorSpaceName:NSCalibratedRGBColorSpace
                              bytesPerRow:0
                             bitsPerPixel:0];
    if (buffer == nil) {
        DPRINT((@"can't alloc NSBitmapImageRep\n"));
    } else {
        // ؤ
        [self->preview_ release];
        self->preview_ = buffer;

        // ܲ
        sbmp = [self->bitmap_ pixelmap];
        dbmp = [self->preview_ bitmapData];
        sskip = (w & 1);
        dskip = ((unsigned int)([self->preview_ bytesPerRow]) - (w * 3));
        y = h;
        do {
            x = w;
            do {
                c = [palette palette:*(sbmp)];
                (sbmp)++;
                *(dbmp) = [c red];
                (dbmp)++;
                *(dbmp) = [c green];
                (dbmp)++;
                *(dbmp) = [c blue];
                (dbmp)++;
                (x)--;
            } while (x != 0);
            sbmp += sskip;
            dbmp += dskip;
            (y)--;
        } while (y != 0);
    }

    return;
}
#endif   // 0


// ----------------------------------------- instance - public - ե
//
// ¸
//
//  Call
//    None
//
//  Return
//    function : data
//
-(NSData *)createFileData
{
    //  class Ǥϲ⤷ʤ
    ;

    return nil;
}


//
// ɤ߹
//
//  Call
//    png : PNG Decoder
//
//  Return
//    function : 
//
-(id)initWithPNGDecoder:(PoCoPNGDecoder *)png
{
    //  class Ǥϲ⤷ʤ
    ;

    return nil;
}

@end




// ============================================================================
@implementation PoCoBitmapLayer

// ---------------------------------------------------------- instance - public
//
// initialize(ʣ)
//
//  Call
//    zone : zone(api ѿ)
//
//  Return
//    function : (ʣ)
//
-(id)copyWithZone:(NSZone *)zone
{
    DPRINT((@"[PoCoBitmapLayer copyWithZone:0x%p]\n", zone));

    // ʣ
    return [super copyWithZone:zone];
}


//
// 
//
//  Call
//    None
//
//  Return
//    function : ֹ
//
-(PoCoLayerType)layerType
{
    return PoCoLayerType_Bitmap;
}


//
// ̾
//
//  Call
//    None
//
//  Return
//    function : ̾ʸ
//
-(NSString *)typeName
{
    return [[NSBundle mainBundle]
               localizedStringForKey:@"BitmapLayerName"
                               value:@"image"
                               table:nil];
}


// ----------------------------------------- instance - public - ե
//
// ¸
//
//  Call
//    isDisplay_ : ɽػߤ̵ͭ(instance ѿ)
//    drawLock_  : ػߤ̵ͭ(instance ѿ)
//    name_      : 쥤䡼̾(instance ѿ)
//    bitmap_    : Խоݤβ(instance ѿ)
//
//  Return
//    function : data
//
-(NSData *)createFileData
{
    NSData *data;
    PoCoPNGEncoder *png;
    NSData *pixdata;

    DPRINT((@"[PoCoBitmapLayer createFileData]"));

    png = [[PoCoPNGEncoder alloc] init];

    // CHUNK 
    [png createChunk:(unsigned char *)("ilAY")];

    // ilAY Ƥ񤭹
    [png writeByteChunk:self->isDisplay_];
    [png writeByteChunk:self->drawLock_];
    if ([self->name_ length] > 0) {
        [png writeChunk:[self->name_ UTF8String]
                 length:((unsigned int)(strlen([self->name_ UTF8String])) + 1)];
    } else {
        [png writeByteChunk:0x00];
    }
    pixdata = [self->bitmap_ createFileData];
    [png writeChunk:[pixdata bytes]
             length:(unsigned int)([pixdata length])];

    // CHUNK Ĥ
    [png closeChunk];

    // NSData 
    data = [png createNSData];
    [png release];

    return data;
}


//
// ɤ߹(ɸ)
//
//  Call
//    png : PNG Decoder
//
//  Return
//    function : 
//
-(id)initWithPNGDecoder:(PoCoPNGDecoder *)png
{
    PoCoBitmap *bmp;
    unsigned char i;

    DPRINT((@"[PoCoBitmapLayer initWithPNGDecoder]"));

    // ꥤ˥饤ƤӽФ
    self = [super init];
    if (self == nil) {
        goto EXIT;
    }

    // ɽݤɤ߹
    if (!([png readByteChunk:&(i)])) {
        DPRINT((@"can't read display.\n"));
        [self release];
        self = nil;
        goto EXIT;
    }
    if (i == 0) {
        [self clearIsDisplay];
    }

    // Խݤɤ߹
    if (!([png readByteChunk:&(i)])) {
        DPRINT((@"can't read display.\n"));
        [self release];
        self = nil;
        goto EXIT;
    }
    if (i != 0) {
        [self setDrawLock];
    }

    // 쥤䡼̾Τɤ߹
    [self->name_ release];
    [self setName:[NSString stringWithUTF8String:[png bytes]]];
    [png readChunk:NULL
            length:((unsigned int)(strlen([png bytes])) + 1)
            readed:NULL];

    // pixelmap ɤ߹
    bmp = [[PoCoBitmap alloc] initWithPNGDecoder:png];
    if (bmp == nil) {
        DPRINT((@"can't load pixelmap.\n"));
        [self release];
        self = nil;
        goto EXIT;
    }
    [self->bitmap_ release];
    self->bitmap_ = bmp;
#if 0
{
    const unsigned char *p = [bmp pixelmap];
    const int w = [bmp width];
    const int h = [bmp height];
    const int r = (w + (w & 0x01));
    int x;
    int y;

    for (y = 0; y < h; (y)++) {
        for (x = 0; x < w; (x)++) {
            printf(" %02x,", p[(y * r) + x]);
        }
        printf("\n");
    }
}
#endif  // 0

EXIT:
    return self;
}


//
// ɤ߹(IDAT CHUNK)
//
//  Call
//    png : PNG Decoder
//    w   : 
//    h   : ⤵
//
//  Return
//    function : 
//
-(id)initWithPNGDecoderIDAT:(PoCoPNGDecoder *)png
                  sizeWidth:(unsigned int)w
                 sizeHeight:(unsigned int)h
{
    PoCoBitmap *bmp;

    DPRINT((@"[PoCoBitmapLayer initWithPNGDecoderIDAT sizeWidth sizeHeight]"));

    // ꥤ˥饤ƤӽФ
    self = [super init];

    // pixelmap ɤ߹
    if (self != nil) {
        bmp = [[PoCoBitmap alloc] initWithPNGDecoderIDAT:png
                                               sizeWidth:w
                                              sizeHeight:h];
        if (bmp == nil) {
            DPRINT((@"can't load pixelmap.\n"));
            [self release];
            self = nil;
        } else {
            [self->bitmap_ release];
            self->bitmap_ = bmp;
        }
    }

    return self;
}

@end




// ============================================================================
@implementation PoCoStringLayer

// ---------------------------------------------------------- instance - public
//
// initialize
//
//  Call
//    None
//
//  Return
//    function : 
//    str_r_   : ʸΥ쥤ΰ(instance ѿ)
//    str_     : оʸ(instance ѿ)
//
-(id)init
{
    // super class ν
    self = [super init];

    // Ȥν
    if (self != nil) {
        self->str_r_ = nil;
        self->str_ = @"";
    }

    return self;
}


//
// deallocate
//
//  Call
//    None
//
//  Return
//    str_r_ : ʸΥ쥤ΰ(instance ѿ)
//    str_   : оʸ(instance ѿ)
//
-(void)dealloc
{
    [self->str_r_ release];
    [self->str_ release];

    // super class β
    [super dealloc];

    return;
}


//
// 
//
//  Call
//    None
//
//  Return
//    function : ֹ
//
-(PoCoLayerType)layerType
{
    return PoCoLayerType_String;
}


//
// ̾
//
//  Call
//    None
//
//  Return
//    function : ̾ʸ
//
-(NSString *)typeName
{
    return [[NSBundle mainBundle]
               localizedStringForKey:@"StringLayerName"
                               value:@"string"
                               table:nil];
}

// ... ʸ layer ͭƤ̤ ...


// ----------------------------------------- instance - public - ե
//
// ¸
//
//  Call
//    None
//
//  Return
//    function : data
//
-(NSData *)createFileData
{
    DPRINT((@"[PoCoStringLayer createFileData]"));

// ... ޤ ...

    return nil;
}


//
// ɤ߹
//
//  Call
//    png : PNG Decoder
//
//  Return
//    function : 
//
-(id)initWithPNGDecoder:(PoCoPNGDecoder *)png
{
    DPRINT((@"[PoCoStringLayer initWithPNGDecoder]"));

// ... ޤ ...

    return nil;
}

@end
