//
//	Pelistina on Cocoa - PoCo -
//	MyDocument(Խоݤδμ紴)
//
//	Copyright (C) 2005-2018 KAENRYUU Koutoku.
//

#import "PoCoMyDocument.h"

#import "PoCoLayer.h"
#import "PoCoPicture.h"
#import "PoCoSelLayer.h"
#import "PoCoColorPattern.h"
#import "PoCoView.h"
#import "PoCoAppController.h"
#import "PoCoEditInfo.h"
#import "PoCoPNG.h"
#import "PoCoColorBuffer.h"


// ѿ
static const unsigned char header[8] = {  // PNG signature
                               0x89, 0x50, 0x4e, 0x47,
                               0x0d, 0x0a, 0x1a, 0x0a
                           };


// ѿ
#if (MAC_OS_X_VERSION_MAX_ALLOWED > MAC_OS_X_VERSION_10_3)
extern NSString* NSPNGPboardType;
#else   // (MAC_OS_X_VERSION_MAX_ALLOWED > MAC_OS_X_VERSION_10_3)
static NSString* NSPNGPboardType = @"Apple PNG pasteboard type";
#endif  // (MAC_OS_X_VERSION_MAX_ALLOWED > MAC_OS_X_VERSION_10_3)

// ============================================================================
@implementation MyDocument

// ------------------------------------------------------------ class - private
//
// PNG إåǧ
//
//  Call
//    data  : ɤ߹
//    range : ɤ߹ϰ
//
//  Return
//    function : YES : 
//               NO  : ۾
//    range    : ɤ߹ϰ(location 򹹿)
//
+(BOOL)checkFileHeader:(NSData *)data
             withRange:(NSRange *)range
{
    BOOL result;
    unsigned char buf[8];

    DPRINT((@"[MyDocument checkFileHeader]\n"));

    // Ƭ 8 bytes ɤ߹
    range->length = 8;
    [data getBytes:buf range:*(range)];
    range->location += range->length;

    // ȹ
    result = (memcmp(buf, header, 8) == 0);

    return result;
}


// ------------------------------------------------------------- class - public
//
// ڡȥܡɤŽդ¹
//
//  Call
//    pb  : ڡȥܡ
//    lyr : ƤФ쥤䡼
//    plt : ƤФѥå
//
//  Return
//    function : YES : 
//               NO  : ۾
//    lyr      : ƤФ쥤䡼
//    plt      : ƤФѥå
//
+(BOOL)readPngFromPasteboard:(NSPasteboard *)pb
                 resultLayer:(PoCoLayerBase **)lyr
               resultPalette:(PoCoPalette *)plt
{
    BOOL result;
    NSString *type;
    NSData *data;
    NSRange range;

    DPRINT((@"[PoCoDocument readPngFromPasteboard]\n"));

    result = NO;                        // ͤϼ
    range.location = 0;
    range.length = 0;

    // ̳ǧ(PNG Τ߼)
    type = [pb availableTypeFromArray:[NSArray arrayWithObject:NSPNGPboardType]];
    if (type == nil) {
        DPRINT((@"not exist PNG data.\n"));
        goto EXIT;
    }

    // Ƽ
    data = [pb dataForType:NSPNGPboardType];
    if (data == nil) {
        DPRINT((@"can't get data from PasteBoard.\n"));
        goto EXIT;
    }

    // إåǧ
    result = [MyDocument checkFileHeader:data
                               withRange:&(range)];
    if (result != YES) {
        DPRINT((@"not match header.\n"));
        goto EXIT;
    }

    // ɤ߹
    result = [PoCoPicture pastePBoardData:data
                                withRange:&(range)
                              resultLayer:lyr
                            resultPalette:plt];
    if (result != YES) {
        DPRINT((@"not paste data.\n"));
        goto EXIT;
    }

    // ｪλ
    result = YES;

EXIT:
    return result;
}


// --------------------------------------------------------- instance - private
//
// եإå
//  PNG signature Ǽ NSData ֤
//
//  Call
//    None
//
//  Return
//    function : data
//
-(NSData *)createFileHeader
{
    DPRINT((@"[MyDocument createFileHeader]\n"));

    return [NSData dataWithBytes:header length:8];
}


//
// եեå
//  IEND Chunk Ǽ NSData ֤
//
//  Call
//    None
//
//  Return
//    function : data
//
-(NSData *)createFileFooter
{
    NSData *data;
    PoCoPNGEncoder *png;

    DPRINT((@"[MyDocument createFileFooter]\n"));

    png = [[PoCoPNGEncoder alloc] init];

    // CHUNK 
    [png createChunk:(unsigned char *)("IEND")];

    // CHUNK Ĥ
    [png closeChunk];

    // NSData 
    data = [png createNSData];
    [png release];

    return data;
}


// ---------------------------------------------------------- instance - public
//
// initialize
//
//  Call
//    None
//
//  Return
//    function : 
//
-(id)init
{
    // ꥤ˥饤ز
    return [self initWidth:640
                initHeight:480
            initResolution:120
              defaultColor:0];
}


//
// initialize(ꥤ˥饤)
//  ˸Ƥ
//
//  Call
//    w : (dot ñ)
//    h : ⤵(dot ñ)
//    r : (dpi)
//    c : 
//
//  Return
//    function     : 
//    picture_     : Խоݲ(instance ѿ)
//    selLayer_    : 쥤䡼(instance ѿ)
//    eraser_      : äѲ(instance ѿ)
//    colorBuffer_ : ݻ(instance ѿ)
//
-(id)initWidth:(int)w
    initHeight:(int)h
initResolution:(int)r
  defaultColor:(unsigned char)c
{
    DPRINT((@"[PoCoMyDocument initWidth initHeight initResolution defaultColor]\n"));

    PoCoErr er;

    // super class 
    self = [super init];

    // Ȥν
    if (self != nil) {
        self->picture_ = nil;
        self->selLayer_ = nil;
        self->eraser_ = nil;
        self->colorBuffer_ = nil;

        // 񸻤γ
        self->picture_ = [[PoCoPicture alloc] init:r];
        if (self->picture_ == nil) {
            DPRINT((@"can't alloc PoCoPicture\n"));
            [self release];
            self = nil;
            goto EXIT;
        }
        self->selLayer_ = [[PoCoSelLayer alloc] init];
        if (self->selLayer_ == nil) {
            DPRINT((@"can't alloc PoCoSelLayer\n"));
            [self release];
            self = nil;
            goto EXIT;
        }
        if ([[(PoCoAppController *)([NSApp delegate]) editInfo] enableColorBuffer]) {
            self->colorBuffer_ = [[PoCoColorBuffer alloc] init];
            if (self->colorBuffer_ == nil) {
                DPRINT((@"can't alloc PoCoColorBuffer\n"));
                [self release];
                self = nil;
                goto EXIT;
            }
        }

        // 쥤䡼ɲ
        er = [self->picture_ addLayer:-1
                                width:w
                               height:h
                                color:c];
        if (er < ER_OK) {
            DPRINT((@"[self->picture_ addLayer:width:height:color:]: %d\n", er));
            [self release];
            self = nil;
            goto EXIT;
        }

        // ä
        [[self undoManager] setLevelsOfUndo:[[(PoCoAppController *)([NSApp delegate]) editInfo] undoMaxLevel]];
    }

EXIT:
    return self;
}


//
// deallocate
//
//  Call
//    None
//
//  Return
//    picture_     : Խоݲ(instance ѿ)
//    selLayer_    : 쥤䡼(instance ѿ)
//    eraser_      : äѲ(instance ѿ)
//    colorBuffer_ : ݻ(instance ѿ)
//
-(void)dealloc
{
    DPRINT((@"[PoCoMyDocument dealloc]\n"));

    // 񸻤β
    [self->eraser_ release];
    [self->selLayer_ release];
    [self->picture_ release];
    [self->colorBuffer_ release];
    self->picture_ = nil;
    self->selLayer_ = nil;
    self->eraser_ = nil;
    self->colorBuffer_ = nil;

    // super class β
    [super dealloc];

    return;
}


//
// Խоݲμ
//
//  Call
//    picture_ : Խоݲ(instance ѿ)
//
//  Return
//    function : Խоݲ
//
-(PoCoPicture *)picture
{
    return self->picture_;
}


//
// 쥤䡼μ
//
//  Call
//    selLayer_ : 쥤䡼(instance ѿ)
//
//  Return
//    function : 쥤䡼
//
-(PoCoSelLayer *)selLayer
{
    return self->selLayer_;
}


//
// ݻ
//
//  Call
//    colorBuffer_ : ݻ(instance ѿ)
//
//  Return
//    function : ݻ
//
-(PoCoColorBuffer *)colorBuffer
{
    return self->colorBuffer_;
}


//
// view 
//
//  Call
//    view_ : ɽ(outlet)
//
//  Return
//    function : view
//
-(PoCoView *)view
{
    return self->view_;
}


//
// äѲ
//
//  Call
//    picture_  : Խоݲ(instance ѿ)
//    selLayer_ : 쥤䡼(instance ѿ)
//
//  Return
//    eraser_ : äѲ(instance ѿ)
//
-(void)setEraser
{
    // ʬ
    [self clearEraser];

    // ߤΥ쥤䡼򤽤Τޤ޵
    self->eraser_ = [[PoCoColorPattern alloc]
                       initWithBitmap:[[self->picture_ layer:[self->selLayer_ sel]] bitmap]];

    return;
}


//
// äѲ
//
//  Call
//    None
//
//  Return
//    eraser_ : äѲ(instance ѿ)
//
-(void)clearEraser
{
    [self->eraser_ release];
    self->eraser_ = nil;

    return;
}


//
// äѲ
//
//  Call
//    eraser_ : äѲ(instance ѿ)
//
//  Return
//    function : äѲ
//
-(PoCoColorPattern *)eraser
{
    return self->eraser_;
}


//
// ɥڤؤ
//
//  Call
//    note     : (api ѿ)
//    view_    : ɽ(outlet)
//    picture_ : Խоݲ(instance ѿ)
//
//  Return
//    None
//
-(void)windowDidBecomeMain:(NSNotification *)note
{
    PoCoRect *r;

    // ƥɥɽڤؤ
    [[NSNotificationCenter defaultCenter]
        postNotificationName:PoCoChangePicture
                      object:self];

    // Խǽ򹹿()
    [self->view_ didChangeDrawTool];

    // Խι
    r = [self->picture_ bitmapPoCoRect];
    [[(PoCoAppController *)([NSApp delegate]) editInfo] setPictureRect:r];
    [r release];

    // ä
    [[self undoManager] setLevelsOfUndo:[[(PoCoAppController *)([NSApp delegate]) editInfo] undoMaxLevel]];

    return;
}


//
// äΰϤ
//
//  Call
//    sender : ɥ(api ѿ)
//
//  Return
//    function : ä
//
-(NSUndoManager *)windowWillReturnUndoManager:(NSWindow *)sender
{
    return [self undoManager];
}


//
// 
//
//  Call
//    type : 
//
//  Return
//    None
//
-(void)setUpdateState:(NSDocumentChangeType)type
{
    [self updateChangeCount:type];

    return;
}


//
// ѥåȹ
//
//  Call
//    view_    : ɽ(outlet)
//    picture_ : Խоݲ(instance ѿ)
//
//  Return
//    None
//
-(void)didUpdatePalette
{
    [self->picture_ didUpdatePalette];
    [self->view_ setNeedsDisplay:YES];

    return;
}


//
// 
//
//  Call
//    r     : ()
//    view_ : ɽ(outlet)
//
//  Return
//    None
//
-(void)didUpdatePicture:(PoCoRect *)r
{
    // 祦ɥι
    if (r == nil) {
        [self->view_ setNeedsDisplay:YES];
    } else {
        [self->view_ setNeedsDisplayInOriginalRect:r];
    }

    return;
}


//
// Խڤؤ
//  ڤؤǽֹ PoCoEditInfo ˵Ƥ
//
//  Call
//    view_ : ɽ(outlet)
//
//  Return
//    None
//
-(void)didChangeDrawTool
{
    [self->view_ didChangeDrawTool];

    return;
}


//
// 쥤䡼ѹ 
//
//  Call
//    view_ : ɽ(outlet)
//
//  Return
//    None
//
-(void)didLayerResize
{
    // view_ ɽ
    [self->view_ layerResize:YES];

    return;
}


//
// ɽɥ֥륯å
//
//  Call
//    p     : ɽɥǤμκǽ PD 
//    view_ : ɽ(outlet)
//
//  Return
//    None
//
-(void)dclickOnSubView:(PoCoPoint *)p
{
    // view_ ɽ
    [self->view_ doubleClickedOnSubView:p];

    return;
}


//
// ɥ饤ɽ
//
//  Call
//    view_ : ɽ(outlet)
//
//  Return
//    None
//
-(void)drawGuideLine
{
    [self->view_ drawGuideLine];

    return;
}


//
// Խֲ
//
//  Call
//    view_ : ɽ(outlet)
//
//  Return
//    None
//
-(void)cancelEdit
{
    [self->view_ cancelEdit];

    return;
}


//
// ڡ٥
//
//  Call
//    isPress : key down 
//              YES : keyDown ˤ
//              NO  : keyUp ˤ
//    view_   : ɽ(outlet)
//
//  Return
//    None
//
-(void)spaceKeyEvent:(BOOL)isPress
{
    [self->view_ spaceKeyEvent:isPress];

    return;
}


//
//  nib file
//  Xcode ɸ
//
//  Call
//    None
//
//  Return
//    function : nib file name
//
-(NSString *)windowNibName
{
    return @"MyDocument";
}


//
// nib ɤ߹߸
//  Xcode ɸ
//
//  Call
//    aController : window (api )
//    view_       : ɽ(outlet)
//    picture_    : Խоݲ(instance ѿ)
//    scroller_   : scroller(outlet)
//
//  Return
//    None
//
-(void)windowControllerDidLoadNib:(NSWindowController *)aController
{
    PoCoRect *r;
    PoCoPoint *p;

    [super windowControllerDidLoadNib:aController];

    // view_ κ׵ȯ
    [self->view_ layerResize:YES];
    [[NSNotificationCenter defaultCenter]
        postNotificationName:PoCoChangePicture
                      object:self];

    // mouse move ǽˤ
    [[aController window] setAcceptsMouseMovedEvents:YES];
    [[aController window] makeFirstResponder:self->view_];

    // ǽ˳֤򵭲
    if ([[(PoCoAppController *)([NSApp delegate]) editInfo] saveDocWindowPos]) {
        [aController setWindowFrameAutosaveName:@"MainViewWindowPos"];
    }

    // ԽϤμΤ
    [self->view_ didChangeDrawTool];

    // scrollbar ɽˡ
    switch ([[(PoCoAppController *)([NSApp delegate]) editInfo] showScrollerView]) {
        default:
        case PoCoScrollerType_default:
            // OS ˰¸
            ;
            break;
        case PoCoScrollerType_always:
            // 
            [self->scroller_ setScrollerStyle:NSScrollerStyleLegacy];
            break;
        case PoCoScrollerType_overlay:
            // Ŭ
            [self->scroller_ setScrollerStyle:NSScrollerStyleOverlay];
            break;
    }

    // 濴ɽ
    r = [self->picture_ bitmapPoCoRect];
    p = [[PoCoPoint alloc] initX:([r  width] / 2)
                           initY:([r height] / 2)];
    [self->view_ centering];
    [[(PoCoAppController *)([NSApp delegate]) editInfo] setLastPos:p];
    [p release];
    [r release];

    return;
}


//
// ¸μ¹
//  Xcode ɸ
//
//  Call
//    aType    : ̾(api )
//    outError : error(api )
//
//  Return
//    function : ¸
//
#if (MAC_OS_X_VERSION_MAX_ALLOWED > MAC_OS_X_VERSION_10_3)
-(NSData *)dataOfType:(NSString *)aType
                error:(NSError **)outError
{
    NSMutableData *data;

    DPRINT((@"[PoCoDocument dataOfType:%@ error]\n", aType));

    data = nil;
    *(outError) = nil;

    // ̤б
    if (!([aType isEqualToString:@"PoCoPicture"])) {
        goto EXIT;
    }

    // ɥ饤õ
    [self drawGuideLine];

    // 
    data = [NSMutableData data];
    [data appendData:[self createFileHeader]];
    [data appendData:[self->picture_ createFileData:[[(PoCoAppController *)([NSApp delegate]) editInfo] grayToAlpha]]];
    [data appendData:[self createFileFooter]];

    // ɥ饤ɽ
    [self drawGuideLine];

EXIT:
    return data;
}
#else   // (MAC_OS_X_VERSION_MAX_ALLOWED > MAC_OS_X_VERSION_10_3)
-(NSData *)dataRepresentationOfType:(NSString *)aType
{
    NSMutableData *data;

    DPRINT((@"[PoCoDocument dataRepresentationOfType:%@]\n", aType));

    data = nil;

    // ̤б
    if (!([aType isEqualToString:@"PoCoPicture"])) {
        goto EXIT;
    }

    // ɥ饤õ
    [self drawGuideLine];

    // 
    data = [NSMutableData data];
    [data appendData:[self createFileHeader]];
    [data appendData:[self->picture_ createFileData:[[(PoCoAppController *)([NSApp delegate]) editInfo] grayToAlpha]]];
    [data appendData:[self createFileFooter]];

    // ɥ饤ɽ
    [self drawGuideLine];

EXIT:
    return data;
}
#endif  // (MAC_OS_X_VERSION_MAX_ALLOWED > MAC_OS_X_VERSION_10_3)


//
// ɤ߹ߤμ¹
//  Xcode ɸ
//
//  Call
//    data     : ɤ߹(api )
//    aType    : ̾(api )
//    outError : error(api )
//    picture_ : Խоݲ(instance ѿ)
//
//  Return
//    function : YES : 
//               NO  : ۾
//
#if (MAC_OS_X_VERSION_MAX_ALLOWED > MAC_OS_X_VERSION_10_3)
-(BOOL)readFromData:(NSData *)data
             ofType:(NSString *)aType
              error:(NSError **)outError
{
    BOOL    result;
    NSRange range;

    DPRINT((@"[PoCoDocument readFromData ofType:%@ error]\n", aType));

    result = NO;                        // ͤϼ
    *(outError) = nil;
    range.location = 0;
    range.length = 0;

    // ǧ
    if (!([aType isEqualToString:@"PoCoPicture"])) {
        goto EXIT;
    }

    // إåǧ
    result = [MyDocument checkFileHeader:data
                               withRange:&(range)];
    if (result != YES) {
        DPRINT((@"not match header.\n"));
        goto EXIT;
    }

    // ɤ߹
    result = [self->picture_ loadFileData:data
                                withRange:&(range)];
    if (result != YES) {
        DPRINT((@"can't load data.\n"));
        goto EXIT;
    }

    // 
    result = YES;

EXIT:
    return result;
}
#else   // (MAC_OS_X_VERSION_MAX_ALLOWED > MAC_OS_X_VERSION_10_3)
-(BOOL)loadDataRepresentation:(NSData *)data
                       ofType:(NSString *)aType
{
    BOOL    result;
    NSRange range;

    DPRINT((@"[PoCoDocument loadDataRepresentation ofType:%@]\n", aType));

    result = NO;                        // ͤϼ
    range.location = 0;
    range.length = 0;

    // ǧ
    if (!([aType isEqualToString:@"PoCoPicture"])) {
        goto EXIT;
    }

    // إåǧ
    result = [MyDocument checkFileHeader:data
                               withRange:&(range)];
    if (result != YES) {
        DPRINT((@"not match header.\n"));
        goto EXIT;
    }

    // ɤ߹
    result = [self->picture_ loadFileData:data
                                withRange:&(range)];
    if (result != YES) {
        DPRINT((@"can't load data.\n"));
        goto EXIT;
    }

    // 
    result = YES;

EXIT:
    return result;
}
#endif  // (MAC_OS_X_VERSION_MAX_ALLOWED > MAC_OS_X_VERSION_10_3)


//
// ڡȥܡɤŽդ¹
//
//  Call
//    pb       : ĥդ
//    r        : ڤϰ
//    index    : оݥ쥤䡼
//    picture_ : Խоݲ(instance ѿ)
//
//  Return
//    None
//
-(void)writePngToPasteboard:(NSPasteboard *)pb
                   withRect:(PoCoRect *)r
                    atIndex:(int)index
{
    NSMutableData *data;

    DPRINT((@"[PoCoDocument writePngToPasteboard]\n"));

    // ȴ(current layer  PNG Ǽ)
    data = [NSMutableData data];
    [data appendData:[self createFileHeader]];
    [data appendData:[self->picture_ createPBoardData:r
                                              atIndex:index
                                              toAlpha:[[(PoCoAppController *)([NSApp delegate]) editInfo] grayToAlpha]]];
    [data appendData:[self createFileFooter]];
    if (data != nil) {
        // Ͽ
        [pb declareTypes:[NSArray arrayWithObject:NSPNGPboardType]
                   owner:self];

        // Ƥ
        [pb setData:data
            forType:NSPNGPboardType];
    }

    return;
}

@end
