//
//	Pelistina on Cocoa - PoCo -
//	PoCoView class
//
//	Copyright (C) 2005-2019 KAENRYUU Koutoku.
//

#import "PoCoView.h"

#import <Carbon/Carbon.h>

#import "PoCoAppController.h"
#import "PoCoSelLayer.h"
#import "PoCoEditInfo.h"
#import "PoCoPalette.h"
#import "PoCoPicture.h"
#import "PoCoMyDocument.h"

#import "PoCoDrawBase.h"
#import "PoCoDrawSelection.h"
#import "PoCoDrawFactory.h"

#import "PoCoDocumentSettingPanel.h"
#import "PoCoCanvasResizePanel.h"
#import "PoCoImportPalettePanel.h"
#import "PoCoAutoGradationPanel.h"
#import "PoCoColorReplacePanel.h"
#import "PoCoTexturePanel.h"
#import "PoCoToolbarWindow.h"

#import "PoCoControllerFactory.h"
#import "PoCoControllerPaletteIncrementer.h"
#import "PoCoControllerPaletteDecrementer.h"
#import "PoCoControllerPaletteAttributeSetter.h"


// 
static const int FRAME_GAP = 50;


// ѿ
static const int ZOOM_FACTOR[] = {      // ɽΨơ֥(0.1%ñ)
      10,   20,   25,   50,  100,  200,  250,  300,   400,   500,   600,   700,   750,   800,   900,
    1000,
    2000, 3000, 4000, 5000, 6000, 7000, 8000, 9000, 10000, 11000, 12000, 13000, 14000, 15000, 16000
};

// ============================================================================
@implementation PoCoView

// --------------------------------------------------------- instance - private
//
// Ψ˹碌򻻽
//
//  Call
//    zfact     : ɽΨ
//    frameGap_ : (instance ѿ)
//
//  Return
//    function : 
//
-(int)calcFrameGap:(int)zfact
{
    return self->frameGap_;
}


//
// ǤϰϤμ
//
//  Call
//    rect  : ΰ(׵ΰ)
//    zfact : ɽΨ
//
//  Return
//    function : Ǥκɸ
//
-(PoCoRect *)calcSrcRect:(NSRect)rect
                   zfact:(int)zfact
{
    PoCoRect *r;
    const int gap = [self calcFrameGap:zfact];

    // ɽΰ/ΰ򻻽
    r = [[PoCoRect alloc] initNSRect:rect];
    if (zfact == 1000) {
        // 
        [r shiftX:-(gap)];
        [r shiftY:-(gap)];
    } else {
        // /̾(üФʤΤǶ)
        [r   setLeft:(((rect.origin.x -              gap) * 1000) / zfact)];
        [r    setTop:(((rect.origin.y -              gap) * 1000) / zfact)];
        [r  setRight:(((rect.origin.x +  rect.size.width) * 1000) / zfact)];
        [r setBottom:(((rect.origin.y + rect.size.height) * 1000) / zfact)];
        if (((int)((rect.origin.x + rect.size.width) * 1000) % zfact) != 0) {
            [r setRight:([r right] + 1)];
        }
        if (((int)((rect.origin.y + rect.size.height) * 1000) % zfact) != 0) {
            [r setBottom:([r bottom] + 1)];
        }
    }

    return r;
}


//
// μ¹
//
//  Call
//    rect      : ΰ(׵ΰ)
//    zfact     : ɽΨ
//    document_ : Խоݤθ(outlet)
//
//  Return
//    None
//
-(void)execDraw:(NSRect)rect
          zfact:(int)zfact
{
    NSRect dr;                          // ɽΰ
    PoCoRect *sr;                       // ΰ
    NSBitmapImageRep *img;              // ɽ
    const int gap = [self calcFrameGap:zfact];

//    DPRINT((@"execDraw rect:{%f, %f, %f, %f}]\n", rect.origin.x, rect.origin.y, rect.size.width, rect.size.height));

    // ɽΰ/ΰ򻻽
    sr = [self calcSrcRect:rect
                     zfact:zfact];
    dr = rect;
    if (zfact == 1000) {
        // 
        if (rect.origin.x < gap) {
            dr.origin.x += (gap - rect.origin.x);
        }
        if (rect.origin.y < gap) {
            dr.origin.y += (gap - rect.origin.y);
        }
    } else {
        // /̾(üФʤΤǶ)
        if (rect.origin.x <= gap) {
            dr.origin.x = gap;
        } else {
            dr.origin.x = ((([sr left] * zfact) / 1000) + gap);
        }
        if (rect.origin.y <= gap) {
            dr.origin.y = gap;
        } else {
            dr.origin.y = ((([sr top] * zfact) / 1000) + gap);
        }
    }

    // ɽ
    img = [[self->document_ picture] getMainImage:sr
                                        withScale:zfact
                                   withSupplement:[self->editInfo_ supplement]];
    if (img != nil) {
        // ɽ¹
        [img drawAtPoint:dr.origin];

        // ɽ release
        [img release];
    }
    [sr release];

    return;
}


//
// ɥɸɸؤѴ
//
//  Call
//    p         : PD (ɥ)
//    zfactNum_ : ɽΨֹ(instance ѿ)
//
//  Return
//    function : ɸ
//
-(PoCoPoint *)toPictPos:(NSPoint)p
{
    int x;
    int y;
    const int zfact = ZOOM_FACTOR[self->zfactNum_];
    const int gap = [self calcFrameGap:zfact];

    if (zfact == 1000) {
        x = ((int)(p.x) - gap);
        y = ((int)(p.y) - gap);
    } else {
        x = ((((int)(p.x - gap)) * 1000) / zfact);
        y = ((((int)(p.y - gap)) * 1000) / zfact);
    }

    return [[PoCoPoint alloc] initX:x initY:y];
}


//
// PD ֤ã
//
//  Call
//    p         : PD (ɥ)
//    document_ : Խоݤθ(outlet)
//
//  Return
//    editInfo_ : Խ(instance ѿ)
//
-(void)setInfoPDPos:(NSPoint)p
{
    PoCoPoint *pos;
    PoCoRect *r;

    r = [[self->document_ picture] bitmapPoCoRect];

    // ΨȿǤ֤򻻽
    pos = [self toPictPos:p];

    // ϰ˼
    if ([pos x] < [r left]) {
        [pos setX:[r left]];
    } else if ([pos x] >= [r right]) {
        [pos setX:([r right] - 1)];
    }
    if ([pos y] < [r top]) {
        [pos setY:[r top]];
    } else if ([pos y] >= [r bottom]) {
        [pos setY:([r bottom] - 1)];
    }

    // 
    [self->editInfo_ setPDPos:pos];
    [pos release];
    [r release];

    return;
}


//
// ɽΨѹ
//
//  Call
//    num       : ѹͽΨֹ
//    zfactNum_ : ɽΨֹ(instance ѿ)
//    editInfo_ : Խ(instance ѿ)
//
//  Return
//    zfactNum_ : ɽΨֹ(instance ѿ)
//
-(void)changeZoom:(int)num
{
    int zchg;
    PoCoPoint *p;
    PoCoRect *r;

    r = nil;

    if (self->zfactNum_ != num) {
        // ɥ饤õ
        [self drawGuideLine];

        // ߤɽϰϤ
        r = [self calcSrcRect:[self visibleRect]
                        zfact:ZOOM_FACTOR[self->zfactNum_]];
        p = [self->editInfo_ lastPos];
        zchg = ((ZOOM_FACTOR[self->zfactNum_] * 1000) / ZOOM_FACTOR[num]);
        [r   setLeft:([p x] - ((([p x] - [r   left]) * zchg) / 1000))];
        [r    setTop:([p y] - ((([p y] - [r    top]) * zchg) / 1000))];
        [r  setRight:([p x] - ((([p x] - [r  right]) * zchg) / 1000))];
        [r setBottom:([p y] - ((([p y] - [r bottom]) * zchg) / 1000))];

        // 
        self->zfactNum_ = num;

        // 쥤䡼ѷ view ϰϤ򹹿
        [self layerResize:NO];

        // ɽ֤Ĵƺ׵ȯ
        [self scrollPoint:[self toDispRect:r].origin];

        // ɥ饤ɽ
        [self drawGuideLine];
    }
    [r release];

    return;
}


#if (MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_9)
// ------------------------------------------------------------- class - public
//
// responsive scroll
//
//  Call
//    None
//
//  Return
//    function :  YES
//
+(BOOL)isCompatibleWithResponsiveScrolling
{
    return YES;
}
#endif  // (MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_9)


// ---------------------------------------------------------- instance - public
//
// initialize
//
//  Call
//    frameRect : (api )
//
//  Return
//    function          : 
//    zfactNum_         : ɽΨֹ(instance ѿ)
//    frameGap_         : (instance ѿ)
//    isDrawing_        : (instance ѿ)
//    isAltMode_        : ALT Ʊ(instance ѿ)
//    isSpaceMode_      : SPACE Ʊ(instance ѿ)
//    editInfo_         : Խ(instance ѿ)
//    drawTool_         : Խǽμ(instance ѿ)
//    prevContinuation_ : Ϣ³/Ϣ³(instance ѿ)
//    prevDrawMode_     : 赡ǽ(instance ѿ)
//    prevPenStyle_     : Υڥ(instance ѿ)
//
-(id)initWithFrame:(NSRect)frameRect
{
    NSNotificationCenter *notes;

    DPRINT((@"[PoCoView initWithFrame:{%f, %f, %f, %f}]\n", frameRect.origin.x, frameRect.origin.y, frameRect.size.width, frameRect.size.height));

    // super class ν
    self = [super initWithFrame:frameRect];

    // Ȥν
    if (self != nil) {
        self->zfactNum_ = 15;
        self->frameGap_ = FRAME_GAP;
        self->isDrawing_ = NO;
        self->isAltMode_ = NO;
        self->isSpaceMode_ = NO;
        self->editInfo_ = nil;
        self->drawTool_ = nil;
        self->prevDrawMode_ = PoCoDrawModeType_MAX;
        self->prevPenStyle_ = PoCoPenStyleType_MAX;

        // Խμ
        self->editInfo_ = [(PoCoAppController *)([NSApp delegate]) editInfo];
        [self->editInfo_ retain];
        self->prevContinuation_ = [self->editInfo_ continuationType];

        // 򻻽
        self->frameGap_ = (int)(MIN([[NSScreen mainScreen] frame].size.width, [[NSScreen mainScreen] frame].size.height) * 0.1);

        // observer Ͽ
        notes = [NSNotificationCenter defaultCenter];
        [notes addObserver:self
                  selector:@selector(undoManagementing:)
                      name:NSUndoManagerWillUndoChangeNotification 
                    object:nil];
        [notes addObserver:self
                  selector:@selector(undoManagementing:)
                      name:NSUndoManagerDidUndoChangeNotification 
                    object:nil];
        [notes addObserver:self
                  selector:@selector(undoManagementing:)
                      name:NSUndoManagerWillRedoChangeNotification 
                    object:nil];
        [notes addObserver:self
                  selector:@selector(undoManagementing:)
                      name:NSUndoManagerDidRedoChangeNotification 
                    object:nil];

#if (MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_6)
        // touch event 
        [self setAcceptsTouchEvents:YES];
        [self setWantsRestingTouches:YES];
#endif  // MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_6
    }

    return self;
}


//
// deallocate
//
//  Call
//    None
//
//  Return
//    editInfo_ : Խ(instance ѿ)
//    drawTool_ : Խǽμ(instance ѿ)
//
-(void)dealloc
{
    DPRINT((@"[PoCoView dealloc]\n"));

    // observer Ͽ
    [[NSNotificationCenter defaultCenter] removeObserver:self];

    // 񸻤β
    [self->editInfo_ release];
    [self->drawTool_ release];
    self->editInfo_ = nil;
    self->drawTool_ = nil;

    // super class β
    [super dealloc];

    return;
}


//
// 쥤䡼ѷȼ
//
//  Call
//    isRedraw  : YES : ׵ȯԤ
//                NO  : ׵ȯԤʤ
//    document_ : Խоݤθ(outlet)
//    zfactNum_ : ɽΨֹ(instance ѿ)
//
//  Return
//    None
//
-(void)layerResize:(BOOL)isRedraw
{
    NSRect r;
    const int zfact = ZOOM_FACTOR[self->zfactNum_];
    const int gap = [self calcFrameGap:zfact];

    // view ΰ򹹿
    r = [[self->document_ picture] bitmapRect];
    r.size.width = ((r.size.width * zfact) / 1000);
    r.size.height = ((r.size.height * zfact) / 1000);
    r.size.width += (gap << 1);
    r.size.height += (gap << 1);
    [self setFrame:r];

    // ׵ȯ
    if (isRedraw) {
        [self setNeedsDisplay:YES];
    }

    return;
}


//
// 濴ɽ
//
//  Call
//    None
//
//  Return
//    None
//
-(void)centering
{
    NSRect fr = [self frame];
    NSRect vr = [self visibleRect];

    vr.origin.x = ((fr.size.width - vr.size.width) / 2);
    vr.origin.y = ((fr.size.height - vr.size.height) / 2);
    if (!([self scrollRectToVisible:vr])) {
        [self setNeedsDisplay:YES];
    }

    return;
}


//
// 
//  κɸϤǻؼ
//
//  Call
//    r : ΰ
//
//  Return
//    None
//
-(void)setNeedsDisplayInOriginalRect:(PoCoRect *)r
{
    // ɽȤ֤
    [self setNeedsDisplayInRect:[self toDispRect:r]];

    return;
}


//
// Խڤؤ
//  ڤؤǽֹ PoCoEditInfo ˵Ƥ
//
//  Call
//    document_         : Խоݤθ(outlet)
//    editInfo_         : Խ(instance ѿ)
//    prevContinuation_ : Ϣ³/Ϣ³(instance ѿ)
//    prevDrawMode_     : 赡ǽ(instance ѿ)
//    prevPenStyle_     : Υڥ(instance ѿ)
//
//  Return
//    drawTool_         : Խǽμ(instance ѿ)
//    prevContinuation_ : Ϣ³/Ϣ³(instance ѿ)
//    prevDrawMode_     : 赡ǽ(instance ѿ)
//    prevPenStyle_     : Υڥ(instance ѿ)
//
-(void)didChangeDrawTool
{
    DPRINT((@"[PoCoView didChangeDrawTool]\n"));

    if (([self->editInfo_ drawModeType] != self->prevDrawMode_) ||
        ([self->editInfo_ penStyleType] != self->prevPenStyle_) ||
        ([self->editInfo_ continuationType] != self->prevContinuation_)) {
        // μΤ˴
        [self->drawTool_ release];
        self->drawTool_ = nil;

        // ԽǽμΤ
        self->drawTool_ = [PoCoDrawFactory create:self->document_
                                         withDraw:[self->editInfo_ drawModeType]
                                          withPen:[self->editInfo_ penStyleType]];
    }
    self->prevContinuation_ = [self->editInfo_ continuationType];
    self->prevDrawMode_ = [self->editInfo_ drawModeType];
    self->prevPenStyle_ = [self->editInfo_ penStyleType];

    return;
}


//
// ɽɥ֥륯å
//
//  Call
//    dp        : ɽɥǤμκǽ PD 
//    zfactNum_ : ɽΨֹ(instance ѿ)
//    editInfo_ : Խ(instance ѿ)
//
//  Return
//    None
//
-(void)doubleClickedOnSubView:(PoCoPoint *)dp
{
    NSPoint p1;
    NSPoint p2;
    NSRect vr;
    PoCoPoint *lp;
    PoCoRect *dr;
    const int zfact = ZOOM_FACTOR[self->zfactNum_];
    const int gap = [self calcFrameGap:zfact];

    // Ʊͭʾ硢ɽ֤ư
    if ((dp != nil) &&
        ([self->editInfo_ syncSubView])) {
        lp = [self->editInfo_ lastPos];
        vr = [self visibleRect];
        dr = [self calcSrcRect:vr
                         zfact:zfact];

        // ɽ֤򻻽
        p1.x = (float)(gap + (([dp x] * zfact) / 1000));
        p1.y = (float)(gap + (([dp y] * zfact) / 1000));

        // ֤򻻽
        p2 = vr.origin;
        if ((p1.x < vr.origin.x) ||
            (p1.x > (vr.origin.x + vr.size.width))) {
            p2.x = (p1.x - (float)(((([lp x] - [dr left]) * zfact) / 1000)));
        }
        if ((p1.y < vr.origin.y) ||
            (p1.y > (vr.origin.y + vr.size.height))) {
            p2.y = (p1.y - (float)(((([lp y] - [dr top]) * zfact) / 1000)));
        }

        // ¹
        if ((p2.x != vr.origin.x) || (p2.y != vr.origin.y)) {
            [self scrollPoint:p2];
        }

        // ǽ PD ֤Ȥ
        [self->editInfo_ setLastPos:dp];
    }

    return;
}


//
// 
//
//  Call
//    rect      : ΰ(api )
//    zfactNum_ : ɽΨֹ(instance ѿ)
//    editInfo_ : Խ(instance ѿ)
//
//  Return
//    None
//
-(void)drawRect:(NSRect)rect
{
    long l;
    NSInteger cnt;
    const NSRect *r;
    PoCoRect *tr;

//    DPRINT((@"[PoCoView rect:{%f, %f, %f, %f}]\n", rect.origin.x, rect.origin.y, rect.size.width, rect.size.height));

    // liveResize 
    if ((!([self->editInfo_ enableLiveResize])) && ([self inLiveResize])) {
        return;
    }

    // ɽϰϤã
    tr = [self calcSrcRect:[self visibleRect]
                     zfact:ZOOM_FACTOR[self->zfactNum_]];
    [self->editInfo_ setPictureView:tr];
    [tr release];

    // ɽ¹
    [self getRectsBeingDrawn:&(r)
                       count:&(cnt)];
    [[NSGraphicsContext currentContext] setImageInterpolation:NSImageInterpolationNone];
    for (l = 0; l < cnt; (l)++) {
        [self execDraw:r[l] 
                 zfact:ZOOM_FACTOR[self->zfactNum_]];
    }

    return;
}


//
// ɸϤȿž
//
//  Call
//    None
//
//  Return
//    function : YES
//
-(BOOL)isFlipped
{
    return YES;
}


//
// ˥塼򹹿
//
//  Call
//    menu      : menu(api ѿ)
//    editInfo_ : Խ(instance ѿ)
//
//  Return
//    function : 
//
-(BOOL)validateMenuItem:(NSMenuItem *)menu
{
    BOOL result;
    PoCoColor *col = [[[self->document_ picture] palette] palette:[[self->editInfo_ selColor] num]];

    result = YES;

    if (([menu action] == @selector(copy:)) ||
        ([menu action] == @selector(delete:)) ||
        ([menu action] == @selector(flipHorizontal:)) ||
        ([menu action] == @selector(flipVertical:)) ||
        ([menu action] == @selector(autoGradation:)) ||
        ([menu action] == @selector(colorReplace:)) ||
        ([menu action] == @selector(selectClear:)) ||
        ([menu action] == @selector(clipImage:))) {
        result = (([[self->editInfo_ selRect] empty]) ? NO : YES);
    } else if ([menu action] == @selector(texture:)) {
        if ([[self->editInfo_ selRect] empty]) {
            // ϰϤ̵
            result = NO;
        } else if ([[self->document_ selLayer] sel] < 1) {
            // Υ쥤䡼̵(إ쥤䡼򤷤Ƥʤ)
            result = NO;
        }
    } else if (([menu action] == @selector(colorElementAll:)) ||
               ([menu action] == @selector(colorElementRed:)) ||
               ([menu action] == @selector(colorElementGreen:)) ||
               ([menu action] == @selector(colorElementBlue:))) {
        if ([self->editInfo_ lockPalette]) {
            // ѥåȸ
            result = NO;
        } else if ([self->editInfo_ colorMode] != PoCoColorMode_RGB) {
            // RGB 黻ǤϤʤ
            result = NO;
        }
    } else if ([menu action] == @selector(colorAttributeMask:)) {
        [menu setState:(([col isMask]) ? NSOnState : NSOffState)];
    } else if ([menu action] == @selector(colorAttribueDropper:)) {
        [menu setState:(([col noDropper]) ? NSOnState : NSOffState)];
    } else if ([menu action] == @selector(colorAttribueTransparent:)) {
        [menu setState:(([col isTrans]) ? NSOnState : NSOffState)];
    } else if ([menu action] == @selector(useUnderLayer:)) {
        [menu setState:(([self->editInfo_ useUnderPattern]) ? NSOnState : NSOffState)];
    }

    return result;
}


//
// ΰ褫饦ɥΰѴ
//
//  Call
//    r         : ɸ
//    zfactNum_ : ɽΨֹ(instance ѿ)
//
//  Return
//    function : ΰ
//
-(NSRect)toDispRect:(PoCoRect *)r
{
    NSRect nr;
    const int zfact = ZOOM_FACTOR[self->zfactNum_];
    const int gap = [self calcFrameGap:zfact];

    if (zfact == 1000) {
        // 
        nr.origin.x = (float)([r left] + gap);
        nr.origin.y = (float)([r top] + gap);
        nr.size.width = (float)([r width]);
        nr.size.height = (float)([r height]);
    } else {
        // /̾(üФʤΤǶ)
        nr.origin.x = (float)((([r left] * zfact) / 1000) + gap);
        nr.origin.y = (float)((([r top] * zfact) / 1000) + gap);
        nr.size.width = (float)(([r width] * zfact) / 1000);
        nr.size.height = (float)(([r height] * zfact) / 1000);
    }

    return nr;
}


//
// ɽΨ˱ϥɥγĥ
//  ̾ɽξ˥ϥɥ礭뤿ͤ
//
//  Call
//    zfactNum_ : ɽΨֹ(instance ѿ)
//
//  Return
//    function : 
//
-(int)handleRectGap;
{
    const int zfact = ZOOM_FACTOR[self->zfactNum_];

    return ((zfact < 1000) ? (2000 / zfact) : 0);
}


//
// Pointer ϰϻ
//
//  Call
//    drawTool_ : Խǽμ(instance ѿ)
//
//  Return
//    None
//
-(void)resetCursorRects
{
    // ԽǽȤ˹
    [self->drawTool_ updateCursorRect];

    return;
}


//
// ɥ饤ɽ
//
//  Call
//    drawTool_ : Խǽμ(instance ѿ)
//
//  Return
//    None
//
-(void)drawGuideLine
{
    [self->drawTool_ drawGuideLine];

    return;
}


//
// Խֲ
//
//  Call
//    drawTool_ : Խǽμ(instance ѿ)
//
//  Return
//    None
//
-(void)cancelEdit
{
    [self->drawTool_ cancelEdit];

    return;
}


//
// ڡ٥
//
//  Call
//    isPress    : key down 
//                 YES : keyDown ˤ
//                 NO  : keyUp ˤ
//    isDrawing_ : (instance ѿ)
//
//  Return
//    isAltMode_   : ALT Ʊ(instance ѿ)
//    isSpaceMode_ : SPACE Ʊ(instance ѿ)
//
-(void)spaceKeyEvent:(BOOL)isPress
{
    if (!(self->isDrawing_)) {
        self->isAltMode_ = isPress;
        self->isSpaceMode_ = isPress;
    }

    return;
}


//
// undo/redo ¹(observer)
//  WillUndoDidUndoWillRedoDidRedo γΤ
//  ȤȤȿžʤΤǡ
//    ä¹(Ȥ)ȿž򤫤ƾäƤޤ
//    ä¹Ԥƾ֤
//    ˤ⤦١Ȥȿž
//  ȤʼƧΤǡ٤ƤΤƱȤʤ롣
//
//  Call
//    note      : 
//    document_ : Խоݤθ(outlet)
//    drawTool_ : Խǽμ(instance ѿ)
//
//  Return
//    None
//
-(void)undoManagementing:(NSNotification *)note
{
    DPRINT((@"[PoCoView undoManagementing]\n"));

    if ([[NSDocumentController sharedDocumentController] currentDocument] == self->document_) {
        // ɥ饤(ȿž)
        [self->drawTool_ drawGuideLine];
        if (([[note name] compare:NSUndoManagerWillUndoChangeNotification] == NSOrderedSame) ||
            ([[note name] compare:NSUndoManagerWillRedoChangeNotification] == NSOrderedSame)) {
            [self->drawTool_ recreatePasteImage];
        }
    }

    return;
}


// ---------------------------------- instance - public - PasteBoardmenu 
//
// ڡȥܡɤʣ
//
//  Call
//    sender    : ˥塼ܤؤΥ󥹥(api ѿ)
//    document_ : Խоݤθ(outlet)
//    editInfo_ : Խ(instance ѿ)
//    drawTool_ : Խǽμ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)copy:(id)sender
{
    // Ⱦõ
    [self->drawTool_ drawGuideLine];

    // ڡȥܡɤĥդ
    [self->document_ writePngToPasteboard:[NSPasteboard generalPasteboard]
                                 withRect:[self->editInfo_ selRect]
                                  atIndex:[[self->document_ selLayer] sel]];

    // ɽ
    [self->drawTool_ drawGuideLine];

    return;
}


//
// ڡȥܡɤʣ
//
//  Call
//    sender    : ˥塼ܤؤΥ󥹥(api ѿ)
//    document_ : Խоݤθ(outlet)
//    drawTool_ : Խǽμ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)paste:(id)sender
{
    BOOL result;
    PoCoLayerBase *lyr;
    PoCoPalette *plt;
    PoCoImportPalettePanel *panel;

    lyr = nil;
    plt = nil;

    // ػߥ쥤䡼ʤ鲿⤷ʤ
    if ([[[self->document_ picture] layer:[[self->document_ selLayer] sel]] drawLock]) {
        NSBeep();
        goto EXIT;
    }

    // ѥåȤʣ
    plt = [[[self->document_ picture] palette] copy];

    // ¹
    [[self->document_ picture] pushPaletteTable];
    result = [MyDocument readPngFromPasteboard:[NSPasteboard generalPasteboard]
                                   resultLayer:&(lyr)
                                 resultPalette:plt];
    [[self->document_ picture] popPaletteTable];
    if ((result == YES) && (lyr != nil)) {
        // ʣ̥ѥå
        if (!([[[self->document_ picture] palette] isEqualPalette:plt
                                                       checkAttr:NO])) {
            panel = [[PoCoImportPalettePanel alloc] initWithDoc:self->document_
                                              withTargetPalette:plt];
            if (panel != nil) {
                [[panel window] center];
                [panel startWindow];
                [panel release];
            }
        }

        // Ⱦõ
        [self->drawTool_ drawGuideLine];

        // DrawSelection ڤؤ
        [[(PoCoAppController *)([NSApp delegate]) toolbarWindow] setDrawModeAtType:PoCoDrawModeType_Selection];

        // ϰ drawTool_ ˰Ϥ
        [self->drawTool_ pasteBitmap:lyr];
    } else {
        NSBeep();
    }

    // Ƥ
    [plt release];
    [lyr release];

EXIT:
    return;
}


//
// ɥå
//
//  Call
//    isLocal : flag
//
//  Return
//    function : 
//
-(NSDragOperation)draggingSourceOperationMaskForLocal:(BOOL)isLocal
{
    return NSDragOperationCopy;
}


//
// 
//
//  Call
//    sender    : ˥塼ܤؤΥ󥹥(api ѿ)
//    drawTool_ : Խǽμ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)delete:(id)sender
{
    // ¹
    [self->drawTool_ delete];

    return;
}


//
// ʿȿž
//
//  Call
//    sender    : ˥塼ܤؤΥ󥹥(api ѿ)
//    drawTool_ : Խǽμ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)flipHorizontal:(id)sender
{
    // ¹
    [self->drawTool_ flipImage:YES];

    return;
}


//
// ľȿž
//
//  Call
//    sender    : ˥塼ܤؤΥ󥹥(api ѿ)
//    drawTool_ : Խǽμ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)flipVertical:(id)sender
{
    [self->drawTool_ flipImage:NO];

    return;
}


//
// ưǡ
//
//  Call
//    sender    : ˥塼ܤؤΥ󥹥(api ѿ)
//    document_ : Խоݤθ(outlet)
//    editInfo_ : Խ(instance ѿ)
//    drawTool_ : Խǽμ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)autoGradation:(id)sender
{
    PoCoAutoGradationPanel *panel;

    panel = [[PoCoAutoGradationPanel alloc] initWithDoc:self->document_
                                           withEditInfo:self->editInfo_
                                              withImage:[self->drawTool_ originalImage]
                                              withShape:[self->drawTool_ originalShape]];
    if (panel != nil) {
        [[panel window] center];
        [panel startWindow];
        if ([panel isOk]) {
            [self->drawTool_ autoGrad:[panel penSize]
                           isAdjacent:[panel isAdjacent]
                               matrix:[panel matrix]
                         withSizePair:[panel sizePair]];
        }
        [panel release];
    }

    return;
}


//
// ִ
//
//  Call
//    sender    : ˥塼ܤؤΥ󥹥(api ѿ)
//    document_ : Խоݤθ(outlet)
//    editInfo_ : Խ(instance ѿ)
//    drawTool_ : Խǽμ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)colorReplace:(id)sender
{
    PoCoColorReplacePanel *panel;

    panel = [[PoCoColorReplacePanel alloc] initWithDoc:self->document_
                                          withEditInfo:self->editInfo_];
    if (panel != nil) {
        [[panel window] center];
        [panel startWindow];
        if ([panel isOk]) {
            [self->drawTool_ colorReplace:[panel matrix]];
        }
        [panel release];
    }

    return;
}


//
// ƥ
//
//  Call
//    sender    : ˥塼ܤؤΥ󥹥(api ѿ)
//    document_ : Խоݤθ(outlet)
//    editInfo_ : Խ(instance ѿ)
//    drawTool_ : Խǽμ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)texture:(id)sender
{
    PoCoTexturePanel *panel;

    panel = [[PoCoTexturePanel alloc] initWithDoc:self->document_
                                     withEditInfo:self->editInfo_];
    if (panel != nil) {
        [[panel window] center];
        [panel startWindow];
        if ([panel isOk]) {
            [self->drawTool_ texture:[panel base]
                        withGradient:[panel gradient]];
        }
        [panel release];
    }

    return;
}


//   
// 
//
//  Call
//    sender    : ˥塼ܤؤΥ󥹥(api ѿ)
//    drawTool_ : Խǽμ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)selectAll:(id)sender
{
    // DrawSelection ڤؤ
    [[(PoCoAppController *)([NSApp delegate]) toolbarWindow] setDrawModeAtType:PoCoDrawModeType_Selection];
    [self->drawTool_ selectAll];

    return;
}


//
// 
//
//  Call
//    sender    : ˥塼ܤؤΥ󥹥(api ѿ)
//    drawTool_ : Խǽμ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)selectClear:(id)sender
{
    [self->drawTool_ selectClear];

    return;
}


//
// ɽ
//
//  Call
//    sender    : ˥塼ܤؤΥ󥹥(api ѿ)
//    slider_   : Ψ slider(outlet)
//    zfactNum_ : ɽΨֹ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)zoomIn:(id)sender
{
    int num;

    num = (self->zfactNum_ + 1);

    // ɽȿ
    if ((num >= (int)([self->slider_ minValue])) &&
        (num <= (int)([self->slider_ maxValue]))) {
        [self->slider_ setFloatValue:(float)(num)];
        [self changeZoom:num];
    }

    return;
}


//
// ɽ̾
//
//  Call
//    sender    : ˥塼ܤؤΥ󥹥(api ѿ)
//    slider_   : Ψ slider(outlet)
//    zfactNum_ : ɽΨֹ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)zoomOut:(id)sender
{
    int num;

    num = (self->zfactNum_ - 1);

    // ɽȿ
    if ((num >= (int)([self->slider_ minValue])) &&
        (num <= (int)([self->slider_ maxValue]))) {
        [self->slider_ setFloatValue:(float)(num)];
        [self changeZoom:num];
    }

    return;
}


//
// 
//
//  Call
//    sender    : ˥塼ܤؤΥ󥹥(api ѿ)
//    slider_   : Ψ slider(outlet)
//    zfactNum_ : ɽΨֹ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)actualSize:(id)sender
{
    int num;

    num = ((int)([self->slider_ maxValue] - [self->slider_ minValue]) / 2);

    [self->slider_ setFloatValue:(float)(num)];
    [self changeZoom:num];

    return;
}


//
// ɥ
//
//  Call
//    sender    : о(api )
//    document_ : Խоݤθ(outlet)
//
//  Return
//    None
//
-(IBAction)documentSetting:(id)sender
{
    PoCoDocumentSettingPanel *panel;

    panel = [[PoCoDocumentSettingPanel alloc] initWithDoc:self->document_];
    if (panel != nil) {
        [[panel window] center];
        [panel startWindow];
        [panel release];
    }

    return;
}


//
// ѹ
//
//  Call
//    sender    : о(api )
//    document_ : Խоݤθ(outlet)
//
//  Return
//    None
//
-(IBAction)canvasResize:(id)sender
{
    PoCoCanvasResizePanel *panel;

    panel = [[PoCoCanvasResizePanel alloc] initWithDoc:self->document_];
    if (panel != nil) {
        [[panel window] center];
        [panel startWindow];
        [panel release];
    }

    return;
}


//
// ڤȴ
//
//  Call
//    sender    : о(api )
//    editInfo_ : Խ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)clipImage:(id)sender
{
    PoCoRect *r;

    r = nil;

    // ϰϤʤ鲿⤷ʤ
    if ([[self->editInfo_ selRect] empty]) {
        goto EXIT;
    }

    // ϰϤζȤ
    r = [[PoCoRect alloc] initPoCoRect:[self->editInfo_ selRect]];

    // Խֲ
    [self->document_ cancelEdit];

    // ¹
    [[(PoCoAppController *)([NSApp delegate]) factory] createImageClipper:YES
                                                                     rect:r];

EXIT:
    [r release];
    return;
}


// --------------------------------------- instance - public - ѥåϢ
//
// ο
//
//  Call
//    sender    : (api )
//    editInfo_ : Խ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)nextColor:(id)sender
{
    PoCoSelColor *col;
    int num;

    col = [self->editInfo_ selColor];
    if (([col isPattern]) || ([col isUnder])) {
        // 顼ѥ󡢲إ쥤䡼ڤؤʤ
        ;
    } else {
        num = (((([col num] + 1) >= COLOR_MAX) ? -1 : [col num]) + 1);
        [self->editInfo_ setSelColor:num];

        // ڤؤ
        [[NSNotificationCenter defaultCenter]
            postNotificationName:PoCoChangeColor
                          object:nil];
    }

    return;
}


//
// ο
//
//  Call
//    sender    : (api )
//    editInfo_ : Խ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)prevColor:(id)sender
{
    PoCoSelColor *col;
    int num;

    col = [self->editInfo_ selColor];
    if (([col isPattern]) || ([col isUnder])) {
        // 顼ѥ󡢲إ쥤䡼ڤؤʤ
        ;
    } else {
        num = ((([col num] <= 0) ? COLOR_MAX : [col num]) - 1);
        [self->editInfo_ setSelColor:num];

        // ڤؤ
        [[NSNotificationCenter defaultCenter]
            postNotificationName:PoCoChangeColor
                          object:nil];
    }

    return;
}


//
// إ쥤䡼
//
//  Call
//    sender    : (api )
//    editInfo_ : Խ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)useUnderLayer:(id)sender
{
    // ȿž
    [self->editInfo_ toggleUseUnderPattern];

    // ڤؤ(Ȥ)
    [[NSNotificationCenter defaultCenter] postNotificationName:PoCoChangeColor
                                                        object:nil];


    return;
}


//
// (٤)
//
//  Call
//    sender    : (api )
//    editInfo_ : Խ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)colorElementAll:(id)sender
{
    if ([self->editInfo_ lockPalette]) {
        // ѥåȸ
        ;
    } else if ([self->editInfo_ colorMode] != PoCoColorMode_RGB) {
        // RGB 黻ǤϤʤ
        ;
    } else if ([sender tag] < 0) {
        [[(PoCoAppController *)([NSApp delegate]) factory]
            createPaletteDecrementer:YES
                                 num:[[self->editInfo_ selColor] num]
                                 red:YES
                               green:YES
                                blue:YES
                                step:(unsigned int)(-([sender tag]))];
    } else {
        [[(PoCoAppController *)([NSApp delegate]) factory]
            createPaletteIncrementer:YES
                                 num:[[self->editInfo_ selColor] num]
                                 red:YES
                               green:YES
                                blue:YES
                                step:(unsigned int)([sender tag])];
    }

    return;
}


//
// ()
//
//  Call
//    sender    : (api )
//    editInfo_ : Խ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)colorElementRed:(id)sender
{
    if ([self->editInfo_ lockPalette]) {
        // ѥåȸ
        ;
    } else if ([self->editInfo_ colorMode] != PoCoColorMode_RGB) {
        // RGB 黻ǤϤʤ
        ;
    } else if ([sender tag] < 0) {
        [[(PoCoAppController *)([NSApp delegate]) factory]
            createPaletteDecrementer:YES
                                 num:[[self->editInfo_ selColor] num]
                                 red:YES
                               green:NO
                                blue:NO
                                step:(unsigned int)(-([sender tag]))];
    } else {
        [[(PoCoAppController *)([NSApp delegate]) factory]
            createPaletteIncrementer:YES
                                 num:[[self->editInfo_ selColor] num]
                                 red:YES
                               green:NO
                                blue:NO
                                step:(unsigned int)([sender tag])];
    }

    return;
}


//
// ()
//
//  Call
//    sender    : (api )
//    editInfo_ : Խ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)colorElementGreen:(id)sender
{
    if ([self->editInfo_ lockPalette]) {
        // ѥåȸ
        ;
    } else if ([self->editInfo_ colorMode] != PoCoColorMode_RGB) {
        // RGB 黻ǤϤʤ
        ;
    } else if ([sender tag] < 0) {
        [[(PoCoAppController *)([NSApp delegate]) factory]
            createPaletteDecrementer:YES
                                 num:[[self->editInfo_ selColor] num]
                                 red:NO
                               green:YES
                                blue:NO
                                step:(unsigned int)(-([sender tag]))];
    } else {
        [[(PoCoAppController *)([NSApp delegate]) factory]
            createPaletteIncrementer:YES
                                 num:[[self->editInfo_ selColor] num]
                                 red:NO
                               green:YES
                                blue:NO
                                step:(unsigned int)([sender tag])];
    }

    return;
}


//
// ()
//
//  Call
//    sender    : (api )
//    editInfo_ : Խ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)colorElementBlue:(id)sender
{
    if ([self->editInfo_ lockPalette]) {
        // ѥåȸ
        ;
    } else if ([self->editInfo_ colorMode] != PoCoColorMode_RGB) {
        // RGB 黻ǤϤʤ
        ;
    } else if ([sender tag] < 0) {
        [[(PoCoAppController *)([NSApp delegate]) factory]
            createPaletteDecrementer:YES
                                 num:[[self->editInfo_ selColor] num]
                                 red:NO
                               green:NO
                                blue:YES
                                step:(unsigned int)(-([sender tag]))];
    } else {
        [[(PoCoAppController *)([NSApp delegate]) factory]
            createPaletteIncrementer:YES
                                 num:[[self->editInfo_ selColor] num]
                                 red:NO
                               green:NO
                                blue:YES
                                step:(unsigned int)([sender tag])];
    }

    return;
}


//
// ɤػ
//
//  Call
//    sender    : (api )
//    editInfo_ : Խ(instance ѿ)
//
//  Return
//    None
//
-(IBAction)colorAttributeMask:(id)sender
{
    PoCoSelColor *scol = [self->editInfo_ selColor];
    PoCoColor *col = [[[self->document_ picture] palette] palette:[scol num]];

    [[(PoCoAppController *)([NSApp delegate]) factory]
        createPaletteAttributeSetter:YES
                                 num:[scol num]
                             setType:(([col isMask]) ? NO : YES)
                                mask:YES
                             dropper:NO
                               trans:NO];

    return;
}


//
// ۤػ
//
//  Call
//    sender : (api )
//
//  Return
//    None
//
-(IBAction)colorAttribueDropper:(id)sender
{
    PoCoSelColor *scol = [self->editInfo_ selColor];
    PoCoColor *col = [[[self->document_ picture] palette] palette:[scol num]];

    [[(PoCoAppController *)([NSApp delegate]) factory]
        createPaletteAttributeSetter:YES
                                 num:[scol num]
                             setType:(([col noDropper]) ? NO : YES)
                                mask:NO
                             dropper:YES
                               trans:NO];

    return;
}


//
// Ʃ
//
//  Call
//    sender : (api )
//
//  Return
//    None
//
-(IBAction)colorAttribueTransparent:(id)sender
{
    PoCoSelColor *scol = [self->editInfo_ selColor];
    PoCoColor *col = [[[self->document_ picture] palette] palette:[scol num]];

    [[(PoCoAppController *)([NSApp delegate]) factory]
        createPaletteAttributeSetter:YES
                                 num:[scol num]
                             setType:(([col isTrans]) ? NO : YES)
                                mask:NO
                             dropper:NO
                               trans:YES];

    return;
}


// ----------------------------------------- instance - public - ٥Ƚ
//
// ܥ󥤥٥Ȥμ
//
//  Call
//    evt : ȯ٥(api ѿ)
// 
//  Return
//    function : 
//
-(BOOL)acceptsFirstMouse:(NSEvent *)evt
{
    // Ƚ̵
    ;

    return YES;
}


#if (MAC_OS_X_VERSION_MAX_ALLOWED > MAC_OS_X_VERSION_10_3)
//
// ڥ
//
//  Call
//    evt       : ȯ٥(api ѿ)
//    editInfo_ : Խ(instance ѿ)
//
//  Return
//    editInfo_ : Խ(instance ѿ)
//
-(void)tabletProximity:(NSEvent *)evt
{
//    DPRINT((@"\nenter : %s\npointingDeviceType : %d\nPointingDeviceID : %d\n", [evt isEnteringProximity] ? "YES" : "NO", [evt pointingDeviceType], [evt pointingDeviceID]));

    if ([self->editInfo_ enableEraser]) {
        [self->editInfo_ setEraserType:([evt pointingDeviceType] == NSEraserPointingDevice)];
    }

    return;
}
#endif  // MAC_OS_X_VERSION_MAX_ALLOWED > MAC_OS_X_VERSION_10_3


//
// ܥ
//
//  Call
//    evt          : ȯ٥(api ѿ)
//    isDrawing_   : (instance ѿ)
//    editInfo_    : Խ(instance ѿ)
//    drawTool_    : Խǽμ(instance ѿ)
//    isSpaceMode_ : SPACE Ʊ(instance ѿ)
//
//  Return
//    isDrawing_ : (instance ѿ)
//    isAltMode_ : ALT Ʊ(instance ѿ)
//
-(void)mouseDown:(NSEvent *)evt
{
    NSPoint p;

    // PD ּ
    p = [PoCoPoint corrNSPoint:[self convertPoint:[evt locationInWindow]
                                         fromView:nil]];

    // PD ֤ã
    [self setInfoPDPos:p];

    // Alt ƱȽ
    if (self->isDrawing_) {
        // Ǥ
        ;
    } else if (!(self->isSpaceMode_)) {
        self->isAltMode_ = ((([evt modifierFlags] & NSCommandKeyMask) == 0) && (([evt modifierFlags] & NSAlternateKeyMask) != 0));
    }

    // ¹
    if (self->isAltMode_) {
        // [Alt] Ʊ
        ;
    } else {
        // ¹
        [self->drawTool_ mouseDown:evt];
        self->isDrawing_ = YES;
    }

    // Խμ¹Ԥ
    if ([self->editInfo_ syncSubView]) {
        [[NSNotificationCenter defaultCenter]
            postNotificationName:PoCoExecDrawing
                          object:nil];
    }

    return;
}


//
// ɥå
//
//  Call
//    evt        : ȯ٥(api ѿ)
//    isAltMode_ : ALT Ʊ(instance ѿ)
//    drawTool_  : Խǽμ(instance ѿ)
//
//  Return
//    None
//
-(void)mouseDragged:(NSEvent *)evt
{
    NSPoint p;

    // PD ּ
    p = [PoCoPoint corrNSPoint:[self convertPoint:[evt locationInWindow]
                                         fromView:nil]];

    // PD ֤ã
    [self setInfoPDPos:p];

    // ԽǽȤʬ
    if (self->isAltMode_) {
        // [Alt] Ʊ
        [self->drawTool_ dragMove:evt];
    } else {
        // auto-scroll ¹
        if ([self->drawTool_ canAutoScroll]) {
            [self autoscroll:evt];
        }

        // ¹
        [self->drawTool_ mouseDrag:evt];
    }

    // Խμ¹Ԥ
    if ([self->editInfo_ syncSubView]) {
        [[NSNotificationCenter defaultCenter]
            postNotificationName:PoCoExecDrawing
                          object:nil];
    }

    return;
}


//
// ܥ꡼
//
//  Call
//    evt        : ȯ٥(api ѿ)
//    isAltMode_ : ALT Ʊ(instance ѿ)
//    editInfo_  : Խ(instance ѿ)
//    drawTool_  : Խǽμ(instance ѿ)
//
//  Return
//    isDrawing_ : (instance ѿ)
//    editInfo_  : Խ(instance ѿ)
//
-(void)mouseUp:(NSEvent *)evt
{
    // ԽǽȤʬ
    if (self->isAltMode_) {
        // [Alt] Ʊ
        [self->editInfo_ setLastPos:[self->editInfo_ pdPos]];
    } else {
        // ¹
        [self->drawTool_ mouseUp:evt];
        self->isDrawing_ = NO;
    }

    // Խμ¹Ԥ
    if ([self->editInfo_ syncSubView]) {
        [[NSNotificationCenter defaultCenter]
            postNotificationName:PoCoExecDrawing
                          object:nil];
    }

    return;
}


//
// ܥ
//  Խʤ鿧ۤ
//  ԽʤԽ
//
//  Call
//    evt       : ȯ٥(api ѿ)
//    drawTool_ : Խǽμ(instance ѿ)
//
//  Return
//    None
//
-(void)rightMouseDown:(NSEvent *)evt
{
    NSPoint p;

    // PD ּ
    p = [PoCoPoint corrNSPoint:[self convertPoint:[evt locationInWindow]
                                         fromView:nil]];
    if ([self mouse:p
             inRect:[self visibleRect]]) {
        // PD ֤ã
        [self setInfoPDPos:p];

        // ¹
        [self->drawTool_ rightMouseDown:evt];
    }

    return;
}


//
// ܥɥå
//
//  Call
//    evt       : ȯ٥(api ѿ)
//    drawTool_ : Խǽμ(instance ѿ)
//
//  Return
//    None
//
-(void)rightMouseDragged:(NSEvent *)evt
{
    NSPoint p;

    // PD ּ
    p = [PoCoPoint corrNSPoint:[self convertPoint:[evt locationInWindow]
                                         fromView:nil]];

    // PD ֤ã
    [self setInfoPDPos:p];

    // ¹
    [self->drawTool_ rightMouseDrag:evt];

    return;
}


//
// ܥ꡼
//
//  Call
//    evt       : ȯ٥(api ѿ)
//    drawTool_ : Խǽμ(instance ѿ)
//
//  Return
//    None
//
-(void)rightMouseUp:(NSEvent *)evt
{
    NSPoint p;

    // PD ּ
    p = [PoCoPoint corrNSPoint:[self convertPoint:[evt locationInWindow]
                                         fromView:nil]];

    // PD ֤ã
    [self setInfoPDPos:p];

    // ¹
    [self->drawTool_ rightMouseUp:evt];

    return;
}


//
// 
//
//  Call
//    evt        : ȯ٥(api ѿ)
//    isDrawing_ : (instance ѿ)
//    drawTool_  : Խǽμ(instance ѿ)
//
//  Return
//    isAltMode_   : ALT Ʊ(instance ѿ)
//    isSpaceMode_ : SPACE Ʊ(instance ѿ)
//
-(void)keyDown:(NSEvent *)evt
{
    if ((!(self->isDrawing_)) && ([evt keyCode] == kVK_Space)) {
        // ڡϾˤꤺˤ
        if ([evt isARepeat]) {
            // key repeat ̵
            ;
        } else {
            self->isAltMode_ = YES;
            self->isSpaceMode_ = YES;
        }
    } else {
        // ¹
        [self->drawTool_ keyDown:evt];
    }

    return;
}


//
// ꡼
//
//  Call
//    evt       : ȯ٥(api ѿ)
//    drawTool_ : Խǽμ(instance ѿ)
//
//  Return
//    isDrawing_   : (instance ѿ)
//    isAltMode_   : ALT Ʊ(instance ѿ)
//    isSpaceMode_ : SPACE Ʊ(instance ѿ)
//
-(void)keyUp:(NSEvent *)evt
{
    if ((!(self->isDrawing_)) && ([evt keyCode] == kVK_Space)) {
        // ڡϾˤꤺˤ
        self->isAltMode_ = NO;
        self->isSpaceMode_ = NO;
    } else {
        // ¹
        [self->drawTool_ keyUp:evt];
    }

    return;
}


//
// PD ư
//
//  Call
//    evt        : ȯ٥(api ѿ)
//    isAltMode_ : ALT Ʊ(instance ѿ)
//    drawTool_  : Խǽμ(instance ѿ)
//
//  Return
//    None
//
-(void)mouseMoved:(NSEvent *)evt
{
    NSPoint p;

    // PD ּ
    p = [PoCoPoint corrNSPoint:[self convertPoint:[evt locationInWindow]
                                         fromView:nil]];
    if ([self mouse:p
             inRect:[self visibleRect]]) {
        // PD ֤ã
        [self setInfoPDPos:p];

        // ԽǽȤʬ
        if (self->isAltMode_) {
            // [Alt] Ʊ
            ;
        } else {
            // ¹
            [self->drawTool_ mouseMove:evt];
        }
    }

    return;
}


//
// wheel ž
//
//  Call
//    evt       : ȯ٥(api ѿ)
//    slider_   : Ψ slider(outlet)
//    zfactNum_ : ɽΨֹ(instance ѿ)
//
//  Return
//    None
//
-(void)scrollWheel:(NSEvent *)evt
{
    int num;

    if ([evt modifierFlags] & NSControlKeyMask) {
        // ɽΨѹ
        if (((int)([evt deltaX]) == 0) &&
            ((int)([evt deltaY]) == 0) &&
            ((int)([evt deltaZ]) == 0)) {
            // ̵ϲ⤷ʤ(Wacom Tablet Driver  bug)
            ;
        } else {
            if (((int)([evt deltaX]) < 0) || ((int)([evt deltaY]) < 0)) {
                num = (self->zfactNum_ - 1);
            } else {
                num = (self->zfactNum_ + 1);
            }

            // ɽȿ
            if ((num >= (int)([self->slider_ minValue])) &&
                (num <= (int)([self->slider_ maxValue]))) {
                [self->slider_ setFloatValue:(float)(num)];
                [self changeZoom:num];
            }
        }
    } else {
        // ̾ wheel ž
        [super scrollWheel:evt];
    }

    return;
}


#if (MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_6)
//
// touch 
//
//  Call
//    evt : ȯ٥(api ѿ)
//
//  Return
//    None
//
-(void)touchesBeganWithEvent:(NSEvent *)evt
{
    // ⤷ʤ
    ;

    // super class ز
    [super touchesBeganWithEvent:evt];

    return;
}


//
// touch ư
//
//  Call
//    evt : ȯ٥(api ѿ)
//
//  Return
//    None
//
-(void)touchesMovedWithEvent:(NSEvent *)evt
{
    // pointer 
    [[self window] invalidateCursorRectsForView:self];

    // super class ز
    [super touchesMovedWithEvent:evt];

    return;
}


//
// touch λ
//
//  Call
//    evt : ȯ٥(api ѿ)
//
//  Return
//    None
//
-(void)touchesEndedWithEvent:(NSEvent *)evt
{
    // pointer 
    [[self window] invalidateCursorRectsForView:self];

    // super class ز
    [super touchesEndedWithEvent:evt];

    return;
}


//
// touch 
//
//  Call
//    evt : ȯ٥(api ѿ)
//
//  Return
//    None
//
-(void)touchesCancelledWithEvent:(NSEvent *)evt
{
    // ⤷ʤ(ݤʤ)
    ;

    // super class ز
    [super touchesCancelledWithEvent:evt];

    return;
}
#endif  // MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_6


// -------------------------------------------- instance - public - IBAction 
//
// Ψ slider δ
//
//  Call
//    sender : Ƥ slider(api )
//
//  Return
//    None
//
-(IBAction)zoomFactor:(id)sender
{
    [self changeZoom:(int)(floor([sender floatValue]))];

    return;
}


// ---------------------------------------- instance - public - liveResize 
//
// live resize 
//
//  Call
//    editInfo_ : Խ(instance ѿ)
//
//  Return
//    None
//
-(void)viewWillStartLiveResize
{
    if ([self->editInfo_ enableLiveResize]) {
        [super viewWillStartLiveResize];
    }

    return;
}


//
// live resize λ
//
//  Call
//    editInfo_ : Խ(instance ѿ)
//
//  Return
//    None
//
-(void)viewDidEndLiveResize
{
    if (!([self->editInfo_ enableLiveResize])) {
        [self setNeedsDisplay:YES];
    }

    return;
}

@end
