//
//	Pelistina on Cocoa - PoCo -
//	ѥåȥɥ
//
//	Copyright (C) 2005-2018 KAENRYUU Koutoku.
//

#import "PoCoPaletteWindow.h"

#import <Carbon/Carbon.h>

#import "PoCoAppController.h"
#import "PoCoEditInfo.h"
#import "PoCoMyDocument.h"
#import "PoCoPicture.h"
#import "PoCoPalette.h"
#import "PoCoPaletteView.h"
#import "PoCoPaletteInfoView.h"
#import "PoCoPaletteAttributeView.h"
#import "PoCoPaletteGradationView.h"
#import "PoCoPaletteExchangeView.h"

#import "PoCoControllerFactory.h"
#import "PoCoControllerPaletteAttributeSetter.h"
#import "PoCoControllerPictureColorExchanger.h"
#import "PoCoControllerPictureColorPaster.h"
#import "PoCoControllerPaletteNoEditter.h"

// ѿ
static NSString *DISCLOSE_NAME = @"PoCoPaletteWindowDiscloseState";
static NSString *MASK_SELRANGE = @"PoCoPaletteWindowMaskSelectionRange";
static NSString *DROP_SELRANGE = @"PoCoPaletteWindowNoDropperSelectionRange";
static NSString *TRNS_SELRANGE = @"PoCoPaletteWindowTransparentSelectionRange";

// ============================================================================
@implementation PoCoPaletteWindow

// ------------------------------------------------------------- class - public
//
// 
//
//  Call
//    None
//
//  Return
//    None
//
+(void)initialize
{
    NSMutableDictionary *dic;

    dic = [NSMutableDictionary dictionary];

    // ƽͤ
    [dic setObject:[NSNumber numberWithBool:NO]
            forKey:DISCLOSE_NAME];
    [dic setObject:[NSNumber numberWithInt:1]
            forKey:MASK_SELRANGE];
    [dic setObject:[NSNumber numberWithInt:1]
            forKey:DROP_SELRANGE];
    [dic setObject:[NSNumber numberWithInt:1]
            forKey:TRNS_SELRANGE];

    // default 
    [[NSUserDefaults standardUserDefaults] registerDefaults:dic];

    return;
}


// ---------------------------------------------------------- instance - public
//
// initialize
//
//  Call
//    None
//
//  Return
//    function : 
//    docCntl_ : document controller(instance ѿ)
//
-(id)init
{
    // super class 
    self = [super initWithWindowNibName:@"PoCoPaletteWindow"];

    // Ȥ
    if (self != nil) {
        // document controller 
        self->docCntl_ = [NSDocumentController sharedDocumentController];
        [self->docCntl_ retain];

        // ¸
        [self setWindowFrameAutosaveName:@"PaletteWindowPos"];
    }

    return self;
}


//
// deallocate
//
//  Call
//    None
//
//  Return
//    docCntl_ : document controller(instance ѿ)
//
-(void)dealloc
{
    // 񸻤
    [self->docCntl_ release];
    self->docCntl_ = nil;

    // super class β
    [super dealloc];

    return;
}


//
// ɥɤ߹
//
//  Call
//    mode_            : 黻⡼(outlet)
//    disclose_        : ɥȽ(outlet)
//    paletteView_     : ѥåȰ(outlet)
//    paletteInfoView_ : ѥåȾܺ(outlet)
//    scroller_        : scroller(outlet)
//
//  Return
//    None
//
-(void)windowDidLoad
{
    BOOL  stat;
    NSRect r = [[self window] frame];

    // user default μ
    stat = [[NSUserDefaults standardUserDefaults] boolForKey:DISCLOSE_NAME];
    [self->disclose_ setState:stat];

    // window ѷ
    r.size.width = ((stat) ? 128 : 431);
    [[self window] setFrame:r
                    display:YES
                    animate:NO];
    [self->paletteInfoView_ setFirstState];
    [self->paletteView_ setDisclosed:stat];

    // 黻⡼ɤ
    switch ([[(PoCoAppController *)([NSApp delegate]) editInfo] colorMode]) {
        case PoCoColorMode_RGB:
        default:
            [self->mode_ setState:1 atRow:0 column:0];
            [self->mode_ setState:0 atRow:0 column:1];
            break;
        case PoCoColorMode_HLS:
            [self->mode_ setState:0 atRow:0 column:0];
            [self->mode_ setState:1 atRow:0 column:1];
            break;
    }

    // scrollbar ɽˤ
    [self->scroller_ setScrollerStyle:NSScrollerStyleLegacy];

    // ػ
    [self->scroller_ setVerticalScrollElasticity:NSScrollElasticityNone];

    return;
}


//
// ɥĤ(delegate method)
//
//  Call
//    note : (api ѿ)
//
//  Return
//    None
//
-(void)windowWillClose:(NSNotification *)note
{
    [(PoCoAppController *)([NSApp delegate]) closedPaletteWindow];

    return;
}
 

//      
// äΰϤ
//
//  Call
//    sender   : ɥ(api ѿ)
//    docCntl_ : document controller(instance ѿ)
//
//  Return
//    function : ä
//
-(NSUndoManager *)windowWillReturnUndoManager:(NSWindow *)sender
{ 
    return [[self->docCntl_ currentDocument] undoManager];
}


//
// ѥåȸ
//
//  Call
//    sender           : о
//    paletteInfoView_ : ѥåȾܺ(outlet)
//
//  Return
//    None
//
-(void)lockPalette:(id)sender
{
    // Τޤ޲
    [self->paletteInfoView_ lockPalette:sender];

    return;
}


//
// ɥȽ
//
//  Call
//    None
//
//  Return
//    function : 
//
-(BOOL)isDisclose
{
    return [[NSUserDefaults standardUserDefaults] boolForKey:DISCLOSE_NAME];
}


// -------------------------------------------- instance - public - IBAction 
//
// ɥȽ
//
//  Call
//    sender       : о(api ѿ)
//    disclose_    : ɥȽ(outlet)
//    paletteView_ : ѥåȰ(outlet)
//
//  Return
//    disclose_ : ɥȽ(outlet)
//
-(IBAction)changeDisclose:(id)sender
{
    NSRect r = [[self window] frame];
    const BOOL stat = (([[NSUserDefaults standardUserDefaults] boolForKey:DISCLOSE_NAME]) ? NO : YES);

    [self->disclose_ setState:stat];
    r.size.width = ((stat) ? 128 : 431);
    [[self window] setFrame:r
                    display:YES
                    animate:YES];
    [self->paletteView_ setDisclosed:stat];

    // user default ι
    [[NSUserDefaults standardUserDefaults] setBool:stat
                                            forKey:DISCLOSE_NAME];

    return;
}


//
// 黻⡼ڤؤ
//
//  Call
//    sender           : о(api ѿ)
//    mode_            : 黻⡼(outlet)
//    paletteInfoView_ : ѥåȾܺ(outlet)
//
//  Return
//    None
//
-(IBAction)changeMode:(id)sender
{
    switch ((sender == self->mode_) ? [sender selectedColumn] : [sender tag]) {
        case 0:
        default:
            [[(PoCoAppController *)([NSApp delegate]) editInfo] setColorMode:PoCoColorMode_RGB];
            [self->mode_ setState:1 atRow:0 column:0];
            [self->mode_ setState:0 atRow:0 column:1];
            break;
        case 1:
            [[(PoCoAppController *)([NSApp delegate]) editInfo] setColorMode:PoCoColorMode_HLS];
            [self->mode_ setState:0 atRow:0 column:0];
            [self->mode_ setState:1 atRow:0 column:1];
            break;
    }

    // ܺ¦ɽ򹹿
    [self->paletteInfoView_ setNeedsDisplay:YES];

    return;
}


// ----------------------------------------- instance - public - ٥Ȥμ
//
// 
//
//  Call
//    evt              : ȯ٥(api ѿ)
//    paletteInfoView_ : ѥåȾܺ(outlet)
//
//  Return
//    None
//
-(void)keyDown:(NSEvent *)evt
{
    // ڡ٥(keyDown)
    if ([evt keyCode] == kVK_Space) {
        if (!([evt isARepeat])) {
            [[NSNotificationCenter defaultCenter]
                postNotificationName:PoCoSpaceKeyEvent
                              object:[NSNumber numberWithBool:YES]];
        }
    } else if ([evt keyCode] == kVK_UpArrow) {
        [self->paletteInfoView_ scrollUp:nil];
        [self->paletteInfoView_ setNeedsDisplay:YES];
    } else if ([evt keyCode] == kVK_DownArrow) {
        [self->paletteInfoView_ scrollDown:nil];
        [self->paletteInfoView_ setNeedsDisplay:YES];
    } else {
        // super class ز
        [super keyDown:evt];
    }

    return;
}


//
// ꡼
//
//  Call
//    evt : ȯ٥(api ѿ)
//
//  Return
//    None
//
-(void)keyUp:(NSEvent *)evt
{
    // ڡ٥(keyUp)
    if ([evt keyCode] == kVK_Space) {
        [[NSNotificationCenter defaultCenter]
            postNotificationName:PoCoSpaceKeyEvent
                          object:[NSNumber numberWithBool:NO]];
    } else if (([evt keyCode] == kVK_UpArrow) ||
               ([evt keyCode] == kVK_DownArrow)) {
        ;
    } else {
        // super class ز
        [super keyUp:evt];
    }

    return;
}


// --------------------------------- instance - public - °ꥷȴϢ
//
// °ꥷȤ򳫤
//  PoCoPaletteWindow  [attribute] button  action
//
//  Call
//    sender                    : ¹о(api ѿ)
//    attributeSheet_           : °ꥷ(outlet)
//    attributeMaskView_        : ɤػ(outlet)
//    attributeDropperView_     : ۤػ(outlet)
//    attributeTransparentView_ : Ʃ(outlet)
//    rangeOfMask_              : ɤػ߻Ϣ³(outlet)
//    rangeOfDropper_           : ۤػ߻Ϣ³(outlet)
//    rangeOfTransparent_       : ƩϢ³(outlet)
//    docCntl_                  : document controller(instance ѿ)
//
//  Return
//    None
//
-(IBAction)raiseAttributeSheet:(id)sender
{
    int l;
    PoCoColor *col;
    PoCoPalette *palette = [[[self->docCntl_ currentDocument] picture] palette];

    DPRINT((@"open attributeSheet\n"));

    // ͤ
    for (l = 0; l < COLOR_MAX; (l)++) {
        col = [palette palette:l];
        [self->attributeMaskView_        setAttribute:[col    isMask] index:l];
        [self->attributeDropperView_     setAttribute:[col noDropper] index:l];
        [self->attributeTransparentView_ setAttribute:[col   isTrans] index:l];
    }
    [self->attributeMaskView_        setNeedsDisplay:YES];
    [self->attributeDropperView_     setNeedsDisplay:YES];
    [self->attributeTransparentView_ setNeedsDisplay:YES];

    // Ϣ³λ
    l = (int)([[NSUserDefaults standardUserDefaults] integerForKey:MASK_SELRANGE]);
    [self->rangeOfMask_       setIntegerValue:l];
    [self->attributeMaskView_ setRange:l];
    l = (int)([[NSUserDefaults standardUserDefaults] integerForKey:DROP_SELRANGE]);
    [self->rangeOfDropper_       setIntegerValue:l];
    [self->attributeDropperView_ setRange:l];
    l = (int)([[NSUserDefaults standardUserDefaults] integerForKey:TRNS_SELRANGE]);
    [self->rangeOfTransparent_       setIntegerValue:l];
    [self->attributeTransparentView_ setRange:l];

    // °ꥷȤ򳫤
    [NSApp beginSheet:self->attributeSheet_
       modalForWindow:[self window]
        modalDelegate:self
       didEndSelector:@selector(attributeSheetDidEnd:returnCode:contextInfo:)
          contextInfo:nil];

    return;
}


//
// °ꥷȤĤ
//  AttributeSheet  [cancel] [set] button  action
//
//  Call
//    sender              : ¹о(api ѿ)
//    attributeSheet_     : °ꥷ(outlet)
//    rangeOfMask_        : ɤػ߻Ϣ³(outlet)
//    rangeOfDropper_     : ۤػ߻Ϣ³(outlet)
//    rangeOfTransparent_ : ƩϢ³(outlet)
//
//  Return
//    None
//
-(IBAction)endAttributeSheet:(id)sender
{
    DPRINT((@"will close attributeSheet\n"));

    // Ϣ³λ򵭲
    [[NSUserDefaults standardUserDefaults] setInteger:[self->rangeOfMask_ integerValue]
                                               forKey:MASK_SELRANGE];
    [[NSUserDefaults standardUserDefaults] setInteger:[self->rangeOfDropper_ integerValue]
                                               forKey:DROP_SELRANGE];
    [[NSUserDefaults standardUserDefaults] setInteger:[self->rangeOfTransparent_ integerValue]
                                               forKey:TRNS_SELRANGE];

    // °ꥷȤĤ
    [self->attributeSheet_ orderOut:sender];
    [NSApp endSheet:self->attributeSheet_
         returnCode:[sender tag]];

    return;
}


//
// °ꥷȤĤ줿
//    sheet 򳫤Ȥ˻ꤷ selector
//
//  Call
//    sheet                     : Ĥsheet (api ѿ)
//    returnCode                : λ֤(api ѿ)
//    contextInfo               : (api ѿ)
//    attributeMaskView_        : ɤػ(outlet)
//    attributeDropperView_     : ۤػ(outlet)
//    attributeTransparentView_ : Ʃ(outlet)
//
//  Return
//    None
//
-(void)attributeSheetDidEnd:(NSWindow *)sheet
                 returnCode:(int)returnCode
                contextInfo:(void *)contextInfo
{
    DPRINT((@"did close attributeSheet : %d\n", returnCode));

    // [set] ȴƤ顢ȿ
    if (returnCode == 0) {
        // 
        [[(PoCoAppController *)([NSApp delegate]) factory]
            createPaletteAttributeSetter:YES
                                   start:0
                                     end:COLOR_MAX
                                    mask:(const BOOL *)([self->attributeMaskView_ attribute])
                                 dropper:(const BOOL *)([self->attributeDropperView_ attribute])
                                   trans:(const BOOL *)([self->attributeTransparentView_ attribute])];
    }

    return;
}


//
// ɤػ߻Ϣ³
//
//  Call
//    sender             : о(api ѿ)
//    attributeMaskView_ : ɤػ(outlet)
//
//  Return
//    None
//
-(void)rangeOfMask:(id)sender
{
    [self->attributeMaskView_ setRange:(int)([sender integerValue])];

    return;
}


//
// ۤػ߻Ϣ³
//
//  Call
//    sender                : о(api ѿ)
//    attributeDropperView_ : ۤػ(outlet)
//
//  Return
//    None
//
-(void)rangeOfDropper:(id)sender
{
    [self->attributeDropperView_ setRange:(int)([sender integerValue])];

    return;
}


//
// ƩϢ³
//
//  Call
//    sender                    : о(api ѿ)
//    attributeTransparentView_ : Ʃ(outlet)
//
//  Return
//    None
//
-(void)rangeOfTransparent:(id)sender
{
    [self->attributeTransparentView_ setRange:(int)([sender integerValue])];

    return;
}


//
// ٤ƤϰϤ
//
//  Call
//    rangeOfMask_        : ɤػ߻Ϣ³(outlet)
//    rangeOfDropper_     : ۤػ߻Ϣ³(outlet)
//    rangeOfTransparent_ : ƩϢ³(outlet)
//
//  Return
//    attributeMaskView_        : ɤػ(outlet)
//    attributeDropperView_     : ۤػ(outlet)
//    attributeTransparentView_ : Ʃ(outlet)
//
-(void)invokeUpdateAllRange
{
    [self->attributeMaskView_        setRange:(int)([self->rangeOfMask_        integerValue])];
    [self->attributeDropperView_     setRange:(int)([self->rangeOfDropper_     integerValue])];
    [self->attributeTransparentView_ setRange:(int)([self->rangeOfTransparent_ integerValue])];

    return;
}


// --------------------------- instance - public - ǡȴϢ
//
// ǡȤ򳫤
//  PoCoPaletteWindow  [gradation] button  action
//
//  Call
//    sender          : ¹о(api ѿ)
//    gradationSheet_ : ǡ(outlet)
//    gradationView_  : оݿΰ(outlet)
//
//  Return
//    None
//
-(IBAction)raiseGradationSheet:(id)sender
{
    DPRINT((@"open gradationSheet\n"));

    // ǡȤ򳫤
    [self->gradationView_ setNeedsDisplay:YES];
    [NSApp beginSheet:self->gradationSheet_
       modalForWindow:[self window]
        modalDelegate:self
       didEndSelector:@selector(gradationSheetDidEnd:returnCode:contextInfo:)
          contextInfo:nil];

    return;
}


//
// ǡȤĤ
//  GradationSheet  [cancel] [set] button  action
//
//  Call
//    sender          : ¹о(api ѿ)
//    gradationSheet_ : ǡ(outlet)
//
//  Return
//    None
//
-(IBAction)endGradationSheet:(id)sender
{
    DPRINT((@"will close gradationSheet\n"));

    // ǡȤĤ
    [self->gradationSheet_ orderOut:sender];
    [NSApp endSheet:self->gradationSheet_
         returnCode:[sender tag]];

    return;
}


//
// ǡȤĤ줿
//  sheet 򳫤Ȥ˻ꤷ selector
//
//  Call
//    sheet          : Ĥshee (api ѿ)
//    returnCode     : λ֤(api ѿ)
//    contextInfo    : (api ѿ
//    gradationView_ : оݿΰ(outlet)
//
//  Return
//    None
//
-(void)gradationSheetDidEnd:(NSWindow *)sheet
                 returnCode:(int)returnCode
                contextInfo:(void *)contextInfo
{
    DPRINT((@"did close gradationSheet : %d\n", returnCode));

    // [set] ȴƤ顢ȿ
    if (returnCode == 0) {
        [[(PoCoAppController *)([NSApp delegate]) factory]
            createPaletteGradationMaker:YES
                                  start:[self->gradationView_ leftNum]
                                    end:[self->gradationView_ rightNum]];
    }

    return;
}


// ----------------------------- instance - public - ѥåؤȴϢ
//
// ѥåؤȤ򳫤
//  PoCoPaletteWindow  [gradation] button  action
//
//  Call
//    sender         : ¹о(api ѿ)
//    exchangeSheet_ : ѥåؤ(outlet)
//    exchangeView_  : ѥåؤΰ(outlet)
//
//  Return
//    None
//
-(IBAction)raiseExchangeSheet:(id)sender
{
    DPRINT((@"open exchangeSheet\n"));

    // 
    [self->exchangeView_ setUp];

    // ѥåؤȤ򳫤
    [self->exchangeView_ setNeedsDisplay:YES];
    [NSApp beginSheet:self->exchangeSheet_
       modalForWindow:[self window]
        modalDelegate:self
       didEndSelector:@selector(exchangeSheetDidEnd:returnCode:contextInfo:)
          contextInfo:nil];

    return;
}


//
// ѥåؤȤĤ
//  GradationSheet  [cancel] [set] button  action
//
//  Call
//    sender         : ¹о(api ѿ)
//    exchangeSheet_ : ѥåؤ(outlet)
//
//  Return
//    None
//
-(IBAction)endExchangeSheet:(id)sender
{
    DPRINT((@"will close exchangeSheet\n"));

    // ѥåؤȤĤ
    [self->exchangeSheet_ orderOut:sender];
    [NSApp endSheet:self->exchangeSheet_
         returnCode:[sender tag]];

    return;
}


//
// ѥåؤȤĤ줿
//  sheet 򳫤Ȥ˻ꤷ selector
//
//  Call
//    sheet         : Ĥsheet(api ѿ)
//    returnCode    : λ֤(api ѿ)
//    contextInfo   : (api ѿ)
//    exchangeView_ : ѥåؤΰ(outlet)
//
//  Return
//    None
//
-(void)exchangeSheetDidEnd:(NSWindow *)sheet
                returnCode:(int)returnCode
               contextInfo:(void *)contextInfo
{
    id  cntl;
    BOOL upd;
    NSEnumerator *iter;
    PoCoPaletteExchangePair *pair;
    PoCoControllerFactory *factory;

    DPRINT((@"did close exchangeSheet : %d\n", returnCode));

    // [set] ȴƤ顢ȿ
    if (returnCode != 0) {
        goto EXIT;
    }

    // factory μ
    factory = [(PoCoAppController *)([NSApp delegate]) factory];

    // Ƥʬ༡¹
    iter = [[self->exchangeView_ exchangeTable] objectEnumerator];
    pair = [iter nextObject];
    upd = NO;
    for ( ; pair != nil; pair = [iter nextObject]) {
       cntl = [factory createPictureColorExchangerPassive:NO
                                                      src:[pair src]
                                                      dst:[pair dst]];
       upd |= [cntl execute];
       [cntl release];
    }

    // 
    if (upd) {
        [factory createPaletteNoEditter:YES
                                   name:[[NSBundle mainBundle] localizedStringForKey:@"PaletteExchange"
                                                                               value:@"exchange palette"
                                                                               table:nil]
                                  index:-1];
    }

EXIT:
    return;
}


// --------------------------------- instance - public - ѥåʣ̥ȴϢ
//
// ѥåʣ̥Ȥ򳫤
//  PoCoPaletteWindow  [gradation] button  action
//
//  Call
//    sender      : ¹о(api ѿ)
//    pasteSheet_ : ѥåʣ̥(outlet)
//    pasteView_  : ѥåʣΰ(outlet)
//
//  Return
//    None
//
-(IBAction)raisePasteSheet:(id)sender
{
    DPRINT((@"open pasteSheet\n"));

    // 
    [self->pasteView_ setUp];

    // ѥåʣ̥Ȥ򳫤
    [self->pasteView_ setNeedsDisplay:YES];
    [NSApp beginSheet:self->pasteSheet_
       modalForWindow:[self window]
        modalDelegate:self
       didEndSelector:@selector(pasteSheetDidEnd:returnCode:contextInfo:)
          contextInfo:nil];

    return;
}


//
// ѥåʣ̥ȤĤ
//  GradationSheet  [cancel] [set] button  action
//
//  Call
//    sender      : ¹о(api ѿ)
//    pasteSheet_ : ѥåʣ̥(outlet)
//
//  Return
//    None
//
-(IBAction)endPasteSheet:(id)sender
{
    DPRINT((@"will close pasteSheet\n"));

    // ѥåʣ̥ȤĤ
    [self->pasteSheet_ orderOut:sender];
    [NSApp endSheet:self->pasteSheet_
         returnCode:[sender tag]];

    return;
}


//
// ѥåʣ̥ȤĤ줿
//  sheet 򳫤Ȥ˻ꤷ selector
//
//  Call
//    sheet       : Ĥsheet(api ѿ)
//    returnCode  : λ֤(api ѿ)
//    contextInfo : (api ѿ)
//    pasteView_  : ѥåʣΰ(outlet)
//
//  Return
//    None
//
-(void)pasteSheetDidEnd:(NSWindow *)sheet
                returnCode:(int)returnCode
               contextInfo:(void *)contextInfo
{
    id  cntl;
    BOOL upd;
    NSEnumerator *iter;
    PoCoPaletteExchangePair *pair;
    PoCoControllerFactory *factory;

    DPRINT((@"did close pasteSheet : %d\n", returnCode));

    // [set] ȴƤ顢ȿ
    if (returnCode != 0) {
        goto EXIT;
    }

    // factory μ
    factory = [(PoCoAppController *)([NSApp delegate]) factory];

    // Ƥʬ༡¹
    iter = [[self->pasteView_ exchangeTable] objectEnumerator];
    pair = [iter nextObject];
    upd = NO;
    for ( ; pair != nil; pair = [iter nextObject]) {
       cntl = [factory createPictureColorPasterPassive:NO
                                                   src:[pair src]
                                                   dst:[pair dst]];
       upd |= [cntl execute];
       [cntl release];
    }

    // 
    if (upd) {
        [factory createPaletteNoEditter:YES
                                   name:[[NSBundle mainBundle] localizedStringForKey:@"PalettePaste"
                                                                               value:@"paste palette"
                                                                               table:nil]
                                  index:-1];
    }

EXIT:
    return;
}

@end
