//
//	Pelistina on Cocoa - PoCo -
//	2ͥѥԽΰ
//
//	Copyright (C) 2005-2015 KAENRYUU Koutoku.
//

#import "PoCoMonochromePatternEditView.h"

#import "PoCoMonochromePattern.h"
#import "PoCoMonochromePatternSampleView.h"

// 
static unsigned int PIX_SIZE = 8;       // 1pixel ΰ(dot ñ)

// ============================================================================
@implementation PoCoMonochromePatternEditView

// --------------------------------------------------------- instance - private
//
// 줿ɸѴ
//
//  Call
//    p  : ºɸ(̾)
//    rb : YES : 
//         NO  : 
//
//  Return
//    function : ѥɸ
//
-(PoCoPoint *)toNormalPos:(PoCoPoint *)p
            isRightBottom:(BOOL)rb
{
    PoCoPoint *np = [[PoCoPoint alloc] initX:([p x] / PIX_SIZE)
                                       initY:([p y] / PIX_SIZE)];

    if (rb) {
        [np moveX:-1 moveY:-1];
    }

    return np;
}


//
// ºɸѴ
//
//  Call
//    x : ѥX
//    y : ѥY
//
//  Return
//    r : ΰ
//
-(NSRect)toRealRectX:(unsigned int)x
               withY:(unsigned int)y
{
    NSRect r;

    r.origin.x = (float)(x * PIX_SIZE);
    r.origin.y = (float)(y * PIX_SIZE);
    r.size.width = (float)(PIX_SIZE);
    r.size.height = (float)(PIX_SIZE);

    return r;
}


//
// 1ʬɽ
//
//  Call
//    pat : оݥѥ
//    rb  : pat  rowbytes
//    x   : ѥX
//    y   : ѥY
//
//  Return
//    None
//
-(void)drawPoint:(const unsigned char *)pat
        rowBytes:(unsigned int)rb
             atX:(unsigned int)x
             atY:(unsigned int)y
{
    if (pat[y * rb + x] != 0) {
        [[NSColor whiteColor] set];
    } else {
        [[NSColor blackColor] set];
    }
    [NSBezierPath fillRect:[self toRealRectX:x withY:y]];

    return;
}


// ---------------------------------------------------------- instance - public
//
// initialize
//
//  Call
//    frame : ΰ(api ѿ)
//
//  Return
//    function : 
//    isSet_   : ޥꤹ뤫ݤ(instance ѿ)
//    pattern_ : Խоݥѥ(instance ѿ)
//
-(id)initWithFrame:(NSRect)frame
{
    DPRINT((@"[PoCoMonochromePatternEditView initWithFrame]\n"));

    // super class ν
    self = [super initWithFrame:frame];

    // Ȥν
    if (self != nil) {
        self->isSet_ = NO;
        self->pattern_ = nil;

        // ԽоݤΥѥ(ʣ)
        self->pattern_ = [[PoCoMonochromePattern alloc] init];
        if (self->pattern_ == nil) {
            DPRINT((@"can't alloc PoCoMonochromePattern\n"));
            [self release];
            self = nil;
        }
    }

    return self;
}


//
// deallocate
//
//  Call
//    None
//
//  Return
//    pattern_ : Խоݥѥ(instance ѿ)
//
-(void)dealloc
{
    DPRINT((@"[PoCoMonochromePatternEditView dealloc]\n"));

    // 񸻤β
    [self->pattern_ release];
    self->pattern_ = nil;

    // super class β
    [super dealloc];

    return;
}


//
// ѥ
//
//  Call
//    pattern_ : Խоݥѥ(instance ѿ)
//
//  Return
//    function : ѥ
//
-(PoCoMonochromePattern *)pattern
{
    return self->pattern_;
}


//
// ѥ
//
//  Call
//    pat         : ѥ
//    sampleView_ : ץɽΰ(outlet)
//
//  Return
//    pattern_ : Խоݥѥ(instance ѿ)
//
-(void)setPattern:(PoCoMonochromePattern *)pat
{
    DPRINT((@"[PoCoMonochromePatternEditView setPattern]\n"));

    [self->pattern_ setPattern:[pat pattern]
                         width:[pat width]
                        height:[pat height]];
    [self setNeedsDisplay:YES];

    // ¦
    [self->sampleView_ setPattern:self->pattern_];
    [self->sampleView_ setNeedsDisplay:YES];

    return;
}


//
// ɽ׵
//
//  Call
//    rect    : ΰ(api ѿ)
//    pattern : Խоݥѥ(instance ѿ)
//
//  Return
//    None
//
-(void)drawRect:(NSRect)rect
{
    unsigned int x;
    unsigned int y;
    PoCoRect *r;
    PoCoPoint *lt;
    PoCoPoint *rb;
    const unsigned char *pat = [self->pattern_ pattern];
    const unsigned int rbyte = ([self->pattern_ width] + ([self->pattern_ width] & 1));

    // ϰϤλ
    r = [[PoCoRect alloc] initNSRect:rect];
    lt = [self toNormalPos:[r  lefttop]
             isRightBottom:NO];
    rb = [self toNormalPos:[r rightbot]
             isRightBottom:YES];
    if ((r == nil) || (lt == nil) || (rb == nil)) {
        ;
    } else {
        if (([lt x] == [rb x]) && ([lt y] == [rb y])) {
            [self drawPoint:pat
                   rowBytes:rbyte
                        atX:[lt x]
                        atY:[lt y]];
        } else if ([lt x] == [rb x]) {
            for (y = [lt y]; y <= [rb y]; (y)++) {
                [self drawPoint:pat
                       rowBytes:rbyte
                            atX:[lt x]
                            atY:y];
            }
        } else if ([lt y] == [rb y]) {
            for (x = [lt x]; x <= [rb x]; (x)++) {
                [self drawPoint:pat
                       rowBytes:rbyte
                            atX:x
                            atY:[lt y]];
            }
        } else {
            for (y = [lt y]; y <= [rb y]; (y)++) {
                for (x = [lt x]; x <= [rb x]; (x)++) {
                    [self drawPoint:pat
                           rowBytes:rbyte
                                atX:x
                                atY:y];
                }
            }
        }
    }
    [r release];
    [lt release];
    [rb release];

    return;
}


//
// ɸϤȿž
//
//  Call
//    None
//
//  Return
//    function : YES
//
-(BOOL)isFlipped
{
    return YES;
}


// --------------------------------- instance - public - ޥ٥ȽϢ
//
// ܥ
//
//  Call
//    evt      : ȯ٥(api ѿ)
//    pattern_ : Խоݥѥ(instance ѿ)
//
//  Return
//    isSet_ : ޥꤹ뤫ݤ(instance ѿ)
//
-(void)mouseDown:(NSEvent *)evt
{
    NSPoint p = [self convertPoint:[evt locationInWindow] fromView:nil];
    PoCoPoint *tp;
    PoCoPoint *dp;
    const unsigned char *pat = [self->pattern_ pattern];
    const unsigned int rb = ([self->pattern_ width] + ([self->pattern_ width] & 1));

    if ([self mouse:p inRect:[self bounds]]) {
        tp = [[PoCoPoint alloc] initNSPoint:p];
        dp = [self toNormalPos:tp
                 isRightBottom:NO];
        self->isSet_ = ((pat[[dp y] * rb + [dp x]] != 0x00) ? NO : YES);
        [tp release];
        [dp release];

        // ɥåˤޤ魯
        [self mouseDragged:evt];
    }

    return;
}


//
// ɥå
//  ľ pattern Ƥ
//
//  Call
//    evt         : ȯ٥(api ѿ)
//    isSet_      : ޥꤹ뤫ݤ(instance ѿ)
//    sampleView_ : ץɽΰ(outlet)
//
//  Return
//    pattern_ : Խоݥѥ(instance ѿ)
//
-(void)mouseDragged:(NSEvent *)evt
{
    NSPoint p = [self convertPoint:[evt locationInWindow] fromView:nil];
    PoCoPoint *tp;
    PoCoPoint *dp;
    unsigned char *pat = (unsigned char *)([self->pattern_ pattern]);
    const unsigned int rb = ([self->pattern_ width] + ([self->pattern_ width] & 1));

    if ([self mouse:p inRect:[self bounds]]) {
        tp = [[PoCoPoint alloc] initNSPoint:p];
        dp = [self toNormalPos:tp
                 isRightBottom:NO];
        pat[[dp y] * rb + [dp x]] = ((self->isSet_) ? 0x01 : 0x00);
        [self setNeedsDisplayInRect:[self toRealRectX:[dp x] withY:[dp y]]];
        [self->sampleView_ setNeedsDisplay:YES];
        [tp release];
        [dp release];
    }

    return;
}

@end
